/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _NNTPGRAB_UTILS_H_
#define _NNTPGRAB_UTILS_H_

#include "nntpgrab_types.h"

#ifdef  __cplusplus
extern "C" {
#endif

/* These declarations are actually from GLib, but to easy the development
 * of external frontends, they have been duplicated here to avoid linkage bloat */

typedef void NGFileMonitor;

/**
 * Free data allocated by NNTPGrab functions
 *
 * @param data      The data which needs to be free'd
 */
void          ng_free(void *data);

/**
 * Append an item to a list
 *
 * @param list      The list where the item needs to be added to (can be NULL to create a new list)
 * @param data      The item which needs to be added
 * @return          The (possibly) new start of the list
 */
NGList       *ng_list_append  (NGList *list, void *data);

/**
 * Insert an item to the beginning of a list
 *
 * @param list      The list where the item needs to be added to (can be NULL to create a new list)
 * @param data      The item which needs to be added
 * @return          The (possibly) new start of the list
 */
NGList       *ng_list_prepend (NGList *list, void *data);

/**
 * Remove an item from a list
 *
 * @param list      The list where the item needs to be removed from
 * @param data      The item which needs to be removed
 * @return          The (possibly) new start of the list
 *
 * @note            If data isn't part of the list, this function will do nothing
 */
NGList       *ng_list_remove  (NGList *list, void *data);

/**
 * Retrieve the length of a list
 *
 * @param list      The list of which the length must be returned
 * @return          The length of the list
 */
unsigned int  ng_list_length  (NGList *list);

/**
 * Free a list
 *
 * @param list      The list which needs to be free'd
 *
 * @note            This function only free's the list itself, NOT the items which were in it!
 */
void          ng_list_free    (NGList *list);

#define ng_list_previous(list)           ((list) ? (((NGList *)(list))->prev) : NULL)
#define ng_list_next(list)               ((list) ? (((NGList *)(list))->next) : NULL)

/* signals */
/**
 * Register a signal handler
 *
 * @param instance      The instance of an GObject, for example NNTPGrabCore or NNTPGrabAutoImport
 * @param signal_name   The name of the signal for which a signaler handler needs to be registered
 * @param cb_handler    The function which need to be called every time the given signal is emit
 * @param data          Data used as extra argument in the callback functions
 *
 * @note                The file nntpgrab_core/marshall.list contains a list of known signals and callback function prototypes for the NNTPGrabCore
 */
void     ng_signal_connect(void *instance, const char *signal_name, NGCallback cb_handler, void *data);

/**
 * Un-register a signal handler
 *
 * @param instance      The instance of an GObject, for example NNTPGrabCore or NNTPGrabAutoImport
 * @param cb_handler    The callback function which need to be unregistered
 * @param data          The callback data used for the given cb_handler
 */
void     ng_signal_handlers_disconnect_by_func(void *instance, NGCallback cb_handler, void *data);

/**
 * Temporary disable a signal handler for a given callback function
 *
 * @param instance      The instance of an GObject, for example NNTPGrabCore or NNTPGrabAutoImport
 * @param cb_handler    The callback function which need to be temporary blocked
 * @param data          The callback data used for the given cb_handler
 */
void     ng_signal_handlers_block_by_func(void *instance, NGCallback cb_handler, void *data);

/**
 * Re-enable a signal handler for a given callback function
 *
 * @param instance      The instance of an GObject, for example NNTPGrabCore or NNTPGrabAutoImport
 * @param cb_handler    The callback function which need to be un-blocked
 * @param data          The callback data used for the given cb_handler
 */
void     ng_signal_handlers_unblock_by_func(void *instance, NGCallback cb_handler, void *data);

/**
 * Start the event handler loop. This function waits for incoming events and passes these on to all the registered event handlers
 *
 * @note                This function only returns as soon as the function ng_event_handler_loop_quit() is called from an event handler or another thread
 * @note                This function doesn't need to be used for the GTK+ frontend as it is equivalent to gtk_main()
 */
void     ng_event_handler_loop_run(void);

/**
 * Stop the event handler loop
 *
 * @note                This function doesn't need to be used for the GTK+ frontend as it is equivalent to gtk_main_quit()
 */
void     ng_event_handler_loop_quit(void);

/* End of GLib declarations */

typedef struct _nntpgrab_folder {
    char folder[2048];
    ngboolean has_subfolders;
} NNTPGrabFolder;

typedef struct _nzb {
    NGList *files;
} NZB;

typedef struct _nzb_file {
    char subject[256];
    char poster[256];
    time_t stamp;
    nguint64 filesize;
    NGList *groups;
    NGList *segments;
} NZBFile;

typedef struct _nzbcreator_group {
    char newsgroup[512];
    int group_id;
} NZBCreatorGroup;

typedef enum _nzbcreator_search_only_in {
    NZBCREATOR_SEARCH_ONLY_IN_SUBJECTS,
    NZBCREATOR_SEARCH_ONLY_IN_FILENAMES,
    NZBCREATOR_SEARCH_ONLY_IN_POSTERS
} NZBCreatorSearchOnlyIn;

typedef struct _nzbcreator_search_opts {
    char query[64];
    char username[64];
    char password[64];
    NZBCreatorSearchOnlyIn search_only_in;
    int max_age;
    ngint64 minimal_file_size;
    ngint64 maximal_file_size;
    int percentage_complete;
    int group_to_search;
    NNTPFileType file_type;
} NZBCreatorSearchOpts;

typedef struct _nzbcreator_file {
    ngint64 file_id;
    char subject[256];
    char poster[256];
    int num_parts_found;
    int num_parts_expected;
    NNTPFileType file_type;
    ngint64 file_size;
    time_t stamp;
    int complete_percentage;
} NZBCreatorFile;

typedef struct _nzbcreator_collection {
    char collection_name[256];
    char newsgroup[256];
    ngint64 total_file_size;
    time_t stamp;
    NGList *files;                   // List of NZBCreatorFile* instances
} NZBCreatorCollection;

typedef struct _nzbcreator_search_result {
    int number_of_hits;
    int number_of_results;
    NGList *collections;             // List of NZBCreatorCollection* instances
} NZBCreatorSearchResult;

/**
 * Compile a regular expression
 *
 * @param regex     The (perl-compatible) regular expression which needs to be compiled
 * @return          A compiled regular expression
 */
NGRegex     *nntpgrab_utils_regex_compile(const char *regex);

/**
 * Free a compiled regular expression
 *
 * @param re        The compiled regular expression
 */
void         nntpgrab_utils_regex_free(NGRegex *re);

/**
 * Perform a match against a compiled regular expression
 *
 * @param re        The compiled regular expression
 * @param line      The data which need to be matched against the compiled regular expression
 * @return          A NULL-terminated array of matches items (needs to be free'd using nntpgrab_utils_regex_matches_free())
 */
const char **nntpgrab_utils_regex_match(NGRegex *re, const char *line);

/**
 * Free an array of matched items
 *
 * @param matches   The NULL-terminated array of matches as returned by the function nntpgrab_utils_regex_match()
 */
void         nntpgrab_utils_regex_matches_free(const char **matches);

/**
 * Strip a subject into useable pieces of data
 *
 * @param subject                   The subject which need to be stripped
 * @param subject_without_partnum   Location for the subject without any part numbers
 * @param file_num                  Location for the number of the file in the set
 * @param total_files               Location for the total number of files in the set
 * @param filename                  Location for the filename
 * @param extension                 Location for the extension of the filename (as an string)
 * @param file_type                 Location for the extension of the filename (as an enumeration)
 * @param par2_startnum             Location for the start number of the PAR2 recovery file
 * @param num_par2_blocks           Location for the number blocks in the PAR2 recovery file
 * @param part_num                  Location for the part number for this subject
 * @param total_parts               Location for the number of parts for this subject
 * @return                          TRUE is the strip was successfull
 *
 * @note                            All the parameters (except subject) can be NULL to ignore them
 * @note                            All the char* parameters need to be free'd using ng_free()
 */
ngboolean        nntpgrab_utils_strip_subject(const char *subject, char **subject_without_partnum, int *file_num, int *total_files, char **filename, char **extension, NNTPFileType *file_type, int *par2_startnum, int *num_par2_blocks, int *part_num, int *total_parts);

/**
 * Find out the file type of a given filename
 *
 * @param filename          The filename whose file type needs to be found out
 * @return                  The file type (as an enumeration)
 *
 * @note                    If the file type couldn't be found out, the value NNTP_FILE_TYPE_UNKNOWN will be returned
 */
NNTPFileType     nntpgrab_utils_get_file_type_of_filename(const char *filename);

/**
 * Transform a file size into a human readable notation
 *
 * @param file_size         The file size which needs to be transformed
 * @return                  The human readable notation for the file size. Needs to be free'd using ng_free()
 */
char            *nntpgrab_utils_calculate_file_size(nguint64 file_size);

/**
 * Calculate the estimated time remaining to complete the given file size
 *
 * @param bytes_received1       The number of bytes received in now() - 10
 * @param bytes_received2       The number of bytes received in now() - 9
 * @param bytes_received3       The number of bytes received in now() - 8
 * @param bytes_received4       The number of bytes received in now() - 7
 * @param bytes_received5       The number of bytes received in now() - 6
 * @param bytes_received6       The number of bytes received in now() - 5
 * @param bytes_received7       The number of bytes received in now() - 4
 * @param bytes_received8       The number of bytes received in now() - 3
 * @param bytes_received9       The number of bytes received in now() - 2
 * @param bytes_received10      The number of bytes received in now() - 1
 * @param file_size             The file size
 * @return                      The number of seconds in which the given file_size can be downloaded
 *
 * @note                        The bytes_received parameters can be retrieved from the signalhandler for the signal 'traffic_monitor_update'
 */
int              nntpgrab_utils_calculate_estimated_time_remaining(int bytes_received1, int bytes_received2, int bytes_received3, int bytes_received4, int bytes_received5, int bytes_received6, int bytes_received7, int bytes_received8, int bytes_received9, int bytes_received10, nguint64 file_size);

/**
 * Transform a estimated time remaining into a human readable value for the time remaining
 *
 * @param estimated_time_remaining      The estimated time remaining (in seconds)
 * @return                              The human readable notation (like '3 minutes and 10 seconds'). Needs to be free'd using ng_free()
 */
char            *nntpgrab_utils_get_readable_time_remaining(int estimated_time_remaining);

/**
 * Transform a estimated time remaining into a human readable value for the expected time to finish
 *
 * @param estimated_time_remaining      The estimated time remaining (in seconds)
 * @return                              The human readable notation (like 'Friday Februari 6 2009 - 22:15'). Needs to be free'd using ng_free()
 */
char            *nntpgrab_utils_get_readable_finished_time(int estimated_time_remaining);

/**
 * Check whether the given text contains invalid UTF-8 characters and if they're found, replace them with something more sane
 *
 * @param text          The text which need to be sanitized
 * @param length        The length of the text
 */
void             nntpgrab_utils_sanitize_text(char *text, int length);

/**
 * Remove the extension from a filename
 *
 * @param filename      The filename whose extension need to be stripped
 */
void             nntpgrab_utils_strip_nzb_extension(char *filename);

/**
 * Remove forbidden characters from the collection name
 *
 * @param collection_name   The collection name which need to be sanitized
 */
void             nntpgrab_utils_sanitize_collection_name(char *collection_name);

/**
 * Retrieves a list containing all the sub-folders which are in the given folder
 *
 * @param parent        The name of the directory whose sub-folders should be retrieved
 * @param folders       Pointer to a NGList*. The list of sub-folders will be placed in this field. This list needs to be free'd using the function nntpgrab_config_free_folder_listing()
 * @return              TRUE on sucess, FALSE if the given directory isn't found or readable
 */
ngboolean        nntpgrab_utils_get_folder_listing(const char *parent, NGList **folders);

/**
 * Free a list of sub-folders
 *
 * @param folders       A list containing subfolders as returned by the function nntpgrab_config_get_folder_listing()
 */
void             nntpgrab_utils_free_folder_listing(NGList *folders);

/**
 * Parse the contents of a NZB file and put the information in a structure
 *
 * @param contents      The contents of the NZB file
 * @param errmsg        Pointer to a char*. If an errors occurs, the reason will be placed in this field. Needs to be freed using ngfree(). Can be NULL to ignore errors
 * @return              A structure containing information about the NZB file. Needs to be free'd using the function nntpgrab_utils_nzb_file_free(). If the value NULL is returned, an error occured and errmsg will be set
 */
NZB             *nntpgrab_utils_parse_nzb_file(const char *contents, char **errmsg);

/**
 * Free a structure containing details about a NZB file
 *
 * @param nzbfile       The structure containing information about the NZB file
 */
void             nntpgrab_utils_nzb_file_free(NZB *nzbfile);

/**
 * Retrieve a list of all the usenet groups which are indexed by the NZBCreator service
 *
 * @param errmsg        Pointer to a char*. If an errors occurs, the reason will be placed in this field. Needs to be freed using ngfree(). Can be NULL to ignore errors
 * @return              A list of all the indexed usenet groups. Needs to be free'd using nntpgrab_utils_nzbcreator_free_groups(). If an error occures, the value NULL will be returned and errmsg will be set
 */
NGList                  *nntpgrab_utils_nzbcreator_get_all_groups(char **errmsg);

/**
 * Free the list of usenet groups which are indexed by the NZBCreator service
 *
 * @param groups        A list of all the indexed usenet groups
 */
void                     nntpgrab_utils_nzbcreator_free_groups(NGList *groups);

/**
 * Perform a search using the NZBCreator service
 *
 * @param opts          A structure containing information about the requested search
 * @param errmsg        Pointer to a char*. If an errors occurs, the reason will be placed in this field. Needs to be freed using ngfree(). Can be NULL to ignore errors
 * @return              A structure containing the search results. Needs to be free'd using nntpgrab_utils_nzbcreator_free_result(). If an error occurs, the value NULL will be returned and errmsg will be set
 */
NZBCreatorSearchResult  *nntpgrab_utils_nzbcreator_perform_search(NZBCreatorSearchOpts opts, char **errmsg);

/**
 * Free the NZBCreator search results
 *
 * @param result        The NZBCreator search results
 */
void                     nntpgrab_utils_nzbcreator_free_result(NZBCreatorSearchResult *result);

/**
 * Generate a NZB file based on a list of file ID's
 *
 * @param file_ids      A list containing file ID's of all the files which needs to be retrieved
 * @param errmsg        Pointer to a char*. If an errors occurs, the reason will be placed in this field. Needs to be freed using ngfree(). Can be NULL to ignore errors
 * @return              The contents of the NZB file containing all the details of the given file_ids. Needs to be free'd using ng_free(). If an error occurs, the value NULL will be returned and errmsg will be set
 */
char                    *nntpgrab_utils_nzbcreator_generate_NZB(NGList *file_ids, char **errmsg);

/**
 * Start the monitoring of a directory
 *
 * @param path          The directory which needs to be monitored for changes
 * @param errmsg        Pointer to a char*. If an errors occurs, the reason will be placed in this field. Needs to be freed using ngfree(). Can be NULL to ignore errors
 * @return              An instance of a NGFileMonitor or NULL in case an error occured (errmsg will be set)
 *
 * @note                The NGFileMonitor is an GObject where the event 'changed' can be caught using ng_signal_connect()
 */
NGFileMonitor   *nntpgrab_utils_monitor_directory(const char *path, char **errmsg);

/**
 * Cancel the monitoring of a directory
 *
 * @param monitor       The instance of the NGFileMonitor which needs to be stopped
 */
void             nntpgrab_utils_cancel_directory_monitor(NGFileMonitor *monitor);

#ifdef  __cplusplus
}
#endif

#endif /* _NNTPGRAB_UTILS_H_ */

