/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include "gui.h"
#include "config_gui.h"
#include "nntpgrab_utils.h"

#if GLIB_CHECK_VERSION(2,16,0)
// auto_import.c
void    enable_auto_import(void);
void    disable_auto_import(void);

static gboolean scroll_to_cell = FALSE;
static gboolean initialized = FALSE;

enum {
    TREE_LOCAL_DIRECTORY,
    TREE_SERVER_DIRECTORY
};
#endif

G_MODULE_EXPORT void
on_chkAutoImportActivateLocal_toggled(GtkWidget *caller, gpointer data)
{
#if GLIB_CHECK_VERSION(2,16,0)
    ConfigGUIOpts opts;
    gboolean val;

    val = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(caller));

    opts = config_gui_get_opts();
    opts.enable_auto_import = val;
    scroll_to_cell = FALSE;
    config_gui_set_opts(opts);
#endif
}

G_MODULE_EXPORT void
on_chkAutoImportActivateServer_toggled(GtkWidget *caller, gpointer data)
{
#if GLIB_CHECK_VERSION(2,16,0)
    ConfigOpts opts;
    gboolean val;

    val = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(caller));

    opts = nntpgrab_config_get_opts(glue);
    opts.enable_auto_import = val;
    scroll_to_cell = FALSE;
    nntpgrab_config_set_opts(glue, opts);
#endif
}

G_MODULE_EXPORT void
on_radAutoImportMove_toggled(GtkWidget *caller, gpointer data)
{
#if GLIB_CHECK_VERSION(2,16,0)
    GtkWidget *radAutoImportMove = nntpgrab_gui_base_get_widget("radAutoImportMove");
    ConfigOpts opts;

    opts = nntpgrab_config_get_opts(glue);
    opts.move_file_after_auto_import = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(radAutoImportMove));
    scroll_to_cell = FALSE;
    nntpgrab_config_set_opts(glue, opts);
#endif
}

#if GLIB_CHECK_VERSION(2,16,0)
static char *
retrieve_path(GtkTreeModel *model, GtkTreeIter tree_iter)
{
    GtkTreeIter iter = tree_iter;
    GtkTreeIter child;
    GtkTreeIter iter_tmp;
    char *path = NULL;
    char *dir_separator = "/";

    // Find out of the server is a UNIX or a Win32 environment
    if (gtk_tree_model_get_iter_first(model, &iter_tmp)) {
        char *root = NULL;

        gtk_tree_model_get(model, &iter_tmp, 0, &root, -1);

        if (root && !strcmp(root, "/")) {
            dir_separator = "/";
        } else {
            dir_separator = "\\";
        }

        g_free(root);
    }

    do {
        char *dirname = NULL;
        char *tmp;

        child = iter;

        gtk_tree_model_get(model, &child, 0, &dirname, -1);

        if (path) {
            if (!strcmp(dirname, "/")) {
                // Don't use the dir separator for the root
                tmp = g_strdup_printf("%s%s", dirname, path);
                g_free(path);
                path = tmp;
            } else {
                tmp = g_strdup_printf("%s%s%s", dirname, dir_separator, path);      // We walk the path from right to left
                g_free(path);
                path = tmp;
            }
        } else {
            path = g_strdup(dirname);
        }

        g_free(dirname);
    } while (gtk_tree_model_iter_parent(model, &iter, &child));

    return path;
}

static void
on_row_expanded(GtkTreeView *tree_view, GtkTreeIter *tree_iter, GtkTreePath *tree_path, gpointer user_data)
{
    char *path = NULL;
    GtkTreeModel *model = gtk_tree_view_get_model(tree_view);
    NGList *folders = NULL;
    GtkTreeIter iter;
    gboolean local = (GPOINTER_TO_INT(user_data) == TREE_LOCAL_DIRECTORY);
    gboolean ret;

    memcpy(&iter, tree_iter, sizeof(GtkTreeIter));

    path = retrieve_path(model, iter);

    if (local) {
        ret = nntpgrab_utils_get_folder_listing(path, &folders);
    } else {
        ret = nntpgrab_config_get_folder_listing(glue, path, &folders);
    }

    if (ret) {
        NGList *list = folders;
        GtkTreeIter iter;

        while (list) {
            NNTPGrabFolder *folder = (NNTPGrabFolder*) list->data;
            gboolean found = FALSE;
            gboolean create_new = TRUE;

            // Do we already have a child with this name ?
            if (gtk_tree_model_iter_children(model, &iter, tree_iter)) {
                char *folder2 = NULL;

                do {
                    gtk_tree_model_get(model, &iter, 0, &folder2, -1);

                    if (!folder2) {
                        create_new = FALSE;
                        break;
                    }

                    if (folder2 && !strcmp(folder->folder, folder2)) {
                        found = TRUE;
                    }

                    g_free(folder2);
                } while (gtk_tree_model_iter_next(model, &iter));
            }

            if (found) {
                list = ng_list_next(list);
                continue;
            }

            if (create_new) {
                gtk_tree_store_append(GTK_TREE_STORE(model), &iter, tree_iter);
            }

            gtk_tree_store_set(GTK_TREE_STORE(model), &iter, 0, folder->folder, -1);

            if (folder->has_subfolders && !gtk_tree_model_iter_has_child(model, &iter)) {
                GtkTreeIter iter_ignore;
                gtk_tree_store_append(GTK_TREE_STORE(model), &iter_ignore, &iter);
            }

            list = ng_list_next(list);
        }

        if (local) {
            nntpgrab_utils_free_folder_listing(folders);
        } else {
            nntpgrab_config_free_folder_listing(glue, folders);
        }
    }

    g_free(path);
}

static void
update_folder_listing(GtkTreeView *tree, GtkTreeStore *store, GtkTreeSelection *selection, const char *directory, gboolean local)
{
    char **values;
    gboolean is_win32_path = FALSE;
    char *dir_separator;
    int i;
    GtkTreeIter parent;
    GtkTreePath *tree_path;
    char *path = NULL;

    gtk_tree_store_clear(store);

    if (strlen(directory) >= 3 && directory[1] == ':' && directory[2] == '\\') {
        is_win32_path = TRUE;
        dir_separator = "\\";
    } else {
        dir_separator = "/";
    }

    memset(&parent, 0, sizeof(GtkTreeIter));

    values = g_strsplit(directory, dir_separator, 0);
    i = 0;
    while (values[i]) {
        GtkTreeIter iter;
        NGList *folders;
        NGList *list;
        GtkTreeIter temp;
        gboolean already_has_child;

        // Ignore empty paths
        if (strlen(values[i]) == 0 && i > 0) {
            i++;
            continue;
        }

        // build the path (necessary for the nntpgrab_config_get_folder_listing() call)
        if (path) {
            char *tmp;
            if (dir_separator[0] == '/' && i == 1) {
                tmp = g_strdup_printf("%s%s", path, values[i]);
            } else {
                tmp = g_strdup_printf("%s%s%s", path, dir_separator, values[i]);
            }
            g_free(path);
            path = tmp;
        } else {
            if (strlen(values[i]) == 0) {
                path = g_strdup("/");
            } else {
                path = g_strdup(values[i]);
            }
        }

        // build the tree
        if (i == 0) {
            // Find out all the available drives on the NNTPGrab server
            // and add these drives to the list
            NGList *list;

            if (local) {
                g_return_if_fail(nntpgrab_utils_get_folder_listing(NULL, &folders));
            } else {
                g_return_if_fail(nntpgrab_config_get_folder_listing(glue, NULL, &folders));
            }

            list = folders;
            while (list) {
                NNTPGrabFolder *folder = (NNTPGrabFolder*) list->data;

                gtk_tree_store_append(store, &iter, NULL);
                gtk_tree_store_set(store, &iter, 0, folder->folder, -1);

                if (!strcmp(folder->folder, path)) {
                    parent = iter;
                }

                if (folder->has_subfolders) {
                    GtkTreeIter iter_ignore;
                    gtk_tree_store_append(store, &iter_ignore, &iter);
                }

                list = ng_list_next(list);
            }

            if (local) {
                nntpgrab_utils_free_folder_listing(folders);
            } else {
                nntpgrab_config_free_folder_listing(glue, folders);
            }
        }

        if (local) {
            g_return_if_fail(nntpgrab_utils_get_folder_listing(path, &folders));
        } else {
            g_return_if_fail(nntpgrab_config_get_folder_listing(glue, path, &folders));
        }

        list = folders;
        temp = parent;

        // Is there already an child created?
        if (gtk_tree_model_iter_children(GTK_TREE_MODEL(store), &iter, &parent)) {
            already_has_child = TRUE;
            if (!strcmp(path, directory)) {
                gtk_tree_store_remove(store, &iter);
                already_has_child = FALSE;
            }
        } else {
            already_has_child = FALSE;
        }

        while (list) {
            NNTPGrabFolder *folder = (NNTPGrabFolder*) list->data;

            if (!already_has_child) {
                gtk_tree_store_append(store, &iter, &parent);
            } else {
                already_has_child = FALSE;
            }

            gtk_tree_store_set(store, &iter, 0, folder->folder, -1);

            if (folder->has_subfolders) {
                GtkTreeIter iter_ignore;
                gtk_tree_store_append(store, &iter_ignore, &iter);
            }

            if (values[i + 1] && !strcmp(folder->folder, values[i + 1])) {
                temp = iter;
            }

            list = ng_list_next(list);
        }

        parent = temp;

        if (local) {
            nntpgrab_utils_free_folder_listing(folders);
        } else {
            nntpgrab_config_free_folder_listing(glue, folders);
        }

        i++;
    }

    tree_path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &parent);
    gtk_tree_view_expand_to_path(tree, tree_path);
    gtk_tree_selection_select_iter(selection, &parent);
    gtk_tree_view_scroll_to_cell(tree, tree_path, NULL, TRUE, 0.5, 0.0);
    gtk_tree_path_free(tree_path);

    g_strfreev(values);
    if (path) {
        g_free(path);
    }
}

static void
on_selection_changed (GtkTreeSelection *selection, gpointer user_data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    char *path;
    ConfigOpts opts;
    ConfigGUIOpts gui_opts;

    if (!gtk_tree_selection_get_selected(selection, &model, &iter)) {
        return;
    }

    scroll_to_cell = FALSE;

    path = retrieve_path(model, iter);

    opts = nntpgrab_config_get_opts(glue);
    gui_opts = config_gui_get_opts();

    switch (GPOINTER_TO_INT(user_data)) {
        case TREE_LOCAL_DIRECTORY:
            if (!strcmp(gui_opts.auto_import_directory, path)) {
                // Nothing changed, ignore
                g_free(path);
                return;
            }

            memset(gui_opts.auto_import_directory, 0, sizeof(gui_opts.auto_import_directory));
            strncpy(gui_opts.auto_import_directory, path, sizeof(gui_opts.auto_import_directory) - 1);

            scroll_to_cell = FALSE;
            config_gui_set_opts(gui_opts);

            break;

        case TREE_SERVER_DIRECTORY:
            if (!strcmp(opts.auto_import_directory, path)) {
                // Nothing changed, ignore
                g_free(path);
                return;
            }

            memset(opts.auto_import_directory, 0, sizeof(opts.auto_import_directory));
            strncpy(opts.auto_import_directory, path, sizeof(opts.auto_import_directory) - 1);

            scroll_to_cell = FALSE;
            nntpgrab_config_set_opts(glue, opts);

            break;

        default:
            g_print(__FILE__ ":%i Invalid user_data: %i\n", __LINE__, GPOINTER_TO_INT(user_data));
            break;
    };

    g_free(path);
}

static gboolean
config_changed_phase2(gpointer data)
{
    GtkWidget *treeAutoImportLocal;
    GtkWidget *treeAutoImportServer;
    GtkTreeStore *storeLocal;
    GtkTreeStore *storeServer;
    GtkWidget *frameAutoImportLocal;
    GtkWidget *frameAutoImportServer;
    GtkWidget *chkAutoImportActivateLocal;
    GtkWidget *chkAutoImportActivateServer;
    GtkWidget *radAutoImportMove;
    GtkWidget *radAutoImportDel;
    GtkTreeSelection *selection;
    ConfigOpts opts;
    ConfigGUIOpts gui_opts;
    static ConfigOpts prev_opts;
    static ConfigGUIOpts prev_gui_opts;
    static gboolean beenhere = FALSE;
    GtkWidget *window = GTK_WIDGET(data);

    if (!beenhere) {
        beenhere = TRUE;
        memset(&prev_opts, 0, sizeof(ConfigOpts));
        memset(&prev_gui_opts, 0, sizeof(ConfigGUIOpts));
    }

    opts = nntpgrab_config_get_opts(glue);
    gui_opts = config_gui_get_opts();

    frameAutoImportLocal = nntpgrab_gui_base_get_widget("frameAutoImportLocal");
    frameAutoImportServer = nntpgrab_gui_base_get_widget("frameAutoImportServer");
    chkAutoImportActivateLocal = nntpgrab_gui_base_get_widget("chkAutoImportActivateLocal");
    chkAutoImportActivateServer = nntpgrab_gui_base_get_widget("chkAutoImportActivateServer");
    radAutoImportMove = nntpgrab_gui_base_get_widget("radAutoImportMove");
    radAutoImportDel = nntpgrab_gui_base_get_widget("radAutoImportDel");

    g_signal_handlers_block_by_func (G_OBJECT (chkAutoImportActivateLocal), G_CALLBACK (on_chkAutoImportActivateLocal_toggled), NULL);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chkAutoImportActivateLocal), gui_opts.enable_auto_import);
    g_signal_handlers_unblock_by_func (G_OBJECT (chkAutoImportActivateLocal), G_CALLBACK (on_chkAutoImportActivateLocal_toggled), NULL);
    g_signal_handlers_block_by_func (G_OBJECT (chkAutoImportActivateServer), G_CALLBACK (on_chkAutoImportActivateServer_toggled), NULL);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chkAutoImportActivateServer), opts.enable_auto_import);
    g_signal_handlers_unblock_by_func (G_OBJECT (chkAutoImportActivateServer), G_CALLBACK (on_chkAutoImportActivateServer_toggled), NULL);
    gtk_widget_set_sensitive(frameAutoImportLocal, gui_opts.enable_auto_import);

    if (!nntpgrab_glue_get_is_standalone(glue)) {
        gtk_widget_set_sensitive(frameAutoImportServer, opts.enable_auto_import);
    }

    if (gui_opts.enable_auto_import != prev_gui_opts.enable_auto_import) {
        if (gui_opts.enable_auto_import) {
            enable_auto_import();
        } else {
            disable_auto_import();
        }
    }

    g_signal_handlers_block_by_func (G_OBJECT(radAutoImportMove), G_CALLBACK(on_radAutoImportMove_toggled), NULL);
    if (opts.move_file_after_auto_import) {
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radAutoImportMove), TRUE);
    } else {
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radAutoImportDel), TRUE);
    }
    g_signal_handlers_unblock_by_func (G_OBJECT(radAutoImportMove), G_CALLBACK(on_radAutoImportMove_toggled), NULL);

    // Did we cause the signal to be emitted ?
    if (!scroll_to_cell) {
        scroll_to_cell = TRUE;
        prev_opts = opts;
        prev_gui_opts = gui_opts;

        return FALSE;
    }

    if (strcmp(gui_opts.auto_import_directory, prev_gui_opts.auto_import_directory)) {
        treeAutoImportLocal = nntpgrab_gui_base_get_widget("treeAutoImportLocal");

        storeLocal = GTK_TREE_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(treeAutoImportLocal)));
        selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeAutoImportLocal));
        g_signal_handlers_block_by_func (G_OBJECT (treeAutoImportLocal), G_CALLBACK (on_row_expanded), GINT_TO_POINTER(TREE_LOCAL_DIRECTORY));
        g_signal_handlers_block_by_func (G_OBJECT (selection), G_CALLBACK (on_selection_changed), GINT_TO_POINTER(TREE_LOCAL_DIRECTORY));
        update_folder_listing(GTK_TREE_VIEW(treeAutoImportLocal), storeLocal, selection, gui_opts.auto_import_directory, TRUE);
        g_signal_handlers_unblock_by_func (G_OBJECT (selection), G_CALLBACK (on_selection_changed), GINT_TO_POINTER(TREE_LOCAL_DIRECTORY));
        g_signal_handlers_unblock_by_func (G_OBJECT (treeAutoImportLocal), G_CALLBACK (on_row_expanded), GINT_TO_POINTER(TREE_LOCAL_DIRECTORY));
    }

    if (strcmp(opts.auto_import_directory, prev_opts.auto_import_directory)) {
        treeAutoImportServer = nntpgrab_gui_base_get_widget("treeAutoImportServer");

        storeServer = GTK_TREE_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(treeAutoImportServer)));
        selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeAutoImportServer));
        g_signal_handlers_block_by_func (G_OBJECT (treeAutoImportServer), G_CALLBACK (on_row_expanded), GINT_TO_POINTER(TREE_SERVER_DIRECTORY));
        g_signal_handlers_block_by_func (G_OBJECT (selection), G_CALLBACK (on_selection_changed), GINT_TO_POINTER(TREE_SERVER_DIRECTORY));
        update_folder_listing(GTK_TREE_VIEW(treeAutoImportServer), storeServer, selection, opts.auto_import_directory, FALSE);
        g_signal_handlers_unblock_by_func (G_OBJECT (selection), G_CALLBACK (on_selection_changed), GINT_TO_POINTER(TREE_SERVER_DIRECTORY));
        g_signal_handlers_unblock_by_func (G_OBJECT (treeAutoImportServer), G_CALLBACK (on_row_expanded), GINT_TO_POINTER(TREE_SERVER_DIRECTORY));
    }

    gtk_widget_hide(window);
    gtk_widget_destroy(window);

    prev_opts = opts;
    prev_gui_opts = gui_opts;

    return FALSE;
}

static void
config_changed(NNTPGrabGlue *glue, gpointer data)
{
    GtkWidget *window = NULL;
    GtkWidget *label;
    char *markup;

    if (!initialized) {
        return;
    }

    if (scroll_to_cell) {
        window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
        gtk_window_set_skip_taskbar_hint(GTK_WINDOW(window), TRUE);
        gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
        label = gtk_label_new(NULL);
        markup = g_markup_printf_escaped("<span size=\"large\" weight=\"bold\">%s</span>", _("The list of available automatic import folders is now being updated.\n\nPlease wait a few seconds"));
        gtk_label_set_markup(GTK_LABEL(label), markup);
        g_free(markup);
        gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
        gtk_container_add(GTK_CONTAINER(window), label);
        gtk_widget_show(label);
        gtk_widget_show(window);
    }

    g_idle_add(config_changed_phase2, window);
}
#endif

void
config_autoimport_initialize(void)
{
#if GLIB_CHECK_VERSION(2,16,0)
    GtkWidget *treeAutoImportLocal;
    GtkWidget *treeAutoImportServer;
    GtkTreeStore *storeLocal;
    GtkTreeStore *storeServer;
    GtkCellRenderer *cellRenderer;
    GtkTreeSelection *selection;
    GtkWidget *chkAutoImportActivateServer = nntpgrab_gui_base_get_widget("chkAutoImportActivateServer");

    treeAutoImportLocal = nntpgrab_gui_base_get_widget("treeAutoImportLocal");
    treeAutoImportServer = nntpgrab_gui_base_get_widget("treeAutoImportServer");

    // Local
    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeAutoImportLocal), FALSE);
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeAutoImportLocal));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);

    storeLocal = gtk_tree_store_new(1, G_TYPE_STRING);
    gtk_tree_view_set_model(GTK_TREE_VIEW(treeAutoImportLocal), GTK_TREE_MODEL(storeLocal));

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeAutoImportLocal), -1, _("Directory"), cellRenderer, "text", 0, NULL);

    g_signal_connect(G_OBJECT(treeAutoImportLocal), "row-expanded", G_CALLBACK(on_row_expanded), GINT_TO_POINTER(TREE_LOCAL_DIRECTORY));
    g_signal_connect(G_OBJECT(selection), "changed", G_CALLBACK(on_selection_changed), GINT_TO_POINTER(TREE_LOCAL_DIRECTORY));

    // Server
    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeAutoImportServer), FALSE);
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeAutoImportServer));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);

    storeServer = gtk_tree_store_new(1, G_TYPE_STRING);
    gtk_tree_view_set_model(GTK_TREE_VIEW(treeAutoImportServer), GTK_TREE_MODEL(storeServer));

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeAutoImportServer), -1, _("Directory"), cellRenderer, "text", 0, NULL);

    g_signal_connect(G_OBJECT(treeAutoImportServer), "row-expanded", G_CALLBACK(on_row_expanded), GINT_TO_POINTER(TREE_SERVER_DIRECTORY));
    g_signal_connect(G_OBJECT(selection), "changed", G_CALLBACK(on_selection_changed), GINT_TO_POINTER(TREE_SERVER_DIRECTORY));

    if (!nntpgrab_glue_get_is_standalone(glue)) {
        gtk_widget_set_sensitive(chkAutoImportActivateServer, TRUE);
    }

    g_signal_connect(config_gui_get_object(), "config_changed", G_CALLBACK(config_changed), GINT_TO_POINTER(TREE_LOCAL_DIRECTORY));
    ng_signal_connect(glue, "config_changed", NG_CALLBACK(config_changed), GINT_TO_POINTER(TREE_SERVER_DIRECTORY));
#endif
}

void
config_autoimport_update_folders(void)
{
#if GLIB_CHECK_VERSION(2,16,0)
    if (initialized) {
        return;
    }
    initialized = TRUE;

    scroll_to_cell = TRUE;
    config_changed(glue, NULL);
#endif
}
