/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <wspiapi.h>
#else
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif
#include <glib.h>
#include <glib/gi18n.h>
#include <glib-object.h>
#include <gmodule.h>
#include "marshalers.h"
#include "nntpgrab_glue.h"
#include "nntpgrab.h"
#include "nntpgrab_internal.h"
#include "nntpgrab_utils.h"

#ifdef _MSC_VER
#include "config.h.win32"
#else
#include "config.h"
#endif

#define NNTPGRAB_TYPE_GLUE              (glue_get_type ())
#define NNTPGRAB_GLUE(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), NNTPGRAB_TYPE_GLUE, NNTPGrabGlue))
#define NNTPGRAB_GLUE_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), NNTPGRAB_TYPE_GLUE, NNTPGrabGlueClass))
#define IS_NNTPGRAB_GLUE(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), NNTPGRAB_TYPE_GLUE))
#define IS_NNTPGRAB_GLUE_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), NNTPGRAB_TYPE_GLUE))
#define NNTPGRAB_GLUE_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), NNTPGRAB_TYPE_GLUE, NNTPGrabGlueClass))

#ifdef WIN32
#define CLOSE(x)    closesocket(x)
#else
#define CLOSE(x)    close(x)
#endif

#ifdef _MSC_VER
#undef atoll
#define atoll _atoi64

#undef gai_strerror
#define gai_strerror gai_strerrorA
#endif

#define DEBUG(x)                            \
    g_print(x);                             \
    nntpgrab_core_emit_debug_message(FALSE, x);

#define READ_TIMEOUT_VALUE      30

typedef struct _connection {
    int socket_id;
    char recv_buf[4096];      // It should be safe to assume that one line is less than 4096 characters
    int recv_buf_len;
} Connection;

struct NNTPGrabGlue
{
    GObject parent;

    Connection socket_listener;
    Connection socket_command;

    int glue_version;

    NNTPGrabCore *standalone_core;
};

static NNTPGrabGlue *glue = NULL;

struct NNTPGrabGlueClass
{
    GObjectClass parent;
};

enum
{
    CONFIG_CHANGED_SIGNAL,
    PART_DOWNLOAD_START_SIGNAL,
    PART_DONE_SIGNAL,
    PART_FAILED_SIGNAL,
    PART_PROGRESS_UPDATE_SIGNAL,
    TRAFFIC_MONITOR_UPDATE_SIGNAL,
    COLLECTION_ADDED_SIGNAL,
    COLLECTION_REMOVED_SIGNAL,
    COLLECTION_MODIFIED_SIGNAL,
    FILE_ADDED_SIGNAL,
    FILE_REMOVED_SIGNAL,
    FILE_DOWNLOAD_STATE_UPDATE_SIGNAL,
    FILE_STATE_CHANGED_SIGNAL,
    CONNECTION_CONNECTING_SIGNAL,
    CONNECTION_CONNECTED_SIGNAL,
    CONNECTION_DISCONNECT_SIGNAL,
    SCHEDULAR_STATE_CHANGED_SIGNAL,
    FATAL_ERROR_SIGNAL,
    WARNING_MESSAGE_SIGNAL,
    DEBUG_MESSAGE_SIGNAL,
    CONNECTION_LOST_SIGNAL,
    TASK_MOVED_SIGNAL,
    COLLECTION_MOVED_SIGNAL,
    PAR2_REPAIR_MESSAGE_SIGNAL,
    ALL_DOWNLOADS_COMPLETED_SIGNAL,
    UNPACK_PROGRESS_UPDATE_SIGNAL,
    UNPACK_MESSAGE_RECEIVED_SIGNAL,
    UNPACK_WORKING_ARCHIVE_CHANGED_SIGNAL,
    PAR2_BEGIN_VERIFY_SIGNAL,
    PAR2_LOAD_PROGRESS_UPDATE_SIGNAL,
    PAR2_RECOVERY_FILE_LOADED_SIGNAL,
    PAR2_FILE_LOADED_SIGNAL,
    PAR2_REPAIR_PROGRESS_UPDATE_SIGNAL,
    PAR2_REPAIR_FAILURE_SIGNAL,
    PAR2_REPAIR_SUCCESS_SIGNAL,
    PAR2_NO_REPAIR_REQUIRED_SIGNAL,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };
static GModule *module = NULL;

G_DEFINE_TYPE(NNTPGrabGlue, glue, G_TYPE_OBJECT)

void     connect_signal_handlers(NNTPGrabCore *core);
void     disconnect_signal_handlers(NNTPGrabCore *core);

static void
glue_init (NNTPGrabGlue *obj)
{
    obj->socket_listener.socket_id = -1;
    obj->socket_command.socket_id = -1;
    obj->glue_version = -1;
    obj->standalone_core = NULL;
    memset(&obj->socket_listener.recv_buf, 0, sizeof(obj->socket_listener.recv_buf));
    memset(&obj->socket_command.recv_buf, 0, sizeof(obj->socket_command.recv_buf));
}

static void
glue_finalize (GObject *obj)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);

    if (glue->standalone_core) {
        disconnect_signal_handlers(glue->standalone_core);
        g_object_unref(glue->standalone_core);
        glue->standalone_core = NULL;
    }

    if (glue->socket_listener.socket_id > 0) {
        CLOSE(glue->socket_listener.socket_id);
        glue->socket_listener.socket_id = -1;
    }

    if (glue->socket_command.socket_id > 0) {
        CLOSE(glue->socket_command.socket_id);
        glue->socket_command.socket_id = -1;
    }
}

static void
glue_class_init (NNTPGrabGlueClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = glue_finalize;

    signals[CONFIG_CHANGED_SIGNAL] =       g_signal_new("config_changed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__VOID,
                                                        G_TYPE_NONE, 0);

    signals[PART_DOWNLOAD_START_SIGNAL] = g_signal_new ("part_download_start",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT,
                                                        G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);

    signals[PART_DONE_SIGNAL] =           g_signal_new ("part_done",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT_INT,
                                                        G_TYPE_NONE, 6, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    signals[PART_FAILED_SIGNAL] =         g_signal_new ("part_failed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT_INT_BOOLEAN,
                                                        G_TYPE_NONE, 7, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_BOOLEAN);

    signals[PART_PROGRESS_UPDATE_SIGNAL] = g_signal_new ("part_progress_update",
                                                         G_OBJECT_CLASS_TYPE (klass),
                                                         G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                         0,
                                                         NULL, NULL,
                                                         nntpgrab_marshal_VOID__STRING_INT_STRING_INT_INT_INT,
                                                         G_TYPE_NONE, 6, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT);

    signals[TRAFFIC_MONITOR_UPDATE_SIGNAL] = g_signal_new ("traffic_monitor_update",
                                                         G_OBJECT_CLASS_TYPE (klass),
                                                         G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                         0,
                                                         NULL, NULL,
                                                         nntpgrab_marshal_VOID__INT_INT_INT_INT_INT_INT_INT_INT_INT_INT_INT64_DOUBLE,
                                                         G_TYPE_NONE, 12, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT64, G_TYPE_DOUBLE);

    signals[COLLECTION_ADDED_SIGNAL] =    g_signal_new ("collection_added",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING,
                                                        G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);

    signals[COLLECTION_REMOVED_SIGNAL] =  g_signal_new ("collection_removed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING,
                                                        G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[COLLECTION_MODIFIED_SIGNAL] =  g_signal_new ("collection_modified",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING,
                                                        G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);

    signals[FILE_ADDED_SIGNAL] =          g_signal_new ("file_added",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING_STRING_UINT64_UINT64_UINT64_UINT64_INT_INT_POINTER,
                                                        G_TYPE_NONE, 10, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_INT, G_TYPE_INT, G_TYPE_POINTER);

    signals[FILE_REMOVED_SIGNAL] =        g_signal_new ("file_removed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING_UINT64_UINT64,
                                                        G_TYPE_NONE, 4, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT64, G_TYPE_UINT64);

    signals[FILE_DOWNLOAD_STATE_UPDATE_SIGNAL] =  g_signal_new ("file_download_state_update",
                                                                G_OBJECT_CLASS_TYPE (klass),
                                                                G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                                0,
                                                                NULL, NULL,
                                                                nntpgrab_marshal_VOID__STRING_STRING_INT_INT_INT_UINT64_UINT64_UINT64_UINT64,
                                                                G_TYPE_NONE, 9, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64);

    signals[FILE_STATE_CHANGED_SIGNAL] =  g_signal_new ("file_state_changed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING_STRING_INT_INT,
                                                        G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    signals[CONNECTION_CONNECTING_SIGNAL] =  g_signal_new   ("connection_connecting",
                                                             G_OBJECT_CLASS_TYPE (klass),
                                                             G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                             0,
                                                             NULL, NULL,
                                                             nntpgrab_marshal_VOID__STRING_INT,
                                                             G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_INT);

    signals[CONNECTION_CONNECTED_SIGNAL] =  g_signal_new    ("connection_connected",
                                                             G_OBJECT_CLASS_TYPE (klass),
                                                             G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                             0,
                                                             NULL, NULL,
                                                             nntpgrab_marshal_VOID__STRING_INT_STRING,
                                                             G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING);

    signals[CONNECTION_DISCONNECT_SIGNAL] = g_signal_new ("connection_disconnect",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_INT_ENUM_STRING,
                                                          G_TYPE_NONE, 4, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_STRING);


    signals[SCHEDULAR_STATE_CHANGED_SIGNAL] = g_signal_new ("schedular_state_changed",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__INT_STRING,
                                                          G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_STRING);


    signals[FATAL_ERROR_SIGNAL] =           g_signal_new ("fatal_error",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING,
                                                          G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[WARNING_MESSAGE_SIGNAL] =       g_signal_new ("warning_message",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING,
                                                          G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[DEBUG_MESSAGE_SIGNAL] =         g_signal_new ("debug_message",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__POINTER_STRING,
                                                          G_TYPE_NONE, 2, G_TYPE_POINTER, G_TYPE_STRING);

    signals[CONNECTION_LOST_SIGNAL] =       g_signal_new ("connection_lost",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING,
                                                          G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[TASK_MOVED_SIGNAL] =            g_signal_new ("task_moved",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING_STRING_INT_INT,
                                                          G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    signals[COLLECTION_MOVED_SIGNAL] =      g_signal_new ("collection_moved",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_INT_INT,
                                                          G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    signals[PAR2_REPAIR_MESSAGE_SIGNAL] =   g_signal_new ("par2_repair_message",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING,
                                                          G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[ALL_DOWNLOADS_COMPLETED_SIGNAL] =  g_signal_new("all_downloads_completed",
                                                            G_OBJECT_CLASS_TYPE (klass),
                                                            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                            0,
                                                            NULL, NULL,
                                                            nntpgrab_marshal_VOID__VOID,
                                                            G_TYPE_NONE, 0);

    signals[UNPACK_PROGRESS_UPDATE_SIGNAL] = g_signal_new("unpack_progress_update",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING_DOUBLE,
                                                          G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_DOUBLE);

    signals[UNPACK_MESSAGE_RECEIVED_SIGNAL] = g_signal_new("unpack_message_received",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING_STRING,
                                                          G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);

    signals[UNPACK_WORKING_ARCHIVE_CHANGED_SIGNAL] = g_signal_new("unpack_working_archive_changed",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING_STRING,
                                                          G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);

    signals[PAR2_BEGIN_VERIFY_SIGNAL] =      g_signal_new("par2_begin_verify",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING,
                                                          G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);

    signals[PAR2_LOAD_PROGRESS_UPDATE_SIGNAL] = g_signal_new("par2_load_progress_update",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING_STRING_DOUBLE,
                                                          G_TYPE_NONE, 4, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_DOUBLE);

    signals[PAR2_RECOVERY_FILE_LOADED_SIGNAL] = g_signal_new("par2_recovery_file_loaded",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING_STRING_INT_INT,
                                                          G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    signals[PAR2_FILE_LOADED_SIGNAL] =       g_signal_new("par2_file_loaded",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING_STRING_INT_INT_INT,
                                                          G_TYPE_NONE, 6, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT);

    signals[PAR2_REPAIR_PROGRESS_UPDATE_SIGNAL] = g_signal_new("par2_repair_progress_update",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING_DOUBLE,
                                                          G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_DOUBLE);

    signals[PAR2_REPAIR_FAILURE_SIGNAL] =    g_signal_new("par2_repair_failure",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING_INT,
                                                          G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);

    signals[PAR2_REPAIR_SUCCESS_SIGNAL] =    g_signal_new("par2_repair_success",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING,
                                                          G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);

    signals[PAR2_NO_REPAIR_REQUIRED_SIGNAL] = g_signal_new("par2_no_repair_required",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING,
                                                          G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);
}

static TaskState
string_to_task_state(const char *str)
{
    TaskState state;

    g_return_val_if_fail(str != NULL, TASK_STATE_FINISHED_COMPLETE);

    if (!strncmp(str, "WAITING_FOR_DOWNLOAD", 20)) {
        state = TASK_STATE_WAITING_FOR_DOWNLOAD;
    } else if (!strncmp(str, "DOWNLOADING", 11)) {
        state = TASK_STATE_DOWNLOADING;
    } else if (!strncmp(str, "WAITING_FOR_DECODE", 18)) {
        state = TASK_STATE_WAITING_FOR_DECODE;
    } else if (!strncmp(str, "DECODING", 8)) {
        state = TASK_STATE_DECODING;
    } else if (!strncmp(str, "FINISHED_COMPLETE", 17)) {
        state = TASK_STATE_FINISHED_COMPLETE;
    } else if (!strncmp(str, "FINISHED_INCOMPLETE", 19)) {
        state = TASK_STATE_FINISHED_INCOMPLETE;
    } else if (!strncmp(str, "FINISHED_NO_PARTS_AVAIL", 23)) {
        state = TASK_STATE_FINISHED_NO_PARTS_AVAIL;
    } else if (!strncmp(str, "SKIPPED", 7)) {
        state = TASK_STATE_SKIPPED;
    } else {
        char *msg;

        state = TASK_STATE_FINISHED_COMPLETE;

        msg = g_strdup_printf(__FILE__ ":%i Unknown task state: %s\n", __LINE__, str);
        DEBUG(msg);
        g_free(msg);
    }

    return state;
}

static void
strip_newline(char *line)
{
    if (strlen(line) <= 1) {
        return;
    }

    if (line[strlen(line) - 1]  == '\n') {
        line[strlen(line) - 1] = '\0';
    }

    if (line[strlen(line) - 1]  == '\r') {
        line[strlen(line) - 1] = '\0';
    }

}

static gboolean
read_line(Connection *sock, char **ret_line, int timeout, gboolean *error_flag)
{
    int i;
    int len;
    char *line;
    time_t stamp;

    g_assert(ret_line);

    if (error_flag) {
        *error_flag = FALSE;
    }

    g_return_val_if_fail(sock > 0, FALSE);

    memset(sock->recv_buf + sock->recv_buf_len, 0, sizeof(sock->recv_buf) - sock->recv_buf_len);

    stamp = time(NULL);

    do {
        struct timeval tv;
        fd_set recv_fds;

        // Check is there is a newline in the buffer
        for (i = 0; i < sock->recv_buf_len; i++) {
            if (sock->recv_buf[i] == '\n') {
                // Newline found, split the buffer into a line and process it
                line = g_strndup(sock->recv_buf, i);

                strip_newline(line);

                memmove(sock->recv_buf, sock->recv_buf + i + 1, sock->recv_buf_len - i - 1);
                sock->recv_buf_len -= i + 1;

                *ret_line = line;

                return TRUE;
            }
        }

        // No newline found, try to read new data from the socket
        do {
            FD_ZERO(&recv_fds);
            FD_SET(sock->socket_id, &recv_fds);

            tv.tv_sec = timeout;
            tv.tv_usec = 0;

            len = select(sock->socket_id + 1, &recv_fds, NULL, NULL, &tv);
            if (len == 0) {
                // Read timeout
                *ret_line = g_strdup("Read timeout");
                return FALSE;
            } else if (len == 1) {
                gboolean beenhere = FALSE;

                do {
                    if (beenhere) {
                        g_usleep(G_USEC_PER_SEC / 10);
                    } else {
                        beenhere = TRUE;
                    }

                    len = recv(sock->socket_id, sock->recv_buf + sock->recv_buf_len, sizeof(sock->recv_buf) - sock->recv_buf_len, 0);
                } while (len == -1 && errno == EAGAIN && stamp + timeout > time(NULL));
            }
        } while (len == -1 && errno == EINTR && stamp + timeout > time(NULL));

        if (len == -1) {
            if (errno == EAGAIN) {
                // Read timeout
                *ret_line = g_strdup("Read timeout");
                return FALSE;
            }

            // Error occured
g_print(__FILE__":%i Error occured: %s\n", __LINE__, strerror(errno));
            if (error_flag) {
                *error_flag = TRUE;
            }

            CLOSE(sock->socket_id);
            sock->socket_id = -1;

            *ret_line = g_strdup_printf("Unexpected error occured while performing read: %s", strerror(errno));

            return FALSE;
        }

        if (len == 0) {
            // EOF, disconnect
g_print(__FILE__":%i EOF detected\n", __LINE__);
            if (error_flag) {
                *error_flag = TRUE;
            }

            CLOSE(sock->socket_id);
            sock->socket_id = -1;

            *ret_line = g_strdup(_("NNTPGrab server has disconnected"));

            return FALSE;
        }

        sock->recv_buf_len += len;
    } while (TRUE);

    g_assert_not_reached();

    return FALSE;
}

static gboolean
write_line(Connection *sock, const char *line)
{
    if (send(sock->socket_id, line, strlen(line), 0) == -1) {
//        imported_funcs.notify_disconnect(DISCONNECT_WRITE_ERROR, strerror(errno), data);

        CLOSE(sock->socket_id);
        sock->socket_id = -1;

        return FALSE;
    }

    return TRUE;
}

static void
connect_to_server(Connection *sock, const char *hostname, int port, char **welcome_msg)
{
    struct addrinfo hints, *res, *ressave;
    int n;
    char *str_port;
    int last_errno;
    struct timeval tv;
#if 0
    SSL *ssl;
#endif

    g_assert(sock);
    g_assert(hostname);
    g_assert(port > 0);
    g_assert(port < 65535);

    if (sock->socket_id >= 0) {
        CLOSE(sock->socket_id);
    }

    sock->socket_id = -1;

    memset(&hints, 0, sizeof(struct addrinfo));

    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    str_port = g_strdup_printf("%i", port);
    n = getaddrinfo(hostname, str_port, &hints, &res);
    g_free(str_port);

    if (n != 0) {
        if (welcome_msg) {
            *welcome_msg = g_strdup(gai_strerror(n));
        }

        return;
    }

    ressave = res;
    last_errno = 0;

    while (res) {
        sock->socket_id = socket(res->ai_family, res->ai_socktype, res->ai_protocol);

#ifdef WIN32
        if ((sock->socket_id == INVALID_SOCKET)) {
#else
        if ((sock->socket_id == -1)) {
#endif
            // The socket couldn't be created
            // Save the errno and try the next item in the list
            last_errno = errno;
            res = res->ai_next;
            sock->socket_id = -1;

            if (welcome_msg) {
                if (*welcome_msg) {
                    g_free(*welcome_msg);
                }

                *welcome_msg = g_strdup(strerror(last_errno));
            }

            continue;
        }

        // Set the connection timeout on the socket
        tv.tv_sec = 5;
        tv.tv_usec = 0;
        setsockopt(sock->socket_id, SOL_SOCKET, SO_RCVTIMEO, (const void*) &tv, sizeof(tv));
        setsockopt(sock->socket_id, SOL_SOCKET, SO_SNDTIMEO, (const void*) &tv, sizeof(tv));

        if (connect(sock->socket_id, res->ai_addr, (int) res->ai_addrlen) == 0) {
            // Connection succesfull
            last_errno = 0;
            break;
        }

        // Connection could not be made, save the errno
#ifdef WIN32
        last_errno = WSAGetLastError();
#else
        if (errno == EINPROGRESS) {
            // The connection could not be established within the connect time limit.
            // This provides a more clear error message
            last_errno = ETIMEDOUT;
        } else {
            last_errno = errno;
        }
#endif

        if (welcome_msg) {
            if (*welcome_msg) {
                g_free(*welcome_msg);
            }

#ifdef WIN32
            switch(WSAGetLastError()) {
                case WSAECONNREFUSED:
                    *welcome_msg = g_strdup(_("Connection refused"));
                    break;

                case WSAETIMEDOUT:
                    *welcome_msg = g_strdup(_("Connection timed out"));
                    break;

                default:
                    *welcome_msg = g_strdup_printf(_("Unknown error occured, WSAGetLastError() = %i"), WSAGetLastError());
                    break;
            }
#else
            *welcome_msg = g_strdup(strerror(last_errno));
#endif
        }

        CLOSE(sock->socket_id);
        sock->socket_id = -1;

        res = res->ai_next;
    }

    if (sock->socket_id == -1) {
        freeaddrinfo(ressave);

        return;
    }

    freeaddrinfo(ressave);

#if 0
    if (use_ssl) {
        if ((ssl = prepare_ssl_connection(*conn_id, errmsg)) == NULL) {
            CLOSE(*conn_id);
            return NNTP_ERROR_SSL_INITIALISE;
        }
    } else {
        ssl = NULL;
    }
#endif

    if (!read_line(sock, welcome_msg, READ_TIMEOUT_VALUE, NULL)) {
        CLOSE(sock->socket_id);
        sock->socket_id = -1;

        return;
    }
}

/*********************************
 Callback functions
 *********************************/
static gboolean abort_flag = FALSE;
static GThread *thread_listener = NULL;

void emit_config_changed(NNTPGrabCore *core);

NNTPGrabCore *
get_core(void)
{
    return (NNTPGrabCore*) glue;
}

static char **
parse_line(const char *line, const char *expected_command, int num_args_expected)
{
    char **args;
    char *ptr_args_start;
    int num;

    g_return_val_if_fail(line != NULL, NULL);
    g_return_val_if_fail(expected_command != NULL, NULL);
    g_return_val_if_fail(num_args_expected >= 0, NULL);

    ptr_args_start = strstr(line, " ");
    if (!ptr_args_start) {
        char *msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n"__FILE__ ":%i Expected command = %s\n", __LINE__, line, __LINE__, expected_command);
        DEBUG(msg);
        g_free(msg);
        return NULL;
    }

    args = g_strsplit(ptr_args_start + 1, "\t", -1);
    g_return_val_if_fail(args != NULL, NULL);

    // Verify the expected command
    if (!strncmp(line, expected_command, (ptr_args_start - line + 1))) {
        char *msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n"__FILE__ ":%i Expected command = %s\n", __LINE__, line, __LINE__, expected_command);
        DEBUG(msg);
        g_free(msg);
        g_strfreev(args);
        return NULL;
    }

    // Verify the argument count
    num = 0;
    while (args[num]) {
        num++;
    }

    if (num != num_args_expected) {
        char *msg = g_strdup_printf(__FILE__ ":%i Invalid number of arguments received: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_strfreev(args);
        return NULL;
    }

    return args;
}

static void
emit_schedular_state_changed(const char *command)
{
    SchedularState state;
    char **values;

    values = parse_line(command, "SCHEDULAR_STATE_CHANGED", 2);
    if (!values) {
        return;
    }

    if (!strncmp(values[0], "RUNNING", 7)) {
        state = SCHEDULAR_STATE_RUNNING;
    } else if (!strncmp(values[0], "STOPPING", 8)) {
        state = SCHEDULAR_STATE_STOPPING;
    } else if (!strncmp(values[0], "STOPPED", 7)) {
        state = SCHEDULAR_STATE_STOPPED;
    } else {
        char *msg;

        state = SCHEDULAR_STATE_STOPPED;
        msg = g_strdup_printf(__FILE__ ":%i Unknown schedular state: %s\n", __LINE__, values[0]);
        DEBUG(msg);
        g_free(msg);
    }

    nntpgrab_core_emit_schedular_state_changed(FALSE, state, (values[1] && strlen(values[1]) > 0) ? values[1] : NULL);

    g_strfreev(values);
}

static void
emit_file_download_state_update(const char *command)
{
    char **values;

    values = parse_line(command, "FILE_DOWNLOAD_STATE_UPDATE", 9);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_file_download_state_update(FALSE, values[0], values[1], atoi(values[2]), atoi(values[3]), atoi(values[4]), atoll(values[5]), atoll(values[6]), atoll(values[7]), atoll(values[8]));

    g_strfreev(values);
}

static void
emit_file_state_changed(const char *command)
{
    TaskState old_state, new_state;
    char **values;
    const char *real_filename;

    values = parse_line(command, "FILE_STATE_CHANGED", 5);
    if (!values) {
        return;
    }

    if (strlen(values[2]) > 0) {
        real_filename = values[2];
    } else {
        real_filename = NULL;
    }
    old_state = string_to_task_state(values[3]);
    new_state = string_to_task_state(values[4]);

    nntpgrab_core_emit_file_state_changed(FALSE, values[0], values[1], real_filename, old_state, new_state);

    g_strfreev(values);
}

static void
emit_traffic_monitor_update(const char *command)
{
    char **values;
    int bytes_received[10];

    values = parse_line(command, "TRAFFIC_MONITOR_UPDATE", 12);
    if (!values) {
        return;
    }

    bytes_received[0] = atoi(values[0]);
    bytes_received[1] = atoi(values[1]);
    bytes_received[2] = atoi(values[2]);
    bytes_received[3] = atoi(values[3]);
    bytes_received[4] = atoi(values[4]);
    bytes_received[5] = atoi(values[5]);
    bytes_received[6] = atoi(values[6]);
    bytes_received[7] = atoi(values[7]);
    bytes_received[8] = atoi(values[8]);
    bytes_received[9] = atoi(values[9]);

    nntpgrab_core_emit_traffic_monitor_update(FALSE, bytes_received, atoi(values[10]));

    g_strfreev(values);
}

static void
emit_connection_connecting(const char *command)
{
    char **values;

    values = parse_line(command, "CONNECTION_CONNECTING", 2);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_connecting(FALSE, values[0], atoi(values[1]));

    g_strfreev(values);
}

static void
emit_connection_connected(const char *command)
{
    char **values;

    values = parse_line(command, "CONNECTION_CONNECTED", 3);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_connected(FALSE, values[0], atoi(values[1]), values[2]);

    g_strfreev(values);
}

static void
emit_connection_disconnect(const char *command)
{
    char **values;
    NNTPDisconnectType disconnect_type;

    values = parse_line(command, "CONNECTION_DISCONNECT", 4);
    if (!values) {
        return;
    }

    if (!strcmp(values[2], "NORMAL")) {
        disconnect_type = DISCONNECT_NORMAL;
    } else if (!strcmp(values[2], "NO_SUCH_HOST")) {
        disconnect_type = DISCONNECT_NO_SUCH_HOST;
    } else if (!strcmp(values[2], "CONNECTION_REFUSED")) {
        disconnect_type = DISCONNECT_CONNECTION_REFUSED;
    } else if (!strcmp(values[2], "TOO_MANY_CONNECTIONS")) {
        disconnect_type = DISCONNECT_TOO_MANY_CONNECTIONS;
    } else if (!strcmp(values[2], "CONNECT_TIMEOUT")) {
        disconnect_type = DISCONNECT_CONNECT_TIMEOUT;
    } else if (!strcmp(values[2], "READ_ERROR")) {
        disconnect_type = DISCONNECT_READ_ERROR;
    } else if (!strcmp(values[2], "READ_TIMEOUT")) {
        disconnect_type = DISCONNECT_READ_TIMEOUT;
    } else if (!strcmp(values[2], "WRITE_ERROR")) {
        disconnect_type = DISCONNECT_WRITE_ERROR;
    } else if (!strcmp(values[2], "IDLE_TIMEOUT")) {
        disconnect_type = DISCONNECT_IDLE_TIMEOUT;
    } else if (!strcmp(values[2], "INVALID_MSG")) {
        disconnect_type = DISCONNECT_INVALID_MSG;
    } else if (!strcmp(values[2], "LOGIN_FAILURE")) {
        disconnect_type = DISCONNECT_LOGIN_FAILURE;
    } else if (!strcmp(values[2], "ERROR_SSL_INITIALISE")) {
        disconnect_type = DISCONNECT_ERROR_SSL_INITIALISE;
    } else if (!strcmp(values[2], "UNEXPECTED")) {
        disconnect_type = DISCONNECT_UNEXPECTED;
    } else {
        char *msg;

        disconnect_type = DISCONNECT_UNEXPECTED;

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);
    }

    nntpgrab_core_emit_disconnect(FALSE, values[0], atoi(values[1]), disconnect_type, values[3]);

    g_strfreev(values);
}

static void
emit_part_download_start(const char *command)
{
    char **values;

    values = parse_line(command, "PART_DOWNLOAD_START", 5);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_part_download_start(FALSE, values[0], atoi(values[1]), values[2], values[3], atoi(values[4]));

    g_strfreev(values);
}

static void
emit_part_done(const char *command)
{
    char **values;

    values = parse_line(command, "PART_DONE", 6);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_part_done(FALSE, values[0], atoi(values[1]), values[2], values[3], atoi(values[4]), atoi(values[5]));

    g_strfreev(values);
}

static void
emit_part_failed(const char *command)
{
    char **values;
    gboolean all_servers_tried;

    values = parse_line(command, "PART_FAILED", 7);
    if (!values) {
        return;
    }

    if (!strcmp(values[6], "TRUE")) {
        all_servers_tried = TRUE;
    } else {
        all_servers_tried = FALSE;
    }

    nntpgrab_core_emit_part_failed(FALSE, values[0], atoi(values[1]), values[2], values[3], atoi(values[4]), atoi(values[5]), all_servers_tried);

    g_strfreev(values);
}

static void
emit_part_progress_update(const char *command)
{
    char **values;

    values = parse_line(command, "PART_PROGRESS_UPDATE", 6);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_part_progress_update(FALSE, values[0], atoi(values[1]), values[2], atoi(values[3]), atoi(values[4]), atoi(values[5]));

    g_strfreev(values);
}

static void
emit_collection_added(const char *command)
{
    char **values;

    values = parse_line(command, "COLLECTION_ADDED", 2);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_collection_added(FALSE, values[0], values[1]);

    g_strfreev(values);
}

static void
emit_collection_modified(const char *command)
{
    char **values;

    values = parse_line(command, "COLLECTION_MODIFIED", 2);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_collection_modified(FALSE, values[0], values[1]);

    g_strfreev(values);
}

static void
emit_file_added(const char *command)
{
    char **values;
    char **values_groups;
    GList *groups = NULL;
    int i;

    values = parse_line(command, "FILE_ADDED", 10);
    if (!values) {
        return;
    }
    values_groups = g_strsplit(values[9], "|", 0);
    i = 0;
    while (values_groups[i]) {
        groups = g_list_append(groups, values_groups[i]);
        i++;
    }

    nntpgrab_core_emit_file_added(  FALSE, values[0], values[1], values[2], atoll(values[3]),
                                    atoll(values[4]), atoll(values[5]), atoll(values[6]),
                                    string_to_task_state(values[7]), atoi(values[8]), groups);

    g_list_free(groups);
    g_strfreev(values_groups);
    g_strfreev(values);
}

static void
emit_file_removed(const char *command)
{
    char **values;

    values = parse_line(command, "FILE_REMOVED", 4);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_file_removed(FALSE, values[0], values[1], atoll(values[2]), atoll(values[3]));

    g_strfreev(values);
}

static void
emit_task_moved(const char *command)
{
    char **values;

    values = parse_line(command, "TASK_MOVED", 5);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_task_moved(FALSE, values[0], values[1], values[2], atoi(values[3]), atoi(values[4]));

    g_strfreev(values);
}

static void
emit_collection_moved(const char *command)
{
    char **values;

    values = parse_line(command, "COLLECTION_MOVED", 3);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_collection_moved(FALSE, values[0], atoi(values[1]), atoi(values[2]));

    g_strfreev(values);
}

static void
emit_unpack_progress_update(const char *command)
{
    char **values;

    values = parse_line(command, "UNPACK_PROGRESS_UPDATE", 3);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_unpack_progress_update(FALSE, values[0], values[1], atof(values[2]));

    g_strfreev(values);
}

static void
emit_unpack_message_received(const char *command)
{
    char **values;

    values = parse_line(command, "UNPACK_MESSAGE_RECEIVED", 3);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_unpack_message_received(FALSE, values[0], values[1], values[2]);

    g_strfreev(values);
}

static void
emit_unpack_working_archive_changed(const char *command)
{
    char **values;

    values = parse_line(command, "UNPACK_WORKING_ARCHIVE_CHANGED", 3);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_unpack_working_archive_changed(FALSE, values[0], values[1], values[2]);

    g_strfreev(values);
}

static void
emit_par2_begin_verify(const char *command)
{
    char **values;

    values = parse_line(command, "PAR2_BEGIN_VERIFY", 2);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_par2_begin_verify(FALSE, values[0], values[1]);

    g_strfreev(values);
}

static void
emit_par2_load_progress_update(const char *command)
{
    char **values;

    values = parse_line(command, "PAR2_LOAD_PROGRESS_UPDATE", 4);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_par2_load_progress_update(FALSE, values[0], values[1], values[2], g_ascii_strtod(values[3], NULL));

    g_strfreev(values);
}

static void
emit_par2_file_loaded(const char *command)
{
    char **values;
    PAR2FileState state;

    values = parse_line(command, "PAR2_FILE_LOADED", 6);
    if (!values) {
        return;
    }

    if (!strcmp(values[3], "MISSING")) {
        state = PAR2_FILE_STATE_MISSING;
    } else if (!strcmp(values[3], "FOUND")) {
        state = PAR2_FILE_STATE_FOUND;
    } else if (!strcmp(values[3], "DAMAGED")) {
        state = PAR2_FILE_STATE_DAMAGED;
    } else if (!strcmp(values[3], "NO_NEW_BLOCKS_FOUND")) {
        state = PAR2_FILE_STATE_NO_NEW_BLOCKS_FOUND;
    } else {
        g_print(__FILE__ ":%i Unknown state found: %s\n", __LINE__, values[3]);
        state = PAR2_FILE_STATE_MISSING;
    }

    nntpgrab_core_emit_par2_file_loaded(FALSE, values[0], values[1], values[2], state, atoi(values[4]), atoi(values[5]));

    g_strfreev(values);
}

static void
emit_par2_recovery_file_loaded(const char *command)
{
    char **values;

    values = parse_line(command, "PAR2_RECOVERY_FILE_LOADED", 5);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_par2_recovery_file_loaded(FALSE, values[0], values[1], values[2], atoi(values[3]), atoi(values[4]));

    g_strfreev(values);
}

static void
emit_par2_repair_progress_update(const char *command)
{
    char **values;

    values = parse_line(command, "PAR2_REPAIR_PROGRESS_UPDATE", 3);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_par2_repair_progress_update(FALSE, values[0], values[1], g_ascii_strtod(values[2], NULL));

    g_strfreev(values);
}

static void
emit_par2_repair_success(const char *command)
{
    char **values;

    values = parse_line(command, "PAR2_REPAIR_SUCCESS", 2);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_par2_repair_success(FALSE, values[0], values[1]);

    g_strfreev(values);
}

static void
emit_par2_repair_failure(const char *command)
{
    char **values;

    values = parse_line(command, "PAR2_REPAIR_FAILURE", 3);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_par2_repair_failure(FALSE, values[0], values[1], atoi(values[2]));

    g_strfreev(values);
}

static void
emit_par2_no_repair_required(const char *command)
{
    char **values;

    values = parse_line(command, "PAR2_NO_REPAIR_REQUIRED", 2);
    if (!values) {
        return;
    }

    nntpgrab_core_emit_par2_no_repair_required(FALSE, values[0], values[1]);

    g_strfreev(values);
}

static gboolean
do_emit_connection_lost(gpointer data)
{
    char *errmsg = (char *) data;

    if (!get_core()) {
        g_free(errmsg);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "connection_lost", errmsg);
    g_free(errmsg);

    return FALSE;
}

void
emit_connection_lost(const char *errmsg)
{
    if (!get_core()) {
        return;
    }

    g_idle_add(do_emit_connection_lost, (gpointer) g_strdup(errmsg));
}

static gboolean
has_signal_handler_pending(int signal_id)
{
    return g_signal_has_handler_pending(glue, signal_id, 0, TRUE);
}

static void
process_message(const char *line)
{
    char *ptr;
    char *command;
    unsigned int hash;

    ptr = strstr(line, " ");
    if (ptr) {
        command = g_strndup(line, ptr - line);
    } else {
        command = g_strdup(line);
    }

    // Generate a hash of the message to optimize the processing
    hash = g_str_hash(command);

    switch(hash) {
        case 1903602327U:
            if (!strcmp(command, "CONFIG_CHANGED")) {
                if (has_signal_handler_pending(signals[CONFIG_CHANGED_SIGNAL])) {
                    nntpgrab_core_emit_config_changed(FALSE);
                }
                break;
            }

        case 1349955479U:
            if (!strncmp(command, "PART_DOWNLOAD_START", 19)) {
                if (has_signal_handler_pending(signals[PART_DOWNLOAD_START_SIGNAL])) {
                    emit_part_download_start(line);
                }
                break;
            }

        case 1006630990U:
            if (!strncmp(command, "PART_DONE", 9)) {
                if (has_signal_handler_pending(signals[PART_DONE_SIGNAL])) {
                    emit_part_done(line);
                }
                break;
            }

        case 1048928777U:
            if (!strncmp(command, "PART_FAILED", 11)) {
                if (has_signal_handler_pending(signals[PART_FAILED_SIGNAL])) {
                    emit_part_failed(line);
                }
                break;
            }

        case 424842768U:
            if (!strncmp(command, "TRAFFIC_MONITOR_UPDATE", 22)) {
                if (has_signal_handler_pending(signals[TRAFFIC_MONITOR_UPDATE_SIGNAL])) {
                    emit_traffic_monitor_update(line);
                }
                break;
            }

        case 1761576655U:
            if (!strncmp(command, "PART_PROGRESS_UPDATE", 20)) {
                if (has_signal_handler_pending(signals[PART_PROGRESS_UPDATE_SIGNAL])) {
                    emit_part_progress_update(line);
                }
                break;
            }

        case 3983341567U:
            if (!strncmp(command, "COLLECTION_ADDED", 16)) {
                if (has_signal_handler_pending(signals[COLLECTION_ADDED_SIGNAL])) {
                    g_assert(strlen(line) > 17);
                    emit_collection_added(line);
                }
                break;
            }

        case 3415304095U:
            if (!strncmp(command, "COLLECTION_REMOVED", 18)) {
                if (has_signal_handler_pending(signals[COLLECTION_REMOVED_SIGNAL])) {
                    g_assert(strlen(line) > 19);
                    nntpgrab_core_emit_collection_removed(FALSE, line + 19);
                }
                break;
            }

        case 2692520778U:
            if (!strncmp(command, "COLLECTION_MODIFIED", 19)) {
                if (has_signal_handler_pending(signals[COLLECTION_MODIFIED_SIGNAL])) {
                    g_assert(strlen(line) > 20);
                    emit_collection_modified(line);
                }
            }
            break;

        case 3528469597U:
            if (!strncmp(command, "FILE_ADDED", 10)) {
                if (has_signal_handler_pending(signals[FILE_ADDED_SIGNAL])) {
                    emit_file_added(line);
                }
                break;
            }

        case 75037821U:
            if (!strncmp(command, "FILE_REMOVED", 12)) {
                if (has_signal_handler_pending(signals[FILE_REMOVED_SIGNAL])) {
                    emit_file_removed(line);
                }
                break;
            }

        case 2847286315U:
            if (!strncmp(command, "FILE_DOWNLOAD_STATE_UPDATE", 26)) {
                if (has_signal_handler_pending(signals[FILE_DOWNLOAD_STATE_UPDATE_SIGNAL])) {
                    emit_file_download_state_update(line);
                }
                break;
            }

        case 1809569603U:
            if (!strncmp(command, "FILE_STATE_CHANGED", 18)) {
                if (has_signal_handler_pending(signals[FILE_STATE_CHANGED_SIGNAL])) {
                    emit_file_state_changed(line);
                }
                break;
            }

        case 581271225U:
            if (!strncmp(command, "CONNECTION_CONNECTING", 21)) {
                if (has_signal_handler_pending(signals[CONNECTION_CONNECTING_SIGNAL])) {
                    emit_connection_connecting(line);
                }
                break;
            }

        case 3898075848U:
            if (!strncmp(command, "CONNECTION_CONNECTED", 20)) {
                if (has_signal_handler_pending(signals[CONNECTION_CONNECTED_SIGNAL])) {
                    emit_connection_connected(line);
                }
                break;
            }

        case 1887328413U:
            if (!strncmp(command, "CONNECTION_DISCONNECT", 21)) {
                if (has_signal_handler_pending(signals[CONNECTION_DISCONNECT_SIGNAL])) {
                    emit_connection_disconnect(line);
                }
                break;
            }


        case 2671763622U:
            if (!strncmp(command, "SCHEDULAR_STATE_CHANGED", 23)) {
                if (has_signal_handler_pending(signals[SCHEDULAR_STATE_CHANGED_SIGNAL])) {
                    emit_schedular_state_changed(line);
                }
                break;
            }

        case 507002573U:
            if (!strncmp(command, "FATAL_ERROR", 11)) {
                if (has_signal_handler_pending(signals[CONNECTION_LOST_SIGNAL])) {
                    g_assert(strlen(line) > 12);
                    emit_connection_lost(line + 13);
                }
                break;
            }

        case 2480026148U:
            if (!strncmp(command, "WARNING_MESSAGE", 15)) {
                if (has_signal_handler_pending(signals[WARNING_MESSAGE_SIGNAL])) {
                    g_assert(strlen(line) > 16);
                    nntpgrab_core_emit_warning(FALSE, line + 16);
                }
                break;
            }

        case 4232637915U:
            if (!strncmp(command, "DEBUG_MESSAGE", 13)) {
                if (has_signal_handler_pending(signals[DEBUG_MESSAGE_SIGNAL])) {
                    g_assert(strlen(line) > 14);
                    nntpgrab_core_emit_debug_message(FALSE, line + 14);
                }
                break;
            }

        case 2708266841U:
            if (!strncmp(command, "TASK_MOVED", 10)) {
                if (has_signal_handler_pending(signals[TASK_MOVED_SIGNAL])) {
                    emit_task_moved(line);
                }
                break;
            }

        case 3994768818U:
            if (!strncmp(command, "COLLECTION_MOVED", 16)) {
                if (has_signal_handler_pending(signals[COLLECTION_MOVED_SIGNAL])) {
                    emit_collection_moved(line);
                }
                break;
            }

        case 1521434467:
            if (!strncmp(command, "PAR2_REPAIR_MESSAGE", 18)) {
                if (has_signal_handler_pending(signals[PAR2_REPAIR_MESSAGE_SIGNAL])) {
                    g_assert(strlen(line) > 19);
                    nntpgrab_core_emit_par2_repair_message(FALSE, line + 19);
                }
                break;
            }

        case 572446713U:
            if (!strcmp(command, "ALL_DOWNLOADS_COMPLETED")) {
                if (has_signal_handler_pending(signals[ALL_DOWNLOADS_COMPLETED_SIGNAL])) {
                    nntpgrab_core_emit_all_downloads_completed(FALSE);
                }
                break;
            }

        case 2407283982U:
            if (!strncmp(command, "UNPACK_PROGRESS_UPDATE", 22)) {
                if (has_signal_handler_pending(signals[UNPACK_PROGRESS_UPDATE_SIGNAL])) {
                    emit_unpack_progress_update(line);
                }
                break;
            }

        case 1946881958U:
            if (!strncmp(command, "UNPACK_MESSAGE_RECEIVED", 23)) {
                if (has_signal_handler_pending(signals[UNPACK_MESSAGE_RECEIVED_SIGNAL])) {
                    emit_unpack_message_received(line);
                }
                break;
            }

        case 3065743804U:
            if (!strncmp(command, "UNPACK_WORKING_ARCHIVE_CHANGED", 30)) {
                if (has_signal_handler_pending(signals[UNPACK_WORKING_ARCHIVE_CHANGED_SIGNAL])) {
                    emit_unpack_working_archive_changed(line);
                }
                break;
            }

        case 3882452637U:
            if (!strncmp(command, "PAR2_BEGIN_VERIFY", 16)) {
                if (has_signal_handler_pending(signals[PAR2_BEGIN_VERIFY_SIGNAL])) {
                    emit_par2_begin_verify(line);
                }
                break;
            }

        case 4089290096U:
            if (!strncmp(command, "PAR2_LOAD_PROGRESS_UPDATE", 24)) {
                if (has_signal_handler_pending(signals[PAR2_LOAD_PROGRESS_UPDATE_SIGNAL])) {
                    emit_par2_load_progress_update(line);
                }
                break;
            }

        case 2740675116U:
            if (!strncmp(command, "PAR2_RECOVERY_FILE_LOADED", 24)) {
                if (has_signal_handler_pending(signals[PAR2_RECOVERY_FILE_LOADED_SIGNAL])) {
                    emit_par2_recovery_file_loaded(line);
                }
                break;
            }

        case 2733496506U:
            if (!strncmp(command, "PAR2_FILE_LOADED", 15)) {
                if (has_signal_handler_pending(signals[PAR2_FILE_LOADED_SIGNAL])) {
                    emit_par2_file_loaded(line);
                }
                break;
            }

        case 3561788055U:
            if (!strncmp(command, "PAR2_REPAIR_PROGRESS_UPDATE", 26)) {
                if (has_signal_handler_pending(signals[PAR2_REPAIR_PROGRESS_UPDATE_SIGNAL])) {
                    emit_par2_repair_progress_update(line);
                }
                break;
            }

        case 3774902502U:
            if (!strncmp(command, "PAR2_REPAIR_FAILURE", 18)) {
                if (has_signal_handler_pending(signals[PAR2_REPAIR_FAILURE_SIGNAL])) {
                    emit_par2_repair_failure(line);
                }
                break;
            }

        case 2994306911U:
            if (!strncmp(command, "PAR2_REPAIR_SUCCESS", 18)) {
                if (has_signal_handler_pending(signals[PAR2_REPAIR_SUCCESS_SIGNAL])) {
                    emit_par2_repair_success(line);
                }
                break;
            }

        case 1706996929U:
            if (!strncmp(command, "PAR2_NO_REPAIR_REQUIRED", 22)) {
                if (has_signal_handler_pending(signals[PAR2_NO_REPAIR_REQUIRED_SIGNAL])) {
                    emit_par2_no_repair_required(line);
                }
                break;
            }

        default:
            {
                char *msg;

                msg = g_strdup_printf("Unknown data received: %s\n", line);
                DEBUG(msg);
                g_free(msg);

                msg = g_strdup_printf("hash = %u\n", hash);
                DEBUG(msg);
                g_free(msg);
            }

            break;
    };

    g_free(command);

    // Give other threads some time to handle the signals
    g_thread_yield();
}

static gpointer
callbacks_thread(gpointer data)
{
    NNTPGrabGlue *core = data;

    do {
        gboolean error_flag;
        char *line = NULL;

        if (!read_line(&core->socket_listener, &line, 1, &error_flag)) {
            if (error_flag) {
                // An error occured!
                char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s", __LINE__, line);

                emit_connection_lost(msg);
                DEBUG(msg);

                g_free(line);
                g_free(msg);

                return NULL;
            }

            if (line) {
                g_free(line);
            }

            // No data received, sleep
            g_usleep(G_USEC_PER_SEC / 10);

            continue;
        }

        process_message(line);

        g_free(line);
    } while (!abort_flag);

    return NULL;
}

static void
initialize_callbacks(NNTPGrabGlue *core)
{
    char *line = NULL;

    abort_flag = FALSE;
    write_line(&core->socket_listener, "MODE LISTENER\n");
    if (!read_line(&core->socket_listener, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        return;
    }

    if (strcmp(line, "OK")) {
        char *msg = g_strdup_printf(__FILE__ ":%i unknown response received: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        return;
    }

    g_free(line);

    thread_listener = g_thread_create(callbacks_thread, core, TRUE, NULL);

    g_assert(thread_listener != NULL);
}

/*********************************
 Public functions
 *********************************/
static GStaticMutex command_mutex = G_STATIC_MUTEX_INIT;

NNTPGrabGlue *
nntpgrab_glue_init(int glue_version, char **err)
{
#ifdef WIN32
    WORD wVersionRequested;
    WSADATA wsaData;
    int ret;
#endif
    static gboolean initialized = FALSE;

    if (initialized) {
        if (err) {
            *err = g_strdup_printf(_("NNTPGrab Core already initialized"));
        }
        return NULL;
    }

    g_type_init();
    if (!g_thread_supported ()) g_thread_init (NULL);

    // Check if the API version matches
    if (glue_version != NNTPGRAB_GLUE_VERSION) {
        if (err) {
            *err = g_strdup_printf(_("NNTPGrab Glue API mismatch (Glue API version = %i, frontend's version = %i)"), NNTPGRAB_GLUE_VERSION, glue_version);
        }
        return NULL;
    }

#ifdef WIN32
    // Initialise Winsock
    wVersionRequested = MAKEWORD( 2, 2 );

    ret = WSAStartup( wVersionRequested, &wsaData );
    if ( ret != 0 ) {
        /* Tell the user that we could not find a usable */
        /* WinSock DLL.                                  */
        if (err) {
            *err = g_strdup_printf(_("Winsock could not be initialised"));
        }
        return NULL;
    }

    /* Confirm that the WinSock DLL supports 2.2.*/
    /* Note that if the DLL supports versions greater    */
    /* than 2.2 in addition to 2.2, it will still return */
    /* 2.2 in wVersion since that is the version we      */
    /* requested.                                        */

    if ( LOBYTE( wsaData.wVersion ) != 2 ||
         HIBYTE( wsaData.wVersion ) != 2 ) {
        /* Tell the user that we could not find a usable */
        /* WinSock DLL.                                  */
        WSACleanup( );

        if (err) {
            *err = g_strdup_printf(_("Winsock 2.2 or higher is required for this program"));
        }

        return NULL;
    }
#endif

    glue = g_object_new(NNTPGRAB_TYPE_GLUE, NULL);
    glue->glue_version = glue_version;

    return glue;
}

static gboolean
ensure_core_library_is_loaded(char **err)
{
    if (module) {
        return TRUE;
    }

#ifdef WIN32
#define EXT "-0.dll"
#elif DARWIN
#define EXT ".0.dylib"
#else
#define EXT ".so.0"
#endif

    module = g_module_open("libnntpgrab" EXT, G_MODULE_BIND_LAZY | G_MODULE_BIND_LOCAL);
    if (!module) {
        if (err) {
            *err = g_strdup_printf(_("Error while loading NNTPGrab Core library:\n%s\n"), g_module_error());
        } else {
            g_error(_("Error while loading NNTPGrab Core library:\n%s\n"), g_module_error());
        }
        return FALSE;
    }

    g_module_make_resident(module);
    return TRUE;
}

static void
bind_func(const char *name, void **func, char **err)
{
    if (!ensure_core_library_is_loaded(err)) {
        return;
    }

    if (!(*func)) {
        g_module_symbol(module, name, func);
    }

    if (!(*func)) {
        if (err) {
            *err = g_strdup_printf(_("Unable to bind the function '%s'\n%s"), name, g_module_error());
        } else {
            g_critical(_("Unable to bind the function '%s'\n%s"), name, g_module_error());
        }
    }
}

NNTPGrabCore *
nntpgrab_core_init(int version, char **err, char **warnings)
{
    static NNTPGrabCore *(*func) (int, char **, char **) = NULL;

    if (!ensure_core_library_is_loaded(err)) {
        return NULL;
    }

    bind_func("nntpgrab_core_init", (void**) &func, err);
    g_return_val_if_fail(func != NULL, NULL);

    return func(version, err, warnings);
}

ngboolean
nntpgrab_glue_connect(NNTPGrabGlue *obj, const char *hostname, int port, const char *username, const char *password, ngboolean use_ssl, char **err, char **warnings)
{
    NNTPGrabGlue *core = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    int api_version = 0;

    g_assert(core != NULL);

    if (!hostname) {
        core->standalone_core = nntpgrab_core_init(NNTPGRAB_API_VERSION, err, warnings);
        if (core->standalone_core) {
            connect_signal_handlers(core->standalone_core);
        }
        return (core->standalone_core != NULL);
    }

    connect_to_server(&core->socket_listener, hostname, port, &line);
    if (core->socket_listener.socket_id < 0) {
        if (err) {
            if (line) {
                *err = g_strdup(line);
            } else {
                *err = g_strdup_printf(__FILE__ ":%i connect_to_server() FAILED", __LINE__);
            }
        }

        if (line) {
            g_free(line);
        }

        return FALSE;
    }

    g_assert(line);

    // API check
    if (sscanf(line, "NNTPGrab Server - API version %i", &api_version) != 1) {
        if (err) {
            *err = g_strdup_printf(__FILE__ ":%i Unknown response received from server:\n%s\n", __LINE__, line);
        }

        g_free(line);

        return FALSE;
    }

    g_free(line);

    if (api_version != NNTPGRAB_API_VERSION) {
        if (err) {
            *err = g_strdup_printf(_("NNTPGrab Server API mismatch (Server API version = %i, frontend's version = %i)"), NNTPGRAB_API_VERSION, api_version);
        }

        return FALSE;
    }

    initialize_callbacks(core);

    g_static_mutex_lock(&command_mutex);

    line = NULL;
    connect_to_server(&core->socket_command, hostname, port, &line);
    if (core->socket_command.socket_id < 0) {
        if (err) {
            *err = g_strdup_printf(__FILE__ ":%i %s", __LINE__, (line ? line : ""));
        }

        if (line) {
            g_free(line);
        }

        g_static_mutex_unlock(&command_mutex);

        return FALSE;
    }

    g_free(line);

    write_line(&core->socket_command, "MODE COMMAND\n");
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (strcmp(line, "OK")) {
        char *msg = g_strdup_printf(__FILE__ ":%i unknown response received: %s\n", __LINE__, (line ? line : ""));
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    g_free(line);

    g_static_mutex_unlock(&command_mutex);

    return TRUE;
}

ngboolean
nntpgrab_glue_get_is_connected(NNTPGrabGlue *obj)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);

    if (glue->socket_listener.socket_id < 0) {
        return FALSE;
    }

    if (glue->socket_command.socket_id < 0) {
        return FALSE;
    }

    return TRUE;
}

void
nntpgrab_glue_cleanup(NNTPGrabGlue *obj)
{
    abort_flag = TRUE;
    if (thread_listener) {
        g_thread_join(thread_listener);
        thread_listener = NULL;
    }

    g_object_unref(obj);
}

void
nntpgrab_glue_kill_server(NNTPGrabGlue *obj)
{
    g_static_mutex_lock(&command_mutex);

    write_line(&glue->socket_command, "KILL_SERVER\n");

    /* The connection will automatically be killed, so it doesn't make
     * sense to do any error handling here */

    g_static_mutex_unlock(&command_mutex);
}

static NGList *
nntpgrab_glue_config_get_avail_servers(NNTPGrabGlue *obj)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    NGList *ret = NULL;

    g_static_mutex_lock(&command_mutex);

    write_line(&glue->socket_command, "CONFIG_GET_AVAIL_SERVERS\n");

    while (TRUE) {
        if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
            NGList *list;
            char *msg;

            msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
            DEBUG(msg);
            g_free(msg);
            g_free(line);

            list = ret;
            while (list) {
                g_free(list->data);
                list = ng_list_next(list);
            }
            g_list_free((GList*) ret);

            g_static_mutex_unlock(&command_mutex);
            return NULL;
        }

        if (!strncmp(line, "INVALID_COMMAND", 15)) {
            char *msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
            DEBUG(msg);
            g_free(msg);
            g_free(line);
            g_static_mutex_unlock(&command_mutex);
            return FALSE;
        }

        if (!strncmp(line, "NULL", 4)) {
            break;
        }

        ret = ng_list_append(ret, line);
    }

    g_static_mutex_unlock(&command_mutex);

    return ret;
}

NGList *
nntpgrab_config_get_avail_servers(NNTPGrabGlue *obj)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static NGList *(*func) (NNTPGrabCore *) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_config_get_avail_servers", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, NULL);
        return func(glue->standalone_core);
    } else {
        return nntpgrab_glue_config_get_avail_servers(glue);
    }
}

void
nntpgrab_config_free_avail_servers(NNTPGrabGlue *obj, NGList *servers)
{
    NGList *list;

    list = servers;
    while (list) {
        g_free(list->data);
        list = ng_list_next(list);
    }

    g_list_free((GList*) servers);
}

static ngboolean
nntpgrab_glue_config_get_server_info(NNTPGrabGlue *obj, const char *servername, ConfigServer *ret)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    ConfigServer server;

    g_assert(servername);
    g_assert(ret);

    g_static_mutex_lock(&command_mutex);

    line = g_strdup_printf("CONFIG_GET_SERVER_INFO %s\n", servername);
    write_line(&glue->socket_command, line);
    g_free(line);

    // Servername
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        char *msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "NULL", 4)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    strncpy(server.servername, line, sizeof(server.servername));
    g_free(line);

    // Hostname
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    strncpy(server.hostname, line, sizeof(server.hostname));
    g_free(line);

    // Port
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    server.port = atoi(line);
    g_free(line);

    // Username
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    strncpy(server.username, line, sizeof(server.username));
    g_free(line);

    // Password
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    strncpy(server.password, line, sizeof(server.password));
    g_free(line);

    // Max threads
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    server.max_threads = atoi(line);
    g_free(line);

    // Priority
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    server.priority = atoi(line);
    g_free(line);

    // Use SSL
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        server.use_ssl = TRUE;
    } else {
        server.use_ssl = FALSE;
    }

    g_free(line);

    // Enabled
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        server.enabled = TRUE;
    } else {
        server.enabled = FALSE;
    }

    g_free(line);

    // NULL
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "NULL", 4)) {
        g_free(line);
        g_static_mutex_unlock(&command_mutex);

        memcpy(ret, &server, sizeof(ConfigServer));
        return TRUE;
    }

    g_free(line);

    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

gboolean
nntpgrab_config_get_server_info(NNTPGrabGlue *obj, const char *servername, ConfigServer *ret)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static gboolean (*func) (NNTPGrabCore *, const char *, ConfigServer *) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_config_get_server_info", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, FALSE);
        return func(glue->standalone_core, servername, ret);
    } else {
        return nntpgrab_glue_config_get_server_info(glue, servername, ret);
    }
}

static gboolean
nntpgrab_glue_config_add_server(NNTPGrabGlue *obj, ConfigServer new_server, char **errmsg)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    char *use_ssl_str;
    char *enabled_str;
    char *msg;

    g_assert(strlen(new_server.servername) > 0);
    g_assert(strlen(new_server.hostname) > 0);

    g_static_mutex_lock(&command_mutex);

    if (new_server.use_ssl) {
        use_ssl_str = "TRUE";
    } else {
        use_ssl_str = "FALSE";
    }

    if (new_server.enabled) {
        enabled_str = "TRUE";
    } else {
        enabled_str = "FALSE";
    }

    line = g_strdup_printf("CONFIG_ADD_SERVER %s\t%s\t%i\t%s\t%s\t%i\t%i\t%s\t%s\n", new_server.servername, new_server.hostname, new_server.port, new_server.username, new_server.password, new_server.max_threads, new_server.priority, use_ssl_str, enabled_str);
    write_line(&glue->socket_command, line);
    g_free(line);

    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        if (errmsg) {
            *errmsg = g_strdup(msg);
        }
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        if (errmsg) {
            *errmsg = g_strdup(msg);
        }
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);

        if (errmsg) {
            if (strlen(line) > 6) {
                *errmsg = g_strdup(line + 6);
            } else {
                *errmsg = g_strdup(_("Unknown error occured while executing command"));
            }
        }

        g_free(line);

        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    if (errmsg) {
        *errmsg = g_strdup(msg);
    }
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

gboolean
nntpgrab_config_add_server(NNTPGrabGlue *obj, ConfigServer new_server, char **errmsg)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static gboolean (*func) (NNTPGrabCore *, ConfigServer, char **) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_config_add_server", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, FALSE);
        return func(glue->standalone_core, new_server, errmsg);
    } else {
        return nntpgrab_glue_config_add_server(glue, new_server, errmsg);
    }
}

static gboolean
nntpgrab_glue_config_del_server(NNTPGrabGlue *obj, const char *servername, char **errmsg)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    char *msg;

    g_assert(servername);

    g_static_mutex_lock(&command_mutex);

    line = g_strdup_printf("CONFIG_DEL_SERVER %s\n", servername);
    write_line(&glue->socket_command, line);
    g_free(line);

    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        if (errmsg) {
            *errmsg = g_strdup(msg);
        }
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        if (errmsg) {
            *errmsg = g_strdup(msg);
        }
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);

        if (errmsg) {
            if (strlen(line) > 6) {
                *errmsg = g_strdup(line + 6);
            } else {
                *errmsg = g_strdup(_("Unknown error occured while executing command"));
            }
        }

        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    if (errmsg) {
        *errmsg = g_strdup(msg);
    }
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

gboolean
nntpgrab_config_del_server(NNTPGrabGlue *obj, const char *servername, char **errmsg)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static gboolean (*func) (NNTPGrabCore *, const char *, char **) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_config_del_server", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, FALSE);
        return func(glue->standalone_core, servername, errmsg);
    } else {
        return nntpgrab_glue_config_del_server(glue, servername, errmsg);
    }
}

static gboolean
nntpgrab_glue_config_edit_server(NNTPGrabGlue *obj, const char *servername, ConfigServer new_server, char **errmsg)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    char *use_ssl_str;
    char *enabled_str;
    char *msg;

    g_assert(servername);
    g_assert(strlen(new_server.servername) > 0);
    g_assert(strlen(new_server.hostname) > 0);

    g_static_mutex_lock(&command_mutex);

    if (new_server.use_ssl) {
        use_ssl_str = "TRUE";
    } else {
        use_ssl_str = "FALSE";
    }

    if (new_server.enabled) {
        enabled_str = "TRUE";
    } else {
        enabled_str = "FALSE";
    }

    line = g_strdup_printf("CONFIG_EDIT_SERVER %s\t%s\t%s\t%i\t%s\t%s\t%i\t%i\t%s\t%s\n", servername, new_server.servername, new_server.hostname, new_server.port, new_server.username, new_server.password, new_server.max_threads, new_server.priority, use_ssl_str, enabled_str);
    write_line(&glue->socket_command, line);
    g_free(line);

    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        if (errmsg) {
            *errmsg = g_strdup(msg);
        }
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        if (errmsg) {
            *errmsg = g_strdup(msg);
        }
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);
        if (errmsg) {
            if (strlen(line) > 6) {
                *errmsg = g_strdup(line + 6);
            } else {
                *errmsg = g_strdup(_("Unknown error occured while executing command"));
            }
        }

        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    if (errmsg) {
        *errmsg = g_strdup(msg);
    }
    g_free(msg);
    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

gboolean
nntpgrab_config_edit_server(NNTPGrabGlue *obj, const char *servername, ConfigServer new_server, char **errmsg)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static gboolean (*func) (NNTPGrabCore *, const char *, ConfigServer, char **) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_config_edit_server", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, FALSE);
        return func(glue->standalone_core, servername, new_server, errmsg);
    } else {
        return nntpgrab_glue_config_edit_server(glue, servername, new_server, errmsg);
    }
}

static ConfigOpts
nntpgrab_glue_config_get_opts(NNTPGrabGlue *obj)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    ConfigOpts opts;
    char *msg;

    g_static_mutex_lock(&command_mutex);

    memset(&opts, 0, sizeof(opts));

    write_line(&glue->socket_command, "CONFIG_GET_OPTS\n");

    // Download directory
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    strncpy(opts.download_directory, line, sizeof(opts.download_directory) - 1);
    g_free(line);

    // Temp directory
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    strncpy(opts.temp_directory, line, sizeof(opts.temp_directory) - 1);
    g_free(line);

    // Enable intelligent PAR2 downloading
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    if (!strncmp(line, "FALSE", 5)) {
        opts.enable_intelligent_par2_downloading = FALSE;
    } else if (!strncmp(line, "TRUE", 4)) {
        opts.enable_intelligent_par2_downloading = TRUE;
    } else {
        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
    }

    g_free(line);

    // Enable PAR2 repair
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    if (!strncmp(line, "FALSE", 5)) {
        opts.enable_par2_repair = FALSE;
    } else if (!strncmp(line, "TRUE", 4)) {
        opts.enable_par2_repair = TRUE;
    } else {
        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
    }

    g_free(line);

    // Enable auto import
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    if (!strncmp(line, "FALSE", 5)) {
        opts.enable_auto_import = FALSE;
    } else if (!strncmp(line, "TRUE", 4)) {
        opts.enable_auto_import = TRUE;
    } else {
        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
    }

    g_free(line);

    // Auto import directory
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    strncpy(opts.auto_import_directory, line, sizeof(opts.auto_import_directory) - 1);
    g_free(line);

    // Move after import
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    if (!strncmp(line, "FALSE", 5)) {
        opts.move_file_after_auto_import = FALSE;
    } else if (!strncmp(line, "TRUE", 4)) {
        opts.move_file_after_auto_import = TRUE;
    } else {
        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
    }

    g_free(line);

    // Automatic unpack
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    if (!strncmp(line, "FALSE", 5)) {
        opts.enable_auto_unpack = FALSE;
    } else if (!strncmp(line, "TRUE", 4)) {
        opts.enable_auto_unpack = TRUE;
    } else {
        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
    }

    g_free(line);

    // NULL
    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    if (!strncmp(line, "NULL", 4)) {
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return opts;
    }

    g_static_mutex_unlock(&command_mutex);

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);

    return opts;
}

ConfigOpts
nntpgrab_config_get_opts(NNTPGrabGlue *obj)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static ConfigOpts (*func) (NNTPGrabCore *) = NULL;

    if (glue->standalone_core) {
        ConfigOpts dummy;
        memset(&dummy, 0, sizeof(ConfigOpts));

        bind_func("nntpgrab_core_config_get_opts", (void**) &func, NULL);

        g_return_val_if_fail(func != NULL, dummy);

        return func(glue->standalone_core);
    } else {
        return nntpgrab_glue_config_get_opts(glue);
    }
}

static void
nntpgrab_glue_config_set_opts(NNTPGrabGlue *obj, ConfigOpts opts)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    char *enable_intelligent_par2_str;
    char *enable_par2_repair_str;
    char *enable_auto_import_str;
    char *move_file_str;
    char *enable_auto_unpack_str;
    char *msg;

    g_static_mutex_lock(&command_mutex);

    if (opts.enable_intelligent_par2_downloading) {
        enable_intelligent_par2_str = "TRUE";
    } else {
        enable_intelligent_par2_str = "FALSE";
    }

    if (opts.enable_par2_repair) {
        enable_par2_repair_str = "TRUE";
    } else {
        enable_par2_repair_str = "FALSE";
    }

    if (opts.enable_auto_import) {
        enable_auto_import_str = "TRUE";
    } else {
        enable_auto_import_str = "FALSE";
    }

    if (opts.move_file_after_auto_import) {
        move_file_str = "TRUE";
    } else {
        move_file_str = "FALSE";
    }

    if (opts.enable_auto_unpack) {
        enable_auto_unpack_str = "TRUE";
    } else {
        enable_auto_unpack_str = "FALSE";
    }

    line = g_strdup_printf("CONFIG_SET_OPTS %s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n", opts.download_directory, opts.temp_directory, enable_intelligent_par2_str, enable_par2_repair_str, enable_auto_import_str, opts.auto_import_directory, move_file_str, enable_auto_unpack_str);
    write_line(&glue->socket_command, line);
    g_free(line);

    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);
}

void
nntpgrab_config_set_opts(NNTPGrabGlue *obj, ConfigOpts opts)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static void (*func) (NNTPGrabCore *, ConfigOpts) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_config_set_opts", (void**) &func, NULL);
        g_return_if_fail(func != NULL);
        func(glue->standalone_core, opts);
    } else {
        nntpgrab_glue_config_set_opts(glue, opts);
    }
}

static ngboolean
nntpgrab_glue_config_get_folder_listing(NNTPGrabGlue *obj, const char *parent, NGList **folders)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    char **values;
    NNTPGrabFolder *folder;

    g_assert(folders);
    *folders = NULL;

    g_static_mutex_lock(&command_mutex);

    line = g_strdup_printf("CONFIG_GET_FOLDER_LISTING %s\n", (parent ? parent : ""));
    write_line(&glue->socket_command, line);
    g_free(line);

    do {
        if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
            char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
            DEBUG(msg);
            g_free(msg);
            g_free(line);
            g_static_mutex_unlock(&command_mutex);
            return FALSE;
        }

        if (!strncmp(line, "INVALID_COMMAND", 15)) {
            char *msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
            DEBUG(msg);
            g_free(msg);
            g_free(line);
            g_static_mutex_unlock(&command_mutex);
            return FALSE;
        }

        if (!strncmp(line, "FALSE", 5)) {
            g_static_mutex_unlock(&command_mutex);
            g_free(line);
            return FALSE;
        }

        if (!strncmp(line, "NULL", 4)) {
            g_free(line);
            break;
        }

        values = g_strsplit(line, "\t", 0);
        if (values && values[0] && values[1]) {
            folder = g_slice_new0(NNTPGrabFolder);

            strncpy(folder->folder, values[0], sizeof(folder->folder) - 1);
            if (!strcmp(values[1], "TRUE")) {
                folder->has_subfolders = TRUE;
            } else {
                folder->has_subfolders = FALSE;
            }

            *folders = ng_list_append(*folders, folder);
        }

        if (values) {
            g_strfreev(values);
        }

        g_free(line);
    } while (TRUE);

    g_static_mutex_unlock(&command_mutex);

    return TRUE;
}

ngboolean
nntpgrab_config_get_folder_listing(NNTPGrabGlue *obj, const char *parent, NGList **folders)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static ngboolean (*func) (NNTPGrabCore *, const char *, NGList **) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_config_get_folder_listing", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, FALSE);
        return func(glue->standalone_core, parent, folders);
    } else {
        return nntpgrab_glue_config_get_folder_listing(glue, parent, folders);
    }
}

void
nntpgrab_config_free_folder_listing(NNTPGrabGlue *obj, NGList *folders)
{
    nntpgrab_utils_free_folder_listing(folders);
}

static ngboolean
nntpgrab_glue_schedular_start(NNTPGrabGlue *obj)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    char *msg;

    g_static_mutex_lock(&command_mutex);

    write_line(&glue->socket_command, "SCHEDULAR_START\n");

    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

gboolean
nntpgrab_schedular_start(NNTPGrabGlue *obj)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static gboolean (*func) (NNTPGrabCore *) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_schedular_start", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, FALSE);
        return func(glue->standalone_core);
    } else {
        return nntpgrab_glue_schedular_start(glue);
    }
}

static gboolean
nntpgrab_glue_schedular_stop(NNTPGrabGlue *obj, gboolean wait)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    char *wait_str;
    char *msg;

    g_static_mutex_lock(&command_mutex);

    if (wait) {
        wait_str = "TRUE";
    } else {
        wait_str = "FALSE";
    }

    line = g_strdup_printf("SCHEDULAR_STOP %s\n", wait_str);
    write_line(&glue->socket_command, line);
    g_free(line);

    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

gboolean
nntpgrab_schedular_stop(NNTPGrabGlue *obj, gboolean wait)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static gboolean (*func) (NNTPGrabCore *, gboolean) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_schedular_stop", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, FALSE);
        return func(glue->standalone_core, wait);
    } else {
        return nntpgrab_glue_schedular_stop(glue, wait);
    }
}

static SchedularState
nntpgrab_glue_schedular_get_state(NNTPGrabGlue *obj)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    char *msg;

    g_static_mutex_lock(&command_mutex);

    write_line(&glue->socket_command, "SCHEDULAR_GET_STATE\n");

    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "RUNNING", 7)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return SCHEDULAR_STATE_RUNNING;
    }

    if (!strncmp(line, "STOPPING", 8)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return SCHEDULAR_STATE_STOPPING;
    }

    if (!strncmp(line, "STOPPED", 7)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return SCHEDULAR_STATE_STOPPED;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return SCHEDULAR_STATE_STOPPED;
}

SchedularState
nntpgrab_schedular_get_state(NNTPGrabGlue *obj)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static SchedularState (*func) (NNTPGrabCore *) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_schedular_get_state", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, SCHEDULAR_STATE_STOPPED);
        return func(glue->standalone_core);
    } else {
        return nntpgrab_glue_schedular_get_state(glue);
    }
}

static ngboolean
nntpgrab_glue_schedular_add_task_to_queue(NNTPGrabGlue *obj, const char *collection_name, const char *subject, const char *poster, time_t stamp, nguint64 file_size, NGList *groups, NGList *parts, char **errmsg)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    NGList *list;
    char *groups_str = NULL;
    char *parts_str = NULL;
    char *msg;

    g_assert(collection_name);
    g_assert(subject);
    g_assert(poster);
    g_assert(groups);
    g_assert(parts);

    if (errmsg) {
        *errmsg = NULL;
    }

    g_static_mutex_lock(&command_mutex);

    list = groups;
    while (list) {
        if (!groups_str) {
            groups_str = g_strdup(list->data);
        } else {
            char *tmp = g_strdup_printf("%s|%s", groups_str, (char*) list->data);
            g_free(groups_str);
            groups_str = tmp;
        }

        list = ng_list_next(list);
    }

    list = parts;
    while (list) {
        NNTPGrabPart *segment = list->data;

        if (!parts_str) {
            parts_str = g_strdup_printf("%s,%i,%i", segment->message_id, segment->partnum, segment->size);
        } else {
            char *tmp = g_strdup_printf("%s|%s,%i,%i", parts_str, segment->message_id, segment->partnum, segment->size);
            g_free(parts_str);
            parts_str = tmp;
        }

        list = ng_list_next(list);
    }

    line = g_strdup_printf("SCHEDULAR_ADD_TASK_TO_QUEUE %s\t%s\t%s\t%"G_GUINT64_FORMAT"\t%"G_GUINT64_FORMAT"\t%s\t%s\n", collection_name, subject, poster, (guint64) stamp, file_size, groups_str, parts_str);

    write_line(&glue->socket_command, line);
    g_free(line);
    g_free(groups_str);
    g_free(parts_str);

    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);

        if (errmsg) {
            *errmsg = g_strdup_printf(_("Connection lost to the NNTPGrab Server: %s\n"), line);
        }
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);

        if (errmsg) {
            *errmsg = g_strdup_printf(_("INVALID_COMMAND received\n"));
        }
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        if (strlen(line) > 6 && errmsg) {
            *errmsg = g_strdup(line + 6);
        }

        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

ngboolean
nntpgrab_schedular_add_task_to_queue(NNTPGrabGlue *obj, const char *collection_name, const char *subject, const char *poster, time_t stamp, nguint64 file_size, NGList *groups, NGList *parts, char **errmsg)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static ngboolean (*func) (NNTPGrabCore *, const char *, const char *, const char *, time_t, nguint64, NGList *, NGList *, char **) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_schedular_add_task_to_queue", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, FALSE);
        return func(glue->standalone_core, collection_name, subject, poster, stamp, file_size, groups, parts, errmsg);
    } else {
        return nntpgrab_glue_schedular_add_task_to_queue(glue, collection_name, subject, poster, stamp, file_size, groups, parts, errmsg);
    }
}

static ngboolean
nntpgrab_glue_schedular_del_task_from_queue(NNTPGrabGlue *obj, const char *collection_name, const char *subject, char **errmsg)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    char *msg;

    if (errmsg) {
        *errmsg = NULL;
    }

    g_static_mutex_lock(&command_mutex);

    g_assert(collection_name);
    // There is no assertion on the subject because the user can send a request
    // to remove all the items in the given collection

    line = g_strdup_printf("SCHEDULAR_DEL_TASK_FROM_QUEUE %s\t%s\n", collection_name, (subject ? subject : ""));
    write_line(&glue->socket_command, line);
    g_free(line);

    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        if (strlen(line) > 6 && errmsg) {
            *errmsg = g_strdup(line + 6);
        }

        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

ngboolean
nntpgrab_schedular_del_task_from_queue(NNTPGrabGlue *obj, const char *collection_name, const char *subject, char **errmsg)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static ngboolean (*func) (NNTPGrabCore *, const char *, const char *, char **) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_schedular_del_task_from_queue", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, FALSE);
        return func(glue->standalone_core, collection_name, subject, errmsg);
    } else {
        return nntpgrab_glue_schedular_del_task_from_queue(glue, collection_name, subject, errmsg);
    }
}

static ngboolean
nntpgrab_glue_schedular_restart_task(NNTPGrabGlue *obj, const char *collection_name, const char *subject, char **errmsg)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    char *msg;

    if (errmsg) {
        *errmsg = NULL;
    }

    g_static_mutex_lock(&command_mutex);

    g_assert(collection_name);
    // There is no assertion on the subject because the user can send a request
    // to remove all the items in the given collection

    line = g_strdup_printf("SCHEDULAR_RESTART_TASK %s\t%s\n", collection_name, (subject ? subject : ""));
    write_line(&glue->socket_command, line);
    g_free(line);

    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        if (strlen(line) > 6 && errmsg) {
            *errmsg = g_strdup(line + 6);
        }

        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

ngboolean
nntpgrab_schedular_restart_task(NNTPGrabGlue *obj, const char *collection_name, const char *subject, char **errmsg)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static ngboolean (*func) (NNTPGrabCore *, const char *, const char *, char **) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_schedular_restart_task", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, FALSE);
        return func(glue->standalone_core, collection_name, subject, errmsg);
    } else {
        return nntpgrab_glue_schedular_restart_task(glue, collection_name, subject, errmsg);
    }
}

static ngboolean
nntpgrab_glue_schedular_save_queue(NNTPGrabGlue *obj, char **errmsg)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    char *msg;

    if (errmsg) {
        *errmsg = NULL;
    }

    g_static_mutex_lock(&command_mutex);

    write_line(&glue->socket_command, "SCHEDULAR_SAVE_QUEUE\n");

    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        if (strlen(line) > 6 && errmsg) {
            *errmsg = g_strdup(line + 6);
        }

        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

ngboolean
nntpgrab_schedular_save_queue(NNTPGrabGlue *obj, char **errmsg)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static ngboolean (*func) (NNTPGrabCore *, char **) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_schedular_save_queue", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, FALSE);
        return func(glue->standalone_core, errmsg);
    } else {
        return nntpgrab_glue_schedular_save_queue(glue, errmsg);
    }
}

static void
nntpgrab_glue_schedular_foreach_task(NNTPGrabGlue *obj, ForeachCollectionFunc collection_func, ForeachFileFunc file_func, ForeachGroupFunc group_func, void *data)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    char *collection_name = NULL;

    g_static_mutex_lock(&command_mutex);

    write_line(&glue->socket_command, "SCHEDULAR_FOREACH_TASK\n");

    do {
        if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
            char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
            DEBUG(msg);
            g_free(msg);
            g_static_mutex_unlock(&command_mutex);
            g_free(line);
            return;
        }

        if (!strncmp(line, "INVALID_COMMAND", 15)) {
            char *msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
            DEBUG(msg);
            g_free(msg);
            g_static_mutex_unlock(&command_mutex);
            g_free(line);
            return;
        }

        if (!strncmp(line, "NULL", 4)) {
            g_free(line);
            break;
        }

        g_assert(strlen(line) > 1);

        // Collection
        if (!strncmp(line, "COLLECTION", 10)) {
            char **values;
            guint64 file_size;
            guint64 file_size_remaining;

            values = g_strsplit(line + 11, "\t", 0);

            if (!(values && values[0] && values[1] && values[2] && values[3] && values[4])) {
                char *msg;

                if (values) {
                    g_strfreev(values);
                }

                if (collection_name) {
                    g_free(collection_name);
                }

                msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
                DEBUG(msg);
                g_free(msg);

                g_free(line);
                g_static_mutex_unlock(&command_mutex);
                return;
            }

            file_size = g_ascii_strtoull(values[2], (char**) NULL, 10);
            file_size_remaining = g_ascii_strtoull(values[3], (char**) NULL, 10);

            g_free(line);

            if (collection_func) {
                collection_func(values[0], values[1], file_size, file_size_remaining, atoi(values[4]), data);
            }

            collection_name = g_strdup(values[0]);

            g_strfreev(values);

            continue;
        }

        // File
        if (line[0] == '\t') {
            char **values;
            TaskState status;
            guint64 file_size;
            guint64 file_size_remaining;
            gint64 stamp;

            g_assert(collection_name != NULL);

            values = g_strsplit(line + 6, "\t", 0);

            if (!(values && values[0] && values[1] && values[2] && values[3] && values[4] && values[5] && values[6] && values[7] && values[8] && values[9] && values[10] && values[11])) {
                char *msg;

                if (values) {
                    g_strfreev(values);
                }

                msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
                DEBUG(msg);
                g_free(msg);

                g_free(collection_name);
                g_free(line);
                g_static_mutex_unlock(&command_mutex);
                return;
            }

            status = string_to_task_state(values[9]);

#if GLIB_CHECK_VERSION(2,12,0)
            file_size = g_ascii_strtoull(values[3], (char **) NULL, 10);
            file_size_remaining = g_ascii_strtoull(values[4], (char **) NULL, 10);
            stamp = g_ascii_strtoll(values[2], (char **) NULL, 10);
#else
            file_size = strtoull(values[3], (char **) NULL, 10);
            file_size_remaining = strtoull(values[4], (char **) NULL, 10);
            stamp = strtoll(values[2], (char **) NULL, 10);
#endif

            if (file_func) {
                file_func(collection_name, values[0], values[1], stamp, file_size, file_size_remaining, atoi(values[5]), atoi(values[6]), atoi(values[7]), atoi(values[8]), status, values[10], data);
            }

            if (group_func) {
                char **values_groups = g_strsplit(values[11], "|", 0);
                int i = 0;
                while (values_groups[i]) {
                    group_func(collection_name, values[0], values_groups[i], data);
                    i++;
                }
                g_strfreev(values_groups);
            }

            g_strfreev(values);
        }

        g_free(line);
    } while(TRUE);

    if (collection_name) {
        g_free(collection_name);
    }

    g_static_mutex_unlock(&command_mutex);
}

void
nntpgrab_schedular_foreach_task(NNTPGrabGlue *obj, ForeachCollectionFunc collection_func, ForeachFileFunc file_func, ForeachGroupFunc group_func, void *data)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static void (*func) (NNTPGrabCore *, ForeachCollectionFunc, ForeachFileFunc, ForeachGroupFunc, void *) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_schedular_foreach_task", (void**) &func, NULL);
        g_return_if_fail(func != NULL);
        return func(glue->standalone_core, collection_func, file_func, group_func, data);
    } else {
        return nntpgrab_glue_schedular_foreach_task(glue, collection_func, file_func, group_func, data);
    }
}

static ngboolean
nntpgrab_glue_schedular_move_task(NNTPGrabGlue *obj, const char *collection_name_src, const char *subject_src, const char *collection_name_dest, int position_dest)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    char *msg;

    g_static_mutex_lock(&command_mutex);

    g_assert(collection_name_src);
    g_assert(collection_name_dest);
    g_assert(position_dest >= -1);

    // There is no assertion on the subject because the user can send a request
    // to move the given collection itself

    line = g_strdup_printf("SCHEDULAR_MOVE_TASK %s\t%s\t%s\t%i\n", collection_name_src, (subject_src ? subject_src : ""), collection_name_dest, position_dest);
    write_line(&glue->socket_command, line);
    g_free(line);

    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

ngboolean
nntpgrab_schedular_move_task(NNTPGrabGlue *obj, const char *collection_name_src, const char *subject_src, const char *collection_name_dest, int position_dest)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static ngboolean (*func) (NNTPGrabCore *, const char *, const char *, const char *, int) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_schedular_move_task", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, FALSE);
        return func(glue->standalone_core, collection_name_src, subject_src, collection_name_dest, position_dest);
    } else {
        return nntpgrab_glue_schedular_move_task(glue, collection_name_src, subject_src, collection_name_dest, position_dest);
    }
}

static ngboolean
nntpgrab_glue_schedular_move_collection(NNTPGrabGlue *obj, const char *collection_name, int new_position)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    char *line = NULL;
    char *msg;

    g_static_mutex_lock(&command_mutex);

    g_assert(collection_name);
    g_assert(new_position >= -1);

    line = g_strdup_printf("SCHEDULAR_MOVE_COLLECTION %s\t%i\n", collection_name, new_position);
    write_line(&glue->socket_command, line);
    g_free(line);

    if (!read_line(&glue->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

ngboolean
nntpgrab_schedular_move_collection(NNTPGrabGlue *obj, const char *collection_name, int new_position)
{
    NNTPGrabGlue *glue = NNTPGRAB_GLUE(obj);
    static ngboolean (*func) (NNTPGrabCore *, const char *, int) = NULL;

    if (glue->standalone_core) {
        bind_func("nntpgrab_core_schedular_move_collection", (void**) &func, NULL);
        g_return_val_if_fail(func != NULL, FALSE);
        return func(glue->standalone_core, collection_name, new_position);
    } else {
        return nntpgrab_glue_schedular_move_collection(glue, collection_name, new_position);
    }
}

ngboolean
nntpgrab_glue_get_is_standalone(NNTPGrabGlue *core)
{
    g_return_val_if_fail(glue != NULL, FALSE);

    return (glue->standalone_core != NULL);
}
