/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gmodule.h>
#include <glib-object.h>

#include "nntpgrab_internal.h"
#include "nntpgrab_utils.h"
#include "par2.h"
#include "plugin_par2.h"
#include "nntpgrab_plugin_par2.h"

static PluginPar2ExportedFuncs plugin_funcs;

static void emit_repair_message_callback (const char *message);
static void emit_load_progress_update (const char *filename, double progress);
static void emit_recovery_file_loaded (const char *filename, int num_new_packets, int num_blocks_found);
static void emit_file_loaded (const char *filename, PAR2FileState state, int num_blocks_found, int num_blocks_expected);
static void emit_repair_progress_update (double progress);
static void emit_repair_failure (int num_blocks_more_required);
static void emit_repair_success (void);
static void emit_no_repair_required (void);

typedef struct Par2Class Par2Class;

static GType par2_get_type (void);

static const char *active_collection_name = NULL;
static const char *active_par2_filename = NULL;

struct Par2
{
    GObject parent;

    GStaticRWLock rwlock;
};

struct Par2Class
{
    GObjectClass parent;
};

G_DEFINE_TYPE(Par2, par2, G_TYPE_OBJECT);

static void
par2_init (Par2 *obj)
{
    g_static_rw_lock_init(&obj->rwlock);
}

static void
par2_finalize (GObject *obj)
{
    Par2 *par2 = PAR2(obj);

    g_static_rw_lock_free(&par2->rwlock);
}

static void
par2_class_init (Par2Class *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = par2_finalize;
}

gboolean
par2_initialize(char **errmsg)
{
    GModule *plugin = NULL;
    PluginPar2ImportedFuncs imported_funcs;
    plugin_par2_init_func init_func;
    char *filename;

#ifdef WIN32
#define EXT "-0.dll"
#else
#define EXT ".so"
#endif

    if (g_getenv("NNTPGRAB_LIBDIR")) {
        filename = g_strdup_printf("%s/libnntpgrab_plugin_par2%s", g_getenv("NNTPGRAB_LIBDIR"), EXT);
    } else {
        filename = g_strdup_printf("%s/libnntpgrab_plugin_par2%s", PLUGIN_DIR, EXT);
    }

    plugin = g_module_open(filename, G_MODULE_BIND_LOCAL);

    g_free(filename);

    if (!plugin) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    imported_funcs.version = NNTPGRAB_PLUGIN_API_VERSION;
    imported_funcs.emit_repair_message = emit_repair_message_callback;
    imported_funcs.emit_load_progress_update = emit_load_progress_update;
    imported_funcs.emit_recovery_file_loaded = emit_recovery_file_loaded;
    imported_funcs.emit_file_loaded = emit_file_loaded;
    imported_funcs.emit_repair_progress_update = emit_repair_progress_update;
    imported_funcs.emit_repair_failure = emit_repair_failure;
    imported_funcs.emit_repair_success = emit_repair_success;
    imported_funcs.emit_no_repair_required = emit_no_repair_required;

    if (!g_module_symbol(plugin, "nntpgrab_plugin_par2_get_version", (gpointer *) &plugin_funcs.get_version)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (plugin_funcs.get_version() != NNTPGRAB_PLUGIN_API_VERSION) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("PAR2 Plugin API mismatch (Plugin API version = %i, expected = %i)"), plugin_funcs.get_version(), NNTPGRAB_PLUGIN_API_VERSION);
        }

        return FALSE;
    }

    /* All required methods are found, lets start */
    if (!g_module_symbol(plugin, "nntpgrab_plugin_par2_initialize", (gpointer *) &init_func)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (!g_module_symbol(plugin, "nntpgrab_plugin_par2_repair_files", (gpointer *) &plugin_funcs.repair_files)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (!init_func(imported_funcs)) {
        if (errmsg) {
            *errmsg = g_strdup(_("PAR2 Plugin initialization failed!"));
        }

        return FALSE;
    }

    nntpgrab_core_emit_debug_message(FALSE, _("PAR2 plugin initialized"));

    return TRUE;
}

Par2 *
par2_new(void)
{
    Par2 *par2;

    par2 = g_object_new(PAR2_TYPE_OBJECT, NULL);

    return par2;
}

void
par2_destroy(Par2 *obj)
{
    g_object_unref(obj);
}

gboolean
par2_repair_files(Par2 *obj, const char *collection_name, const char *directory, const char *par2filename)
{
    gboolean retval;

    g_assert(plugin_funcs.repair_files);

    active_collection_name = collection_name;
    active_par2_filename = par2filename;
    retval = plugin_funcs.repair_files(directory, par2filename);
    active_collection_name = NULL;
    active_par2_filename = NULL;

    return retval;
}

static void
emit_repair_message_callback(const char *message)
{
    char *msg_stripped = g_strdup(message);

    nntpgrab_utils_sanitize_text(msg_stripped, strlen(msg_stripped));

    nntpgrab_core_emit_par2_repair_message(FALSE, msg_stripped);

    g_free(msg_stripped);
}

static void emit_load_progress_update (const char *filename, double progress)
{
    nntpgrab_core_emit_par2_load_progress_update(FALSE, active_collection_name, active_par2_filename, filename, progress);
}

static void emit_recovery_file_loaded (const char *filename, int num_new_packets, int num_blocks_found)
{
    nntpgrab_core_emit_par2_recovery_file_loaded(FALSE, active_collection_name, active_par2_filename, filename, num_new_packets, num_blocks_found);
}

static void emit_file_loaded (const char *filename, PAR2FileState state, int num_blocks_found, int num_blocks_expected)
{
    nntpgrab_core_emit_par2_file_loaded(FALSE, active_collection_name, active_par2_filename, filename, state, num_blocks_found, num_blocks_expected);
}

static void emit_repair_progress_update (double progress)
{
    nntpgrab_core_emit_par2_repair_progress_update(FALSE, active_collection_name, active_par2_filename, progress);
}

static void emit_repair_failure (int num_blocks_more_required)
{
    nntpgrab_core_emit_par2_repair_failure(FALSE, active_collection_name, active_par2_filename, num_blocks_more_required);
}

static void emit_repair_success (void)
{
    nntpgrab_core_emit_par2_repair_success(FALSE, active_collection_name, active_par2_filename);
}

static void emit_no_repair_required (void)
{
    nntpgrab_core_emit_par2_no_repair_required(FALSE, active_collection_name, active_par2_filename);
}
