/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gmodule.h>
#include <glib-object.h>

#include "nntpgrab_internal.h"
#include "nntpgrab_utils.h"
#include "unpack.h"
#include "plugin_unpack.h"
#include "nntpgrab_plugin_unpack.h"

static PluginUnpackExportedFuncs plugin_funcs;

typedef struct UnpackClass UnpackClass;

static GType unpack_get_type (void);

static const char *active_collection_name = NULL;

static void unpack_progress_update (const char *filename, double progress);
static void unpack_message_received (const char *filename, const char *message);
static void unpack_working_archive_changed (const char *filename, const char *working_archive);

struct Unpack
{
    GObject parent;

    GStaticRWLock rwlock;
};

struct UnpackClass
{
    GObjectClass parent;
};

G_DEFINE_TYPE(Unpack, unpack, G_TYPE_OBJECT);

static void
unpack_init (Unpack *obj)
{
    g_static_rw_lock_init(&obj->rwlock);
}

static void
unpack_finalize (GObject *obj)
{
    Unpack *unpack = UNPACK(obj);

    g_static_rw_lock_free(&unpack->rwlock);
}

static void
unpack_class_init (UnpackClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = unpack_finalize;
}

gboolean
unpack_initialize(char **errmsg)
{
    GModule *plugin = NULL;
    PluginUnpackImportedFuncs imported_funcs;
    plugin_unpack_init_func init_func;
    char *filename;

#ifdef WIN32
#define EXT "-0.dll"
#else
#define EXT ".so"
#endif

    if (g_getenv("NNTPGRAB_LIBDIR")) {
        filename = g_strdup_printf("%s/libnntpgrab_plugin_unpack%s", g_getenv("NNTPGRAB_LIBDIR"), EXT);
    } else {
        filename = g_strdup_printf("%s/libnntpgrab_plugin_unpack%s", PLUGIN_DIR, EXT);
    }

    plugin = g_module_open(filename, G_MODULE_BIND_LOCAL);

    g_free(filename);

    if (!plugin) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    imported_funcs.version = NNTPGRAB_PLUGIN_API_VERSION;
    imported_funcs.unpack_progress_update = unpack_progress_update;
    imported_funcs.unpack_message_received = unpack_message_received;
    imported_funcs.unpack_working_archive_changed = unpack_working_archive_changed;

    if (!g_module_symbol(plugin, "nntpgrab_plugin_unpack_get_version", (gpointer *) &plugin_funcs.get_version)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (plugin_funcs.get_version() != NNTPGRAB_PLUGIN_API_VERSION) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Unpack Plugin API mismatch (Plugin API version = %i, expected = %i)"), plugin_funcs.get_version(), NNTPGRAB_PLUGIN_API_VERSION);
        }

        return FALSE;
    }

    /* All required methods are found, lets start */
    if (!g_module_symbol(plugin, "nntpgrab_plugin_unpack_initialize", (gpointer *) &init_func)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (!g_module_symbol(plugin, "nntpgrab_plugin_unpack_do_unpack", (gpointer *) &plugin_funcs.do_unpack)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (!init_func(imported_funcs)) {
        if (errmsg) {
            *errmsg = g_strdup(_("Unpack Plugin initialization failed!"));
        }

        return FALSE;
    }

    nntpgrab_core_emit_debug_message(FALSE, _("Unpack plugin initialized"));

    return TRUE;
}

Unpack *
unpack_new(void)
{
    Unpack *unpack;

    unpack = g_object_new(UNPACK_TYPE_OBJECT, NULL);

    return unpack;
}

void
unpack_destroy(Unpack *obj)
{
    g_object_unref(obj);
}

gboolean
unpack_do_unpack(Unpack *obj, const char *collection_name, const char *filename, const char *target_directory)
{
    gboolean retval;
    int saved_errno = 0;
    char *errmsg = NULL;

    g_assert(plugin_funcs.do_unpack);

    active_collection_name = collection_name;
    retval = plugin_funcs.do_unpack(filename, target_directory, &saved_errno, &errmsg);
    active_collection_name = NULL;

    return retval;
}

static void
unpack_progress_update (const char *filename, double progress)
{
    nntpgrab_core_emit_unpack_progress_update(FALSE, active_collection_name, filename, progress);
}

static void
unpack_message_received (const char *filename, const char *message)
{
    nntpgrab_core_emit_unpack_message_received(FALSE, active_collection_name, filename, message);
}

static void
unpack_working_archive_changed (const char *filename, const char *working_archive)
{
    nntpgrab_core_emit_unpack_working_archive_changed(FALSE, active_collection_name, filename, working_archive);
}
