/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <glib/gstdio.h>
#include <glib/gi18n.h>

#include "nntpgrab_plugin_decoder.h"
#include "yydecode.h"

// yydecode.c
int decode(const char *inname, const char *forced_outname, struct decoded_file **decoded_list, int *saved_errno, char **errmsg);

static PluginDecoderImportedFuncs imported_funcs;

gboolean
nntpgrab_plugin_decoder_initialize(PluginDecoderImportedFuncs funcs)
{
    imported_funcs = funcs;

    return TRUE;
}

int
nntpgrab_plugin_decoder_get_version(void)
{
    return NNTPGRAB_PLUGIN_API_VERSION;
}

#if 0
static void
add_line_to_message(char **msg, const char *line)
{
    char *tmp;

    if (*msg && strlen(*msg) > 0) {
        tmp = g_strdup_printf("%s\n%s", *msg, line);
        g_free(*msg);
    } else {
        tmp = g_strdup(line);
    }

    *msg = tmp;
}
#endif

NNTPGrabDecoderRes
nntpgrab_plugin_decoder_decode_file(const char *collection_name, NNTPFile *file, const char *temp_directory, const char *target_directory, int *saved_errno, char **errmsg)
{
    int i;
    char *dir;
#if 0
    GList *lines = NULL;
#endif
    NNTPGrabDecoderRes ret = DECODER_RESULT_COMPLETE;
    gboolean file_found = FALSE;
    struct decoded_file *decoded_list = NULL;
    struct decoded_file *list;

    g_return_val_if_fail(collection_name != NULL, DECODER_RESULT_ERROR);
    g_return_val_if_fail(file != NULL, DECODER_RESULT_ERROR);
    g_return_val_if_fail(temp_directory != NULL, DECODER_RESULT_ERROR);
    g_return_val_if_fail(target_directory != NULL, DECODER_RESULT_ERROR);
    g_return_val_if_fail(saved_errno != NULL, DECODER_RESULT_ERROR);

    *saved_errno = 0;

    dir = g_strdup_printf("%s%s%s", target_directory, G_DIR_SEPARATOR_S, collection_name);
    if (g_mkdir_with_parents(dir, 0755) == -1) {
        g_print(_("Creation of folder '%s' failed, errno = %i, %s\n"), dir, errno, strerror(errno));
    }
    g_free(dir);

    // The yydecode code assumes all the resulting files will be saved in the current directory
    // so we need to change to the wanted directory before starting the decode
    dir = g_strdup_printf("%s%s%s%s", target_directory, G_DIR_SEPARATOR_S, collection_name, G_DIR_SEPARATOR_S);
    if (g_chdir(dir) == -1) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Unable to navigate to folder '%s'"), dir);
        }
        *saved_errno = ENOENT;
        g_free(dir);
        return DECODER_RESULT_ERROR;
    }
    g_free(dir);

    // Load all the individual parts
    for (i = 0; i < file->numparts; i++) {
        struct decoded_file *list;
        char *filename = g_strdup_printf("%s%s%s.%i", temp_directory, G_DIR_SEPARATOR_S, file->tmp_filename, i + 1);

        if (!g_file_test(filename, G_FILE_TEST_IS_REGULAR | G_FILE_TEST_EXISTS)) {
            g_free(filename);
            continue;
        }

        file_found = TRUE;

        if (decode(filename, NULL, &decoded_list, saved_errno, errmsg) != EXIT_SUCCESS) {
            g_print("Error during decode\n");
            if (errmsg && *errmsg) {
                g_print("errmsg = %s\n", *errmsg);
            }

            ret = DECODER_RESULT_ERROR;

            while (decoded_list) {
                list = decoded_list->next;

                if (decoded_list->handle) {
                    fclose(decoded_list->handle);
                    decoded_list->handle = NULL;
                }

                if (decoded_list->filename) free(decoded_list->filename);
                if (decoded_list->outname)  free(decoded_list->outname);
                if (decoded_list->status)   free(decoded_list->status);
                free(decoded_list);
                decoded_list = list;
            }
        }

        g_free(filename);

        if (ret == DECODER_RESULT_ERROR) {
            break;
        }
    }

    if (decoded_list) {
        strncpy((char*) file->real_filename, decoded_list->outname, sizeof(file->real_filename) - 1);
        for(i = 0; i < decoded_list->total_parts; i++) {
            switch(decoded_list->status[i]) {
                case part_missing:
                case part_broken:
                    ret = DECODER_RESULT_INCOMPLETE;
                    break;

                case part_intact:
                case part_duplicated:
                    // don't do anything
                    break;
            }
        }

        while (decoded_list) {
            list = decoded_list->next;

            if (decoded_list->handle) {
                fclose(decoded_list->handle);
                decoded_list->handle = NULL;
            }

            if (decoded_list->filename) free(decoded_list->filename);
            if (decoded_list->outname)  free(decoded_list->outname);
            if (decoded_list->status)   free(decoded_list->status);
            free(decoded_list);
            decoded_list = list;
        }
    }

    if (!file_found) {
        if (errmsg) {
            *errmsg = g_strdup(_("No data to decode"));
        }

        if (saved_errno) {
            *saved_errno = -1;
        }

        ret = DECODER_RESULT_NO_PARTS_AVAIL;
    }

#if 0
    if (lines && errmsg) {
        GList *list;
        char *line;

        *errmsg = NULL;

        if (g_list_length(lines) > 4) {
            // Take only the first two and the last two messages
            list = g_list_first(lines);
            line = list->data;
            add_line_to_message(errmsg, line);

            list = g_list_next(list);
            line = list->data;
            add_line_to_message(errmsg, line);

            list = g_list_last(lines);
            list = g_list_previous(list);
            line = list->data;
            add_line_to_message(errmsg, line);

            list = g_list_next(list);
            line = list->data;
            add_line_to_message(errmsg, line);

            add_line_to_message(errmsg, _("(not all output is shown)"));
        } else {
            list = lines;
            while (list) {
                line = list->data;
                add_line_to_message(errmsg, line);

                list = g_list_next(list);
            }
        }
    }
#endif

    if (ret != DECODER_RESULT_ERROR) {
        // Remove all the temp parts on success
        for (i = 0; i < file->numparts; i++) {
            char *filename = g_strdup_printf("%s%s%s.%i", temp_directory, G_DIR_SEPARATOR_S, file->tmp_filename, i + 1);

            g_unlink(filename);

            g_free(filename);
        }
    }

#if 0
    if (lines) {
        GList *list = lines;
        while (list) {
            g_free(list->data);
            list = g_list_next(list);
        }
        g_list_free(lines);
    }
#endif

    // Change the current directory back to another directory to prevent the NNTPGrab process
    // from 'claiming' the directory in which the decoded files were saved
    g_chdir(target_directory);

    return ret;
}
