/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <glib-object.h>
#include <glib/gi18n.h>

#include "nntpgrab_plugin_schedular.h"
#include "schedular_plugin.h"

void
decoder_thread_func(gpointer data, gpointer user_data)
{
    DecoderData *decoder_data = (DecoderData *) data;
    int saved_errno = 0;
    static time_t last_timestamp = 0;
    PAR2VerifyData *verify_data;
    NNTPGrabDecoderRes res;
    char *errmsg = NULL;

    // Do we still need to decode?
    g_static_mutex_lock(&decoder_data->schedular->mutex);
    g_static_mutex_lock(&decoder_data->collection->mutex);
    if (decoder_data->schedular->abort_flag || decoder_data->file->stop_flag) {
        decoder_data->file->now_decoding = FALSE;
        g_static_mutex_unlock(&decoder_data->schedular->mutex);
        g_static_mutex_unlock(&decoder_data->collection->mutex);

        imported_funcs_file_unref(decoder_data->file);
        imported_funcs_collection_unref(decoder_data->collection);

        g_slice_free(DecoderData, decoder_data);

        return;
    }
    g_static_mutex_unlock(&decoder_data->schedular->mutex);
    g_static_mutex_unlock(&decoder_data->collection->mutex);

    imported_funcs_task_state_changed(decoder_data->collection->collection_name, decoder_data->file->subject, NULL, decoder_data->file->status, TASK_STATE_DECODING);

    g_static_mutex_lock(&decoder_data->collection->mutex);
    decoder_data->file->status = TASK_STATE_DECODING;
    g_static_mutex_unlock(&decoder_data->collection->mutex);

    // Perform the actual decoding
    if ((res = imported_funcs_decode_file(decoder_data->collection->collection_name, decoder_data->file, &saved_errno, &errmsg)) == DECODER_RESULT_ERROR) {
        if (saved_errno == ENOSPC) {
            // No free space available, stop the schedular
            if (errmsg) {
                stop_schedular_from_seperate_thread(errmsg);
                g_free(errmsg);
            } else {
                stop_schedular_from_seperate_thread(strerror(saved_errno));
            }

            // Sleep for 1 second to prevent other threads from retrying this file
            g_usleep(G_USEC_PER_SEC * 1);

            g_static_mutex_lock(&decoder_data->collection->mutex);
            decoder_data->file->now_decoding = FALSE;
            imported_funcs_task_state_changed(decoder_data->collection->collection_name, decoder_data->file->subject, NULL, decoder_data->file->status, TASK_STATE_WAITING_FOR_DECODE);
            decoder_data->file->status = TASK_STATE_WAITING_FOR_DECODE;
            g_static_mutex_unlock(&decoder_data->collection->mutex);

            imported_funcs_file_unref(decoder_data->file);
            imported_funcs_collection_unref(decoder_data->collection);

            g_slice_free(DecoderData, decoder_data);

            return;
        } else {
            // Unknown decode error occured, stop the schedular
            char *msg;

            if (errmsg) {
                msg = g_strdup_printf(errmsg);
                g_free(errmsg);
            } else {
                msg = g_strdup_printf(_("Unknown decode error occured, %s (%i)"), strerror(saved_errno), saved_errno);
            }
            stop_schedular_from_seperate_thread(msg);
            g_free(msg);

            // Sleep for 1 second to prevent other threads from retrying this file
            g_usleep(G_USEC_PER_SEC * 1);

            g_static_mutex_lock(&decoder_data->collection->mutex);
            decoder_data->file->now_decoding = FALSE;
            imported_funcs_task_state_changed(decoder_data->collection->collection_name, decoder_data->file->subject, NULL, decoder_data->file->status, TASK_STATE_WAITING_FOR_DECODE);
            decoder_data->file->status = TASK_STATE_WAITING_FOR_DECODE;
            g_static_mutex_unlock(&decoder_data->collection->mutex);

            imported_funcs_file_unref(decoder_data->file);
            imported_funcs_collection_unref(decoder_data->collection);

            g_slice_free(DecoderData, decoder_data);

            return;
        }
    }

    // Mark the file as completed
    g_static_mutex_lock(&decoder_data->collection->mutex);

    if (!decoder_data->file->file_is_new) {
        decoder_data->file->file_is_changed = TRUE;
    }

    // Was the task restarted while the decoding was busy?
    if (decoder_data->file->now_decoding) {
        // Nope, mark the file decode as completed
        TaskState state;

        decoder_data->file->file_is_downloaded = TRUE;
        decoder_data->file->now_decoding = FALSE;

        switch(res) {
            case DECODER_RESULT_INCOMPLETE:
                state = TASK_STATE_FINISHED_INCOMPLETE;
                break;

            case DECODER_RESULT_NO_PARTS_AVAIL:
                state = TASK_STATE_FINISHED_NO_PARTS_AVAIL;
                break;

            case DECODER_RESULT_ERROR:
                state = TASK_STATE_WAITING_FOR_DECODE;
                break;

            default:
            case DECODER_RESULT_COMPLETE:
                state = TASK_STATE_FINISHED_COMPLETE;
                break;
        }

        imported_funcs_task_state_changed(decoder_data->collection->collection_name, decoder_data->file->subject, decoder_data->file->real_filename, decoder_data->file->status, state);
        decoder_data->file->status = state;
    }

    g_static_mutex_unlock(&decoder_data->collection->mutex);

    // Verify this file using PAR2
    verify_data = g_slice_new0(PAR2VerifyData);
    verify_data->file = decoder_data->file;
    verify_data->collection = decoder_data->collection;
    verify_data->poolPar2verify = decoder_data->poolPar2verify;
    verify_data->poolPar2repair = decoder_data->poolPar2repair;
    verify_data->poolUnpack = decoder_data->poolUnpack;

    g_thread_pool_push(decoder_data->poolPar2verify, verify_data, NULL);

    // Keep the reference count intact here (actually we should do both an unref and ref here)

    g_slice_free(DecoderData, decoder_data);

    // Save the download queue only when there has been some time between now and the previous time we've decoded something
    if (time(NULL) > last_timestamp + 60) {
        schedular_plugin_save_queue(NULL);
        last_timestamp = time(NULL);
    }
}

