/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <glib-object.h>
#include <glib/gi18n.h>
#ifdef WIN32
#include <winsock2.h>
#endif

#include "nntpgrab_plugin_schedular.h"
#include "schedular_plugin.h"

void
download_thread_func(gpointer data, gpointer user_data)
{
    DownloadData *download_data = (DownloadData *) data;
    SchedularPlugin *schedular = SCHEDULAR_PLUGIN(user_data);
    NNTPPart *part;
    NNTPFile *file;
    NNTPCollection *collection;
    NNTPGrabErrCode err;
    gboolean abort_flag = FALSE;
    gpointer nntp_connection = NULL;
    gint idle_timer = 0;
    char *reason;
    int i;

    do {
        g_static_mutex_lock(&schedular->mutex);
        abort_flag = schedular->abort_flag;
        g_static_mutex_unlock(&schedular->mutex);

        if (abort_flag) {
            break;
        }

        // Have we got anything to download ?
        if (!(get_next_part_to_download(schedular, download_data->server_id, &collection, &file, &part, download_data->poolDecoder))) {
            idle_timer++;
            if (idle_timer == 10) {
                // We have waited for 10 seconds with no new tasks. Kill the connection
                if (nntp_connection) {
                    imported_funcs_nntp_connection_destroy(nntp_connection);
                    nntp_connection = NULL;
                }
            }

            g_usleep(G_USEC_PER_SEC * 1);
            continue;
        }

        idle_timer = 0;

        // Perform the download
        if (!nntp_connection) {
            err = NNTP_ERROR_NONE;
            nntp_connection = imported_funcs_nntp_connection_initialize(download_data->server->servername, &err);
        }

        g_static_mutex_lock(&schedular->mutex);
        abort_flag = schedular->abort_flag;
        g_static_mutex_unlock(&schedular->mutex);

        if (abort_flag) {
            // Push the file back in the download queue
            update_part_download_status(collection, file, part, download_data->server_id, FALSE, FALSE, FALSE, FALSE, NULL);

            imported_funcs_file_unref(file);
            imported_funcs_collection_unref(collection);

            break;
        }

        if (!nntp_connection) {
            if (err == NNTP_ERROR_HOST_NOT_FOUND) {
                // Configuration error detected, kill this thread
                update_part_download_status(collection, file, part, download_data->server_id, FALSE, FALSE, FALSE, FALSE, NULL);
                reason = g_strdup_printf(_("Unable to detect the IP address belonging to '%s' (servername = %s)"), download_data->server->hostname, download_data->server->servername);
                stop_schedular_from_seperate_thread(reason);
                g_free(reason);

                imported_funcs_file_unref(file);
                imported_funcs_collection_unref(collection);

                break;
            }

            // Unable to connect, wait 30 seconds, push the part back in the queue and start the loop again
            update_part_download_status(collection, file, part, download_data->server_id, FALSE, FALSE, FALSE, FALSE, NULL);

            imported_funcs_file_unref(file);
            imported_funcs_collection_unref(collection);

            for (i = 0; i < 30; i++) {
                g_usleep(G_USEC_PER_SEC);

                g_static_mutex_lock(&schedular->mutex);
                abort_flag = schedular->abort_flag;
                g_static_mutex_unlock(&schedular->mutex);

                if (abort_flag) {
                    break;
                }
            }

            continue;
        }

        // Download the part.
        err = imported_funcs_nntp_connection_get_part(nntp_connection, collection->collection_name, file, part, &reason);
        switch (err) {
            case NNTP_ERROR_NONE:
                update_part_download_status(collection, file, part, download_data->server_id, TRUE, FALSE, FALSE, TRUE, nntp_connection);

                break;

            case NNTP_ERROR_PART_NOT_AVAILABLE:
                update_part_download_status(collection, file, part, download_data->server_id, FALSE, FALSE, FALSE, TRUE, nntp_connection);

                break;

            case NNTP_ERROR_UNABLE_TO_SAVE_PART:
                update_part_download_status(collection, file, part, download_data->server_id, FALSE, FALSE, FALSE, FALSE, nntp_connection);

                stop_schedular_from_seperate_thread(reason);

                if (reason) {
                    g_free(reason);
                }

                // Prevent to start a new download task
                g_usleep(G_USEC_PER_SEC / 10);

                break;

            default:
                // Connection error. Put the part back in the queue and
                // sleep for a few seconds to avoid flooding

                update_part_download_status(collection, file, part, download_data->server_id, FALSE, FALSE, FALSE, FALSE, nntp_connection);

                // Kill the connection
                imported_funcs_nntp_connection_destroy(nntp_connection);
                nntp_connection = NULL;

                for (i = 0; i < 5; i++) {
                    g_usleep(G_USEC_PER_SEC);

                    g_static_mutex_lock(&schedular->mutex);
                    abort_flag = schedular->abort_flag;
                    g_static_mutex_unlock(&schedular->mutex);

                    if (abort_flag) {
                        break;
                    }
                }

                break;
        };

        imported_funcs_file_unref(file);
        imported_funcs_collection_unref(collection);

        g_static_mutex_lock(&schedular->mutex);
        abort_flag = schedular->abort_flag;
        g_static_mutex_unlock(&schedular->mutex);
    } while (!abort_flag);

    if (nntp_connection) {
        imported_funcs_nntp_connection_destroy(nntp_connection);
    }

    g_slice_free(DownloadData, download_data);
}
