/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include <glib-object.h>
#include "nntpgrab_plugin_schedular.h"
#include "schedular_plugin.h"

typedef struct _par2_file_seen {
    char filename_without_extension[256];
    char filename[256];
} PAR2FileSeen;

void
par2repair_thread_func(gpointer data, gpointer user_data)
{
    PAR2RepairData *par2_data = (PAR2RepairData *) data;
    SchedularPlugin *schedular = SCHEDULAR_PLUGIN(user_data);
    GList *list;
    GList *list2;
    GList *files = NULL;
    ConfigOpts opts;
    char *collection_name;

    // Do we still need to repair?
    opts = imported_funcs_config_get_opts();
    g_static_mutex_lock(&schedular->mutex);
    if (schedular->abort_flag || !opts.enable_par2_repair) {
        if (!schedular->abort_flag) {
            if (opts.enable_auto_unpack) {
                UnpackData *unpack_data = g_slice_new0(UnpackData);
                unpack_data->collection = par2_data->collection;
                unpack_data->collection->now_unpacking = TRUE;
                unpack_data->poolPar2verify = par2_data->poolPar2verify;
                unpack_data->poolPar2repair = par2_data->poolPar2repair;
                unpack_data->poolUnpack = par2_data->poolUnpack;
                g_thread_pool_push(par2_data->poolUnpack, unpack_data, NULL);

                imported_funcs_collection_ref(par2_data->collection);
            } else {
                test_if_all_downloads_are_completed(schedular, par2_data->poolPar2verify, par2_data->poolPar2repair, par2_data->poolUnpack);
            }
        }

        g_static_mutex_unlock(&schedular->mutex);

        par2_data->collection->now_repairing = FALSE;
        imported_funcs_collection_unref(par2_data->collection);

        g_slice_free(PAR2RepairData, par2_data);
        return;
    }

    g_static_mutex_unlock(&schedular->mutex);

    g_static_mutex_lock(&par2_data->collection->mutex);

    request_par2_and_unpack_lock();

    collection_name = g_strdup(par2_data->collection->collection_name);

    // Find out all the unique PAR2 sets in this collection
    list = par2_data->collection->files;
    while (list) {
        char filename_without_extension[256];
        gboolean filename_seen = FALSE;
        char *ptr;
        NNTPFile *file = (NNTPFile*) list->data;
        char *path;

        if (!(file->file_type == NNTP_FILE_TYPE_PAR2 || file->file_type == NNTP_FILE_TYPE_PAR)) {
            list = g_list_next(list);
            continue;
        }

        // Does the .par2 file exist?
        path = g_build_filename(opts.download_directory, collection_name, file->real_filename, NULL);
        if (!g_file_test(path, G_FILE_TEST_EXISTS)) {
            g_free(path);
            list = g_list_next(list);
            continue;
        }
        g_free(path);

        memset(filename_without_extension, 0, sizeof(filename_without_extension));
        ptr = g_strrstr(file->real_filename, ".");
        if (!ptr) {
            // Huh? file contains no extension, ignore
            list = g_list_next(list);
            continue;
        }

        strncpy(filename_without_extension, file->real_filename, ptr - file->real_filename);

        // Strip any .volXX+YY part
        ptr = g_strrstr(filename_without_extension, ".");
        if (ptr && !g_strncasecmp(ptr, ".vol", 4)) {
            *ptr = '\0';
        }

        list2 = files;
        while (list2) {
            PAR2FileSeen *file = (PAR2FileSeen*) list2->data;

            if (!strcmp(filename_without_extension, file->filename_without_extension)) {
                // We've already seen this PAR2 set, ignore
                filename_seen = TRUE;
                break;
            }

            list2 = g_list_next(list2);
        }

        if (!filename_seen) {
            PAR2FileSeen *file_seen = g_slice_new0(PAR2FileSeen);

            strncpy(file_seen->filename_without_extension, filename_without_extension, sizeof(file_seen->filename_without_extension) - 1);
            strncpy(file_seen->filename, file->real_filename, sizeof(file_seen->filename) - 1);

            files = g_list_append(files, file_seen);
        }

        list = g_list_next(list);
    }

    g_static_mutex_unlock(&par2_data->collection->mutex);

    // Perform the repair
    list = files;
    while (list) {
        PAR2FileSeen *file_seen = (PAR2FileSeen*) list->data;

        imported_funcs_perform_par2_repair(collection_name, file_seen->filename);

        g_slice_free(PAR2FileSeen, file_seen);
        list = g_list_next(list);
    }
    g_list_free(files);

    g_free(collection_name);

    g_static_mutex_lock(&par2_data->collection->mutex);
    par2_data->collection->now_repairing = FALSE;
    g_static_mutex_unlock(&par2_data->collection->mutex);

    release_par2_and_unpack_lock();

    if (opts.enable_auto_unpack) {
        // All files repaired, unpack all the files
        UnpackData *unpack_data = g_slice_new0(UnpackData);
        unpack_data->collection = par2_data->collection;
        unpack_data->collection->now_unpacking = TRUE;
        unpack_data->poolPar2verify = par2_data->poolPar2verify;
        unpack_data->poolPar2repair = par2_data->poolPar2repair;
        unpack_data->poolUnpack = par2_data->poolUnpack;
        g_thread_pool_push(par2_data->poolUnpack, unpack_data, NULL);

        // Keep the reference count intact here for the collection (actually we should do both an unref and ref here)
    } else {
        test_if_all_downloads_are_completed(schedular, par2_data->poolPar2verify, par2_data->poolPar2repair, par2_data->poolUnpack);
        imported_funcs_collection_unref(par2_data->collection);
    }

    g_slice_free(PAR2RepairData, par2_data);
}
