/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef _MSC_VER
#include "config.h.win32"
#else
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <sqlite3.h>

#include "nntpgrab_plugin_schedular.h"
#include "nntpgrab_types.h"
#include "schedular_plugin.h"

#ifdef _MSC_VER
#undef atoll
#define atoll _atoi64
#endif

#if SIZEOF_LONG == 8
#define GUINT64_FORMAT "li"
#else
#define GUINT64_FORMAT "lli"
#endif

static char *
generate_filename(void)
{
    return g_build_filename(g_get_user_config_dir(), "NNTPGrab", "download_queue.sqlite", NULL);
}

sqlite3 *
db_open(char **errmsg)
{
    sqlite3 *db = NULL;
    char *filename;
    char *dirname;
    int ret;

    // Generate the folder ~/.config/NNTPGrab in case it didn't exist yet
    filename = generate_filename();
    dirname = g_path_get_dirname(filename);
    g_mkdir_with_parents(dirname, 0700);
    g_free(dirname);

    ret = sqlite3_open(filename, &db);
    if (ret > 0) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Unable to open database: %s"), sqlite3_errmsg(db));
        } else {
            g_warning(_("Unable to open database: %s"), sqlite3_errmsg(db));
        }

        g_free(filename);
        sqlite3_close(db);
        return NULL;
    }

    g_free(filename);

    return db;
}

void
db_close(sqlite3 *db)
{
    if (db) {
        sqlite3_close(db);
    }
}

static gboolean
execute_query(sqlite3 *db, const char *query, char **errmsg)
{
    int ret;
    char *err = NULL;

    g_return_val_if_fail(db != NULL, FALSE);
    g_return_val_if_fail(query != NULL, FALSE);

//g_print("Now executing query: %s\n", query);
    ret = sqlite3_exec(db, query, NULL, NULL, &err);
    if (ret != SQLITE_OK) {
        g_return_val_if_fail(err != NULL, FALSE);

        if (errmsg) {
            *errmsg = g_strdup_printf(_("Error occured while performing query '%s':\n%s"), query, err);
            g_print(_("Error occured while performing query '%s':\n%s"), query, err);
        } else {
            g_warning(_("Error occured while performing query '%s':\n%s"), query, err);
        }

        sqlite3_free(err);

        return FALSE;
    } else {
        return TRUE;
    }
}

gboolean
db_begin_transaction(sqlite3 *db, char **errmsg)
{
    return execute_query(db, "BEGIN", errmsg);
}

gboolean
db_end_transaction(sqlite3 *db, char **errmsg)
{
    return execute_query(db, "END", errmsg);
}

static int
import_part(void *data, int num_fields, char **values, char **fieldnames)
{
    NNTPPart *part = g_slice_new0(NNTPPart);
    NNTPFile *file = (NNTPFile*) data;

    g_return_val_if_fail(num_fields == 4, -1);

    file->parts = g_list_append(file->parts, part);

    part->partnum = atoi(values[0]);
    strncpy((char*) part->message_id, values[1], sizeof(part->message_id) - 1);
    part->size = atoi(values[2]);
    part->downloaded = atoi(values[3]);

    file->numparts++;
    file->file_size += part->size;
    if (part->downloaded) {
        file->num_parts_downloaded++;
    } else {
        file->file_size_remaining += part->size;
    }

    return 0;
}

static int
import_group(void *data, int num_fields, char **values, char **fieldnames)
{
    NNTPFile *file = (NNTPFile*) data;

    g_return_val_if_fail(num_fields == 1, -1);

    file->groups = g_list_append(file->groups, g_strdup(values[0]));

    return 0;
}

struct _import_file_data {
    sqlite3 *db;
    NNTPCollection *collection;
};

static int
import_file(void *data, int num_fields, char **values, char **fieldnames)
{
    struct _import_file_data *import_data = (struct _import_file_data*) data;
    NNTPFile *file = imported_funcs_file_new();
    char *query;
    int ret;
    char *err = NULL;
    GList *parts;

    g_return_val_if_fail(num_fields == 11, -1);

    import_data->collection->files = g_list_append(import_data->collection->files, file);

    file->file_id = atoll(values[0]);
    strncpy((char*) file->subject, values[1], sizeof(file->subject) - 1);
    strncpy((char*) file->poster, values[2], sizeof(file->poster) - 1);
    file->stamp = (time_t) atoll(values[3]);
    strncpy((char*) file->tmp_filename, values[4], sizeof(file->real_filename) - 1);
    strncpy((char*) file->real_filename, values[5], sizeof(file->real_filename) - 1);
    file->status = atoi(values[6]);
    file->file_type = atoi(values[7]);
    file->par2_start_num = atoi(values[8]);
    file->par2_end_num = atoi(values[9]);
    file->par2_num_blocks = atoi(values[10]);
    file->position = g_list_length(import_data->collection->files);

    query = sqlite3_mprintf("SELECT partnum, message_id, size, downloaded "
                            "FROM parts "
                            "WHERE file_id = %"GUINT64_FORMAT" "
                            "ORDER BY partnum", file->file_id);
    ret = sqlite3_exec(import_data->db, query, import_part, file, &err);
    if (ret != SQLITE_OK) {
        g_return_val_if_fail(err != NULL, FALSE);

        g_warning(_("Error occured while performing query '%s':\n%s"), query, err);

        sqlite3_free(err);
        sqlite3_free(query);
        return -1;
    }
    sqlite3_free(query);

    query = sqlite3_mprintf("SELECT groupname FROM groups "
                            "WHERE file_id = %"GUINT64_FORMAT, file->file_id);
    ret = sqlite3_exec(import_data->db, query, import_group, file, &err);
    if (ret != SQLITE_OK) {
        g_return_val_if_fail(err != NULL, FALSE);

        g_warning(_("Error occured while performing query '%s':\n%s"), query, err);

        sqlite3_free(err);
        sqlite3_free(query);
        return -1;
    }
    sqlite3_free(query);

    import_data->collection->total_size += file->file_size;
    import_data->collection->total_size_remaining += file->file_size_remaining;

    parts = file->parts;
    file->file_is_downloaded = TRUE;
    while (parts) {
        NNTPPart *part = parts->data;

        if (!part->downloaded) {
            file->file_is_downloaded = FALSE;
        }

        parts = g_list_next(parts);
    }

    switch (file->status) {
        case TASK_STATE_DOWNLOADING:
            file->file_is_downloading = TRUE;
            break;

        case TASK_STATE_WAITING_FOR_DECODE:
        case TASK_STATE_DECODING:
            file->file_is_downloaded = FALSE;
            file->file_is_downloading = TRUE;    /* Needed to requeue the decoding */
            break;

        case TASK_STATE_WAITING_FOR_DOWNLOAD:
        case TASK_STATE_FINISHED_COMPLETE:
        case TASK_STATE_FINISHED_INCOMPLETE:
        case TASK_STATE_FINISHED_NO_PARTS_AVAIL:
        case TASK_STATE_SKIPPED:
            break;
    }

    return 0;
}

struct _import_collection_data {
    sqlite3 *db;
    GList *queue_ret;
};

static int
import_collection(void *data, int num_fields, char **values, char **fieldnames)
{
    NNTPCollection *collection = imported_funcs_collection_new();
    struct _import_collection_data *import_data = (struct _import_collection_data*) data;
    struct _import_file_data *file_data;
    char *query;
    int ret;
    char *err = NULL;

    g_return_val_if_fail(num_fields == 4, -1);

    collection->collection_id = atoll(values[0]);
    strncpy((char*) collection->collection_name, values[1], sizeof(collection->collection_name) - 1);
    strncpy((char*) collection->poster, values[2], sizeof(collection->poster) - 1);
    strncpy((char*) collection->groups, values[3], sizeof(collection->groups) - 1);

    import_data->queue_ret = g_list_append(import_data->queue_ret, collection);
    collection->position = g_list_length(import_data->queue_ret);

    query = sqlite3_mprintf("SELECT file_id, subject, poster, "
                            "stamp, tmp_filename, real_filename, status, file_type, "
                            "par2_start_num, par2_end_num, par2_num_blocks "
                            "FROM files "
                            "WHERE collection_id = %"GUINT64_FORMAT" "
                            "ORDER BY position", collection->collection_id);
    file_data = g_slice_new0(struct _import_file_data);
    file_data->db = import_data->db;
    file_data->collection = collection;
    ret = sqlite3_exec(import_data->db, query, import_file, file_data, &err);
    g_slice_free(struct _import_file_data, file_data);

    if (ret != SQLITE_OK) {
        g_return_val_if_fail(err != NULL, FALSE);

        g_warning(_("Error occured while performing query '%s':\n%s"), query, err);

        sqlite3_free(err);
        sqlite3_free(query);
        return -1;
    }

    sqlite3_free(query);

    return 0;
}

gboolean
load_download_queue(GList **queue_ret, char **errmsg)
{
    char *filename;
    gboolean retval = TRUE;
    int ret;
    char *err = NULL;
    const char *query;
    sqlite3 *db;
    struct _import_collection_data *import_data;

    filename = generate_filename();
    if (!g_file_test(filename, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR)) {
        // File not found, looks like this is the first time NNTPGrab is ever started, ignore
        if (errmsg) {
            *errmsg = g_strdup_printf(_("No database found at %s"), filename);
        } else {
            g_warning(_("No database found at %s"), filename);
        }

        g_free(filename);

        return TRUE;
    }

    g_free(filename);

    if (!(db = db_open(errmsg))) {
        retval = FALSE;
        goto bailout;
    }

    query = "SELECT collection_id, collection_name, poster, groups FROM collections ORDER BY position";
    import_data = g_slice_new0(struct _import_collection_data);
    import_data->db = db;
    import_data->queue_ret = NULL;
    ret = sqlite3_exec(db, query, import_collection, import_data, &err);
    if (ret != SQLITE_OK) {
        g_return_val_if_fail(err != NULL, FALSE);

        if (errmsg) {
            *errmsg = g_strdup_printf(_("Error occured while performing query '%s':\n%s"), query, err);
        } else {
            g_warning(_("Error occured while performing query '%s':\n%s"), query, err);
        }

        sqlite3_free(err);
        retval = FALSE;
    } else {
        *queue_ret = import_data->queue_ret;
    }

    g_slice_free(struct _import_collection_data, import_data);

bailout:
    db_close(db);
    return retval;
}

static gboolean
insert_part(sqlite3 *db, NNTPFile *file, NNTPPart *part, char **errmsg)
{
    char *err = NULL;
    char *query;

    query = sqlite3_mprintf("INSERT INTO parts (file_id, partnum, message_id, size, downloaded) VALUES "
                            " (%" GUINT64_FORMAT ", %i, '%q', %i, %i)",
                            file->file_id, part->partnum, part->message_id, part->size, part->downloaded);
    if (!execute_query(db, query, &err)) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Unable to insert part: %s"), err);
        } else {
            g_warning(_("Unable to insert part: %s"), err);
        }
        g_free(err);
        sqlite3_free(query);

        return FALSE;
    }

    g_free(err);
    sqlite3_free(query);

    return TRUE;
}

static gboolean
update_part(sqlite3 *db, NNTPFile *file, NNTPPart *part, char **errmsg)
{
    char *err = NULL;
    char *query;

    query = sqlite3_mprintf("UPDATE parts SET downloaded=%i "
                            "WHERE file_id=%"GUINT64_FORMAT" "
                            "AND partnum=%i",
                            part->downloaded, file->file_id, part->partnum);
    if (!execute_query(db, query, &err)) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Unable to update part: %s"), err);
        } else {
            g_warning(_("Unable to update part: %s"), err);
        }
        g_free(err);
        sqlite3_free(query);

        return FALSE;
    }

    g_free(err);
    sqlite3_free(query);

    return TRUE;
}

static gboolean
insert_group(sqlite3 *db, NNTPFile *file, const char *group, char **errmsg)
{
    char *err = NULL;
    char *query;

    query = sqlite3_mprintf("INSERT INTO groups (file_id, groupname) VALUES "
                            " (%" GUINT64_FORMAT ", '%q')", file->file_id, group);
    if (!execute_query(db, query, &err)) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Unable to insert group: %s"), err);
        } else {
            g_warning(_("Unable to insert group: %s"), err);
        }
        g_free(err);
        sqlite3_free(query);

        return FALSE;
    }

    g_free(err);
    sqlite3_free(query);

    return TRUE;
}

static gboolean
insert_file(sqlite3 *db, NNTPCollection *collection, NNTPFile *file, int pos, char **errmsg)
{
    GList *parts;
    GList *groups;
    char *err = NULL;
    char *query;

    query = sqlite3_mprintf("INSERT INTO files ("
                            "   collection_id, position, subject, poster, "
                            "   stamp, tmp_filename, real_filename, status, file_type, "
                            "   par2_start_num, par2_end_num, par2_num_blocks "
                            ") VALUES ("
                            "   %"GUINT64_FORMAT", %i, '%q', '%q', "
                            "   %i, '%q', '%q', %i, %i, "
                            "   %i, %i, %i)",
                            collection->collection_id, pos, file->subject, file->poster,
                            (int) file->stamp, file->tmp_filename, file->real_filename, file->status, file->file_type,
                            file->par2_start_num, file->par2_end_num, file->par2_num_blocks);
    if (!execute_query(db, query, &err)) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Unable to insert file: %s"), err);
        } else {
            g_warning(_("Unable to insert file: %s"), err);
        }
        g_free(err);
        sqlite3_free(query);

        return FALSE;
    }

    file->file_id = sqlite3_last_insert_rowid(db);

    sqlite3_free(query);

    // Insert all the individual groups in the database
    groups = file->groups;
    while (groups) {
        const char *group = groups->data;

        if (!insert_group(db, file, group, errmsg)) {
            return FALSE;
        }

        groups = g_list_next(groups);
    }

    // Insert all the individual parts in the database
    parts = file->parts;
    while (parts) {
        NNTPPart *part = parts->data;

        if (!insert_part(db, file, part, errmsg)) {
            return FALSE;
        }

        parts = g_list_next(parts);
    }

    return TRUE;
}

static gboolean
update_file(sqlite3 *db, NNTPCollection *collection, NNTPFile *file, int pos, char **errmsg)
{
    char *err = NULL;
    char *query;
    GList *parts;

    query = sqlite3_mprintf("UPDATE files SET "
                            "   position = %i, "
                            "   tmp_filename = '%q', "
                            "   real_filename = '%q', "
                            "   status = %i, "
                            "   file_type = %i, "
                            "   par2_start_num = %i, "
                            "   par2_end_num = %i, "
                            "   par2_num_blocks = %i "
                            "WHERE file_id = %"GUINT64_FORMAT,
                            pos, file->tmp_filename, file->real_filename, file->status, file->file_type,
                            file->par2_start_num, file->par2_end_num, file->par2_num_blocks,
                            file->file_id);
    if (!execute_query(db, query, &err)) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Unable to update file: %s"), err);
        } else {
            g_warning(_("Unable to update file: %s"), err);
        }
        g_free(err);
        sqlite3_free(query);

        return FALSE;
    }

    // Update all the individual parts in the database
    parts = file->parts;
    while (parts) {
        NNTPPart *part = parts->data;

        if (!update_part(db, file, part, errmsg)) {
            return FALSE;
        }

        parts = g_list_next(parts);
    }

    g_free(err);
    sqlite3_free(query);

    return TRUE;
}

static gboolean
update_file_position(sqlite3 *db, NNTPCollection *collection, NNTPFile *file, int pos, char **errmsg)
{
    char *err = NULL;
    char *query;

    query = sqlite3_mprintf("UPDATE files SET "
                            "   position = %i "
                            "WHERE file_id = %"GUINT64_FORMAT,
                            pos, file->file_id);
    if (!execute_query(db, query, &err)) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Unable to update file position: %s"), err);
        } else {
            g_warning(_("Unable to update file position: %s"), err);
        }
        g_free(err);
        sqlite3_free(query);

        return FALSE;
    }

    g_free(err);
    sqlite3_free(query);

    return TRUE;
}

static gboolean
insert_or_update_files(sqlite3 *db, NNTPCollection *collection, char **errmsg)
{
    GList *files = collection->files;
    int pos = 0;

    while (files) {
        NNTPFile *file = files->data;

        if (file->file_is_new) {
            file->file_is_new = FALSE;
            if (!insert_file(db, collection, file, pos, errmsg)) {
                return FALSE;
            }
        }

        if (file->file_is_changed) {
            file->file_is_changed = FALSE;
            if (!update_file(db, collection, file, pos, errmsg)) {
                return FALSE;
            }
        }

        if (file->position_is_changed) {
            file->position_is_changed = FALSE;
            if (!update_file_position(db, collection, file, pos, errmsg)) {
                return FALSE;
            }
        }

        pos++;
        files = g_list_next(files);
    }

    return TRUE;
}

static gboolean
insert_collection(sqlite3 *db, NNTPCollection *collection, int pos, char **errmsg)
{
    char *err = NULL;
    char *query;
    gboolean retval = TRUE;

    query = sqlite3_mprintf("INSERT INTO collections (position, collection_name, poster, groups) "
                            " VALUES (%i, '%q', '%q', '%q')", pos, collection->collection_name,
                            collection->poster, collection->groups);
    if (!execute_query(db, query, &err)) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Unable to insert collection: %s"), err);
        } else {
            g_warning(_("Unable to insert collection: %s"), err);
        }

        retval = FALSE;
        goto out;
    }

    collection->collection_id = sqlite3_last_insert_rowid(db);

out:
    g_free(err);
    sqlite3_free(query);

    return retval;
}

static gboolean
update_collection(sqlite3 *db, NNTPCollection *collection, int pos, char **errmsg)
{
    char *err = NULL;
    char *query;
    gboolean retval = TRUE;

    query = sqlite3_mprintf("UPDATE collections SET "
                            "position = %i, "
                            "poster = '%q', "
                            "groups = '%q' "
                            "WHERE collection_id = %"GUINT64_FORMAT,
                            pos, collection->poster, collection->groups, collection->collection_id);
    if (!execute_query(db, query, &err)) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Unable to update collection: %s"), err);
        } else {
            g_warning(_("Unable to update collection: %s"), err);
        }

        retval = FALSE;
        return retval;
    }

    g_free(err);
    sqlite3_free(query);

    return retval;
}

static gboolean
create_tables(sqlite3 *db, char **errmsg)
{
    char *err = NULL;

    if (!execute_query( db,
                        "CREATE TABLE collections ("
                        "   collection_id INTEGER PRIMARY KEY AUTOINCREMENT, "
                        "   position INTEGER NOT NULL, "
                        "   collection_name TEXT NOT NULL, "
                        "   poster TEXT NOT NULL, "
                        "   groups TEXT NOT NULL "
                        ")", &err)) {
        goto error_occured;
    }

    if (!execute_query( db,
                        "CREATE TABLE files ("
                        "   file_id INTEGER PRIMARY KEY AUTOINCREMENT, "
                        "   collection_id INTEGER NOT NULL, "
                        "   position INTEGER NOT NULL, "
                        "   subject STRING NOT NULL, "
                        "   poster STRING NOT NULL, "
                        "   stamp TIMESTAMP NOT NULL, "
                        "   tmp_filename STRING NOT NULL DEFAULT \"\", "
                        "   real_filename STRING NOT NULL DEFAULT \"\", "
                        "   status INTEGER NOT NULL DEFAULT 0, "
                        "   file_type INTEGER NOT NULL DEFAULT 0, "
                        "   par2_start_num INTEGER NOT NULL DEFAULT -1, "
                        "   par2_end_num INTEGER NOT NULL DEFAULT -1, "
                        "   par2_num_blocks INTEGER NOT NULL DEFAULT 0 "
                        ")", &err)) {
        goto error_occured;
    }

    if (!execute_query( db,
                        "CREATE TABLE parts ("
                        "   file_id INTEGER NOT NULL, "
                        "   partnum INTEGER NOT NULL, "
                        "   message_id STRING NOT NULL, "
                        "   size INTEGER NOT NULL, "
                        "   downloaded INTEGER NOT NULL, "
                        "   PRIMARY KEY(file_id, partnum) "
                        ")", &err)) {
        goto error_occured;
    }

    if (!execute_query( db,
                        "CREATE TABLE groups ("
                        "   file_id INTEGER NOT NULL, "
                        "   groupname STRING NOT NULL, "
                        "   PRIMARY KEY(file_id, groupname) "
                        ")", &err)) {
        goto error_occured;
    }

    if (!execute_query(db, "CREATE INDEX parts_idx ON parts(file_id)", &err)) {
        goto error_occured;
    }

    if (!execute_query(db, "CREATE INDEX groups_idx ON groups(file_id)", &err)) {
        goto error_occured;
    }

    return TRUE;

error_occured:
    if (errmsg) {
        *errmsg = g_strdup_printf(_("Unable to create tables: %s"), err);
    } else {
        g_warning(_("Unable to create tables: %s"), err);
    }
    g_free(err);

    return FALSE;
}

gboolean
save_download_queue(GList *queue, char **errmsg)
{
    GList *list;
    char *filename;
    char *dirname;
    gboolean retval = TRUE;
    gboolean first_time = FALSE;
    int pos;
    char *err = NULL;
    gboolean pos_is_changed;
    sqlite3 *db;

    // Generate the folder ~/.config/NNTPGrab in case it didn't exist yet
    filename = generate_filename();
    dirname = g_path_get_dirname(filename);
    g_mkdir_with_parents(dirname, 0700);
    g_free(dirname);

    if (!g_file_test(filename, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR)) {
        // File not found, looks like this is the first time NNTPGrab is ever started, create a new database
        first_time = TRUE;
    }

    g_free(filename);

    if (!(db = db_open(errmsg))) {
        retval = FALSE;
        goto out;
    }

    if (first_time) {
        if (!create_tables(db, errmsg)) {
            retval = FALSE;
            goto out;
        }
    }

    if (!db_begin_transaction(db, &err)) {
        goto error_occured;
    }

    list = queue;
    pos = 0;
    pos_is_changed = FALSE;         // As soon as a position change is detected, update the rest of the collections
    while (list) {
        NNTPCollection *collection = (NNTPCollection *) list->data;

        if (collection->collection_is_new) {
            if (!insert_collection(db, collection, pos, &err)) {
                goto error_occured;
            }

            collection->collection_is_new = FALSE;
            pos_is_changed = TRUE;
        } else if (collection->collection_is_changed || pos_is_changed) {
            // As soon as a position change is detected, update the rest of the collections
            pos_is_changed = TRUE;
            collection->collection_is_changed = FALSE;
            if (!update_collection(db, collection, pos, &err)) {
                goto error_occured;
            }
        }

        if (!insert_or_update_files(db, collection, &err)) {
            goto error_occured;
        }

        pos++;
        list = g_list_next(list);
    }

    if (!db_end_transaction(db, &err)) {
        goto error_occured;
    }

out:
    db_close(db);

    return retval;

error_occured:
    if (errmsg) {
        *errmsg = g_strdup_printf(_("Unable to start or complete the transaction: %s"), err);
    } else {
        g_warning(_("Unable to start or complete the transaction: %s"), err);
    }
    g_free(err);
    retval = FALSE;

    goto out;
}

gboolean
remove_file_from_database(sqlite3 *db, NNTPCollection *collection, NNTPFile *file, gboolean is_in_transaction, char **errmsg)
{
    char *err = NULL;
    char *query = NULL;

    if (!is_in_transaction && !db_begin_transaction(db, &err)) {
        goto error_occured;
    }

    query = sqlite3_mprintf("DELETE FROM files WHERE file_id = %"GUINT64_FORMAT, file->file_id);
    if (!execute_query(db, query, &err)) {
        goto error_occured;
    }
    sqlite3_free(query);

    query = sqlite3_mprintf("DELETE FROM parts WHERE file_id = %"GUINT64_FORMAT, file->file_id);
    if (!execute_query(db, query, &err)) {
        goto error_occured;
    }
    sqlite3_free(query);

    query = sqlite3_mprintf("DELETE FROM groups WHERE file_id = %"GUINT64_FORMAT, file->file_id);
    if (!execute_query(db, query, &err)) {
        goto error_occured;
    }

    if (!is_in_transaction && !db_end_transaction(db, &err)) {
        goto error_occured;
    }
    sqlite3_free(query);

    return TRUE;

error_occured:
    if (query) {
        sqlite3_free(query);
    }

    if (errmsg) {
        *errmsg = g_strdup_printf(_("Unable to remove file: %s"), err);
    } else {
        g_warning(_("Unable to remove file: %s"), err);
    }
    g_free(err);

    return FALSE;
}

gboolean
remove_collection_from_database(sqlite3 *db, NNTPCollection *collection, char **errmsg)
{
    char *err = NULL;
    char *query;
    gboolean retval = TRUE;

    query = sqlite3_mprintf("DELETE FROM collections WHERE collection_id = %"GUINT64_FORMAT, collection->collection_id);
    if (!execute_query(db, query, &err)) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Unable to remove collection: %s"), err);
        } else {
            g_warning(_("Unable to remove collection: %s"), err);
        }

        retval = FALSE;
        return retval;
    }

    g_free(err);
    sqlite3_free(query);

    return retval;
}

void
free_download_queue(GList *queue)
{
    GList *collections;

    collections = queue;
    while (collections) {
        NNTPCollection *collection = (NNTPCollection *) collections->data;

        imported_funcs_collection_unref(collection);

        collections = g_list_next(collections);
    }

    g_list_free(queue);
}
