#ifndef _SCHEDULAR_PLUGIN_H_
#define _SCHEDULAR_PLUGIN_H_

#include <glib-object.h>
#include <sqlite3.h>

typedef struct SchedularPlugin SchedularPlugin;

#define SCHEDULAR_PLUGIN_TYPE_OBJECT       (schedular_plugin_get_type ())
#define SCHEDULAR_PLUGIN(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), SCHEDULAR_PLUGIN_TYPE_OBJECT, SchedularPlugin))
#define SCHEDULAR_PLUGIN_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), SCHEDULAR_PLUGIN_TYPE_OBJECT, SchedularPluginClass))
#define IS_SCHEDULAR_PLUGIN(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), SCHEDULAR_PLUGIN_TYPE_OBJECT))
#define IS_SCHEDULAR_PLUGIN_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), SCHEDULAR_PLUGIN_TYPE_OBJECT))
#define SCHEDULAR_PLUGIN_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), SCHEDULAR_PLUGIN_TYPE_OBJECT, SchedularPluginClass))

GType schedular_plugin_get_type (void);

struct SchedularPlugin
{
    GObject parent;

    // The mutex is used for settings properties used by the schedular thread
    GStaticMutex mutex;
    GList *queue;           // Pointers to NNTPCollection*
    GList *servers;         // Pointers to ConfigServer*

    // These are used to keep track of all the jobs for each usenet server
    GAsyncQueue *task_queue[MAX_NNTP_SERVERS];  // All jobs for each usenet server are queue here
    GCond       *task_cond;
    GMutex      *task_mutex;

    // Keep track of which servers are of which priority
    gboolean     is_high_priority_server[MAX_NNTP_SERVERS];
    gboolean     is_normal_priority_server[MAX_NNTP_SERVERS];
    gboolean     is_low_priority_server[MAX_NNTP_SERVERS];
    gboolean     is_server_enabled[MAX_NNTP_SERVERS];

    // The runtime_mutex is used for the creation and destruction of the schedular thread
    GStaticMutex runtime_mutex;
    GThread *thread;
    gboolean abort_flag;
};

typedef struct _unpack_data {
    NNTPCollection  *collection;
    GThreadPool     *poolPar2verify;
    GThreadPool     *poolPar2repair;
    GThreadPool     *poolUnpack;
} UnpackData;

typedef struct _par2repair_data {
    NNTPCollection  *collection;
    GThreadPool     *poolPar2verify;
    GThreadPool     *poolPar2repair;
    GThreadPool     *poolUnpack;
} PAR2RepairData;

typedef struct _par2verify_data {
    NNTPCollection  *collection;
    NNTPFile        *file;
    GThreadPool     *poolPar2verify;
    GThreadPool     *poolPar2repair;
    GThreadPool     *poolUnpack;
} PAR2VerifyData;

typedef struct _decoder_data {
    SchedularPlugin *schedular;
    NNTPCollection  *collection;
    NNTPFile        *file;
    GThreadPool     *poolPar2verify;
    GThreadPool     *poolPar2repair;
    GThreadPool     *poolUnpack;
} DecoderData;

typedef struct _download_data {
    ConfigServer    *server;
    int              server_id;
    GThreadPool     *poolDecoder;
    GThreadPool     *poolPar2verify;
    GThreadPool     *poolPar2repair;
    GThreadPool     *poolUnpack;
} DownloadData;

// schedular.c
void             test_if_all_downloads_are_completed(SchedularPlugin *schedular, GThreadPool *poolPar2verify, GThreadPool *poolPar2repair, GThreadPool *poolUnpack);
void             request_par2_and_unpack_lock(void);
void             release_par2_and_unpack_lock(void);
gboolean         get_next_part_to_download(SchedularPlugin *schedular, int server_id, NNTPCollection **collection, NNTPFile **file, NNTPPart **part, GThreadPool *poolDecoder);
void             update_part_download_status(NNTPCollection *collection, NNTPFile *file, NNTPPart *part, int server_id, gboolean success, gboolean reset_to_zero, gboolean queue_is_being_paused, gboolean dont_retry_anymore, gpointer nntpconnection);
void             stop_schedular_from_seperate_thread(const char *reason);
gboolean         schedular_plugin_save_queue(char **errmsg);

// unpack_thread.c
void             unpack_thread_func(gpointer data, gpointer user_data);

// par2repair_thread.c
void             par2repair_thread_func(gpointer data, gpointer user_data);

// par2verify_thread.c
void             par2verify_thread_func(gpointer data, gpointer user_data);

// decoder_thread.c
void             decoder_thread_func(gpointer data, gpointer user_data);

// download_thread.c
void             download_thread_func(gpointer data, gpointer user_data);

// imported_funcs.c
void             schedular_plugin_set_imported_funcs(PluginSchedularImportedFuncs funcs);

void             imported_funcs_task_added (const NNTPCollection *collection, const NNTPFile *file);
void             imported_funcs_task_removed (const char *collection_name, const char *subject, guint64 total_file_size, guint64 total_file_size_remaining);
void             imported_funcs_task_state_changed (const char *collection_name, const char *subject, const char *real_filename, TaskState old_state, TaskState new_state);
void             imported_funcs_collection_added (const char *collection_name, const char *poster);
void             imported_funcs_collection_removed (const char *collection_name);
void             imported_funcs_collection_modified (const char *collection_name, const char *poster);
void             imported_funcs_notify_schedular_state_changed (SchedularState new_state, const char *reason);
GList           *imported_funcs_config_get_avail_servers (void);
ConfigServer    *imported_funcs_config_get_server_info (const char *servername);
ConfigOpts       imported_funcs_config_get_opts (void);
gpointer         imported_funcs_nntp_connection_initialize (const char *servername, NNTPGrabErrCode *err);
void             imported_funcs_nntp_connection_destroy (gpointer obj);
NNTPGrabErrCode  imported_funcs_nntp_connection_get_part (gpointer obj, const char *collection_name, const NNTPFile *file, const NNTPPart *part, char **reason);
gboolean         imported_funcs_decode_file (const char *collection_name, NNTPFile *file, int *saved_errno, char **errmsg);
void             imported_funcs_emit_fatal_error (const char *errmsg);
void             imported_funcs_emit_part_download_complete (const char *servername, gpointer nntp_connection, const char *collection_name, const char *subject, int partnum, int size);
void             imported_funcs_emit_part_download_failed (const char *servername, gpointer nntp_connection, const char *collection_name, const char *subject, int partnum, int size, gboolean all_servers_tried);
void             imported_funcs_emit_task_moved (const char *orig_collection_name, const char *subject, const char *new_collection_name, int old_position, int new_position);
void             imported_funcs_emit_collection_moved (const char *collection_name, int old_position, int new_position);
void             imported_funcs_emit_file_download_state_update (const char *collection_name, const char *subject, int num_parts_total, int num_parts_done, int num_parts_failed, guint64 file_size, guint64 file_size_remaining, guint64 total_file_size, guint64 total_file_size_remaining);
void             imported_funcs_perform_par2_repair(const char *collection_name, const char *par2filename);
void             imported_funcs_perform_unpack(const char *collection_name, const char *filename);
void             imported_funcs_emit_all_downloads_completed(void);
NNTPFile        *imported_funcs_file_new(void);
void             imported_funcs_file_ref(NNTPFile *file);
void             imported_funcs_file_unref(NNTPFile *file);
NNTPCollection  *imported_funcs_collection_new(void);
void             imported_funcs_collection_ref(NNTPCollection *collection);
void             imported_funcs_collection_unref(NNTPCollection *collection);
NGRegex         *imported_funcs_regex_compile(const char *regex);
const char     **imported_funcs_regex_match(NGRegex *re, const char *text);
void             imported_funcs_regex_matches_free(const char **matches);

// queue_sqlite.c
sqlite3         *db_open(char **errmsg);
void             db_close(sqlite3 *db);
gboolean         db_begin_transaction(sqlite3 *db, char **errmsg);
gboolean         db_end_transaction(sqlite3 *db, char **errmsg);
gboolean         load_download_queue(GList **ret, char **errmsg);
gboolean         save_download_queue(GList *queue, char **errmsg);
gboolean         remove_file_from_database(sqlite3 *db, NNTPCollection *collection, NNTPFile *file, gboolean is_in_transaction, char **errmsg);
gboolean         remove_collection_from_database(sqlite3 *db, NNTPCollection *collection, char **errmsg);
void             free_download_queue(GList *queue);

#endif /* _SCHEDULAR_PLUGIN_H_ */
