/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <glib.h>
#include <glib/gi18n.h>

#include "nntpgrab_plugin_unpack.h"

#if GLIB_CHECK_VERSION(2,16,0)
#include "fr-archive.h"
#include "fr-command.h"
#include "fr-mime.h"
#endif

static PluginUnpackImportedFuncs imported_funcs;

#if GLIB_CHECK_VERSION(2,16,0)
struct _unpack_data {
    FrArchive *archive;
    gboolean now_loading;
    GMutex *mutex;
    GCond *cond;
    gboolean error_occured;
    FrProcErrorType error_type;
    char errmsg[1024];
    char filename[1024];
    char target_directory[1024];
};
#endif

gboolean
nntpgrab_plugin_unpack_initialize(PluginUnpackImportedFuncs funcs)
{
    imported_funcs = funcs;

#if GLIB_CHECK_VERSION(2,16,0)
    fr_command_register_commands();
#endif

    return TRUE;
}

int
nntpgrab_plugin_unpack_get_version(void)
{
    return NNTPGRAB_PLUGIN_API_VERSION;
}

#if GLIB_CHECK_VERSION(2,16,0)
static void
on_start(FrArchive *archive, FrAction action, gpointer data)
{
g_print("start: action = %i\n", action);
}

static void
on_message(FrArchive *archive, const char *message, gpointer data)
{
    struct _unpack_data *unpack_data = data;
    imported_funcs.unpack_message_received(unpack_data->filename, message);
}

static void
on_progress(FrArchive *archive, double progress, gpointer data)
{
    struct _unpack_data *unpack_data = data;
    imported_funcs.unpack_progress_update(unpack_data->filename, progress);
}

static void
on_working_archive(FrArchive *archive, const char *filename, gpointer data)
{
    struct _unpack_data *unpack_data = data;
    char *filename_local = g_filename_from_uri(filename, NULL, NULL);
    char *filename_short;

    if (filename_local) {
        filename_short = g_path_get_basename(filename_local);
        imported_funcs.unpack_working_archive_changed(unpack_data->filename, filename_short);
    } else {
        filename_short = g_path_get_basename(filename);
        imported_funcs.unpack_working_archive_changed(unpack_data->filename, filename_short);
    }

    g_free(filename_short);
    g_free(filename_local);
}

static gboolean
do_start_unpack(gpointer data)
{
    struct _unpack_data *unpack_data = data;
g_print("Now starting with extract phase\n");
    unpack_data->now_loading = FALSE;
    fr_process_clear (unpack_data->archive->process);
    fr_archive_extract_to_local(unpack_data->archive, NULL, unpack_data->target_directory, NULL, FALSE, TRUE, FALSE, NULL);
    fr_process_start (unpack_data->archive->process);

    return FALSE;
}

static void
on_done(FrArchive *archive, FrAction action, FrProcError *error, gpointer data)
{
    /* Signal the unpack thread that we're finished */
    struct _unpack_data *unpack_data = data;

g_print("error->type = %i\n", error->type);
g_print("error->status = %i\n", error->status);
if (error->gerror) {
g_print("error->gerror->message = %s\n", error->gerror->message);
}

    g_print("done\n");

    unpack_data->error_type = error->type;
    unpack_data->error_occured = (error->status != 0 || error->type != 0);

    if (error->gerror && error->gerror->message) {
        strncpy(unpack_data->errmsg, error->gerror->message, sizeof(unpack_data->errmsg) - 1);
    } else if (error->status != 0) {
        strcpy(unpack_data->errmsg, "Unknown error occured while unpacking");
    }

    if (action == FR_ACTION_LISTING_CONTENT && !unpack_data->error_occured) {
        g_idle_add(do_start_unpack, unpack_data);
    }

    if (action == FR_ACTION_EXTRACTING_FILES || unpack_data->error_occured) {
        on_progress(archive, 1.0, unpack_data);
        if (unpack_data->error_occured) {
            on_message(archive, unpack_data->errmsg, unpack_data);
        }

        /* Send a signal to the unpack thread that we're done */
        g_mutex_lock(unpack_data->mutex);
        g_cond_signal(unpack_data->cond);
        g_mutex_unlock(unpack_data->mutex);
    }
}
#endif

gboolean
nntpgrab_plugin_unpack_do_unpack(const char *filename, const char *target_directory, char **errmsg)
{
#if GLIB_CHECK_VERSION(2,16,0)
    struct _unpack_data *unpack_data;
    FrArchive *archive = fr_archive_new();
    gboolean ret;
    char *filename_uri;
    char *filename_short;
    GError *err = NULL;

    filename_uri = g_filename_to_uri(filename, NULL, &err);
    if (!filename_uri) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Unable to convert filename to URI: %s (filename = %s)"), err->message, filename);
        } else {
            g_warning(_("Unable to convert filename to URI: %s (filename = %s)"), err->message, filename);
        }
        g_error_free(err);
        return FALSE;
    }

    /* The File Roller code is heavily integrated in the GLib mainloop
     * As this is a seperate thread, we just have to wait here until
     * we have received a signal that the extracting has completed */
    unpack_data = g_slice_new0(struct _unpack_data);
    unpack_data->cond = g_cond_new();
    unpack_data->mutex = g_mutex_new();
    unpack_data->archive = archive;
    filename_short = g_path_get_basename(filename);
    strncpy(unpack_data->filename, filename_short, sizeof(unpack_data->filename) - 1);
    g_free(filename_short);
    strncpy(unpack_data->target_directory, target_directory, sizeof(unpack_data->target_directory) - 1);

    g_signal_connect(archive, "start", G_CALLBACK(on_start), unpack_data);
    g_signal_connect(archive, "done", G_CALLBACK(on_done), unpack_data);
    g_signal_connect(archive, "progress", G_CALLBACK(on_progress), unpack_data);
    g_signal_connect(archive, "working_archive", G_CALLBACK(on_working_archive), unpack_data);
    g_signal_connect(archive, "message", G_CALLBACK(on_message), unpack_data);

    if (!fr_archive_load_local(archive, filename_uri, NULL)) {
        if (errmsg) {
            *errmsg = g_strdup_printf("Unable to load files to unpack");
        }

        g_object_unref(archive);
        g_slice_free(struct _unpack_data, unpack_data);
        g_free(filename_uri);

        return FALSE;
    }

    /* Wait for a confirmation that extraction was successfull */
    g_mutex_lock(unpack_data->mutex);
    g_cond_wait(unpack_data->cond, unpack_data->mutex);
    g_mutex_unlock(unpack_data->mutex);

    ret = !unpack_data->error_occured;
    if (!ret && errmsg) {
        *errmsg = g_strdup(unpack_data->errmsg);
    }

    g_mutex_free(unpack_data->mutex);
    g_cond_free(unpack_data->cond);
    g_object_unref(archive);
    g_slice_free(struct _unpack_data, unpack_data);

    g_free(filename_uri);

    return ret;
#else
    /* GLib version too old, assume extraction was successfull */
    return TRUE;
#endif
}
