/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef _MSC_VER
#include "config.h.win32"
#else
#include "config.h"
#endif

#include <glib.h>
#include <glib-object.h>

#include "nntpgrab.h"
#include "nntpgrab_utils.h"

static void
config_changed (NNTPGrabCore *core, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup("CONFIG_CHANGED\n"));
}

static void
part_download_start (NNTPGrabCore *core, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("PART_DOWNLOAD_START %s\t%i\t%s\t%s\t%i\n", servername, conn_id, collection_name, subject, part_num));
}

static void
part_done (NNTPGrabCore *core, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int size, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("PART_DONE %s\t%i\t%s\t%s\t%i\t%i\n", servername, conn_id, collection_name, subject, part_num, size));
}

static void
part_failed (NNTPGrabCore *core, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int size, gboolean all_servers_tried, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    const char *all_servers_tried_str;

    if (all_servers_tried) {
        all_servers_tried_str = "TRUE";
    } else {
        all_servers_tried_str = "FALSE";
    }

    g_async_queue_push(queue, g_strdup_printf("PART_FAILED %s\t%i\t%s\t%s\t%i\t%i\t%s\n", servername, conn_id, collection_name, subject, part_num, size, all_servers_tried_str));
}

static void
traffic_monitor_update (NNTPGrabCore *core, int bytes_received1, int bytes_received2, int bytes_received3, int bytes_received4, int bytes_received5, int bytes_received6, int bytes_received7, int bytes_received8, int bytes_received9, int bytes_received10, guint64 stamp, double average, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("TRAFFIC_MONITOR_UPDATE %i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%"G_GUINT64_FORMAT"\t%.2f\n", bytes_received1, bytes_received2, bytes_received3, bytes_received4, bytes_received5, bytes_received6, bytes_received7, bytes_received8, bytes_received9, bytes_received10, (guint64) stamp, average));
}

static void
part_progress_update (NNTPGrabCore *core, const char *servername, int conn_id, const char *subject, int part_num, int bytes_downloaded, int bytes_total, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("PART_PROGRESS_UPDATE %s\t%i\t%s\t%i\t%i\t%i\n", servername, conn_id, subject, part_num, bytes_downloaded, bytes_total));
}

static void
collection_added (NNTPGrabCore *core, const char *collection_name, const char *poster, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("COLLECTION_ADDED %s\t%s\n", collection_name, poster));
}

static void
collection_removed (NNTPGrabCore *core, const char *collection_name, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("COLLECTION_REMOVED %s\n", collection_name));
}

static void
collection_modified (NNTPGrabCore *core, const char *collection_name, const char *poster, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("COLLECTION_MODIFIED %s\t%s\n", collection_name, poster));
}

static void
file_added (NNTPGrabCore *core, const char *collection_name, const char *subject, const char *poster, guint64 stamp, guint64 file_size, guint64 total_size, guint64 total_size_remaining, TaskState status, int num_parts, GList *groups, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    char *groups_str = NULL;
    GList *list;
    char *status_str;

    list = groups;
    while (list) {
        char *tmp;
        const char *groupname = list->data;

        if (groups_str) {
            tmp = g_strdup_printf("%s|%s", groups_str, groupname);
            g_free(groups_str);
            groups_str = tmp;
        } else {
            groups_str = g_strdup(groupname);
        }

        list = g_list_next(list);
    }

    switch (status) {
        case TASK_STATE_WAITING_FOR_DOWNLOAD:
            status_str = "WAITING_FOR_DOWNLOAD";
            break;

        case TASK_STATE_DOWNLOADING:
            status_str = "DOWNLOADING";
            break;

        case TASK_STATE_WAITING_FOR_DECODE:
            status_str = "WAITING_FOR_DECODE";
            break;

        case TASK_STATE_DECODING:
            status_str = "DECODING";
            break;

        case TASK_STATE_SKIPPED:
            status_str = "SKIPPED";
            break;

        case TASK_STATE_FINISHED_INCOMPLETE:
            status_str = "FINISHED_INCOMPLETE";
            break;

        case TASK_STATE_FINISHED_NO_PARTS_AVAIL:
            status_str = "FINISHED_NO_PARTS_AVAIL";
            break;

        case TASK_STATE_FINISHED_COMPLETE:
        default:
            status_str = "FINISHED_COMPLETE";
            break;
    };

    g_async_queue_push(queue, g_strdup_printf("FILE_ADDED %s\t%s\t%s\t%"G_GUINT64_FORMAT"\t%"G_GUINT64_FORMAT"\t%"G_GUINT64_FORMAT"\t%"G_GUINT64_FORMAT"\t%s\t%i\t%s\n", collection_name, subject, poster, (guint64) stamp, file_size, total_size, total_size_remaining, status_str, num_parts, groups_str));

    g_free(groups_str);
}

static void
file_removed (NNTPGrabCore *core, const char *collection_name, const char *subject, guint64 total_file_size, guint64 total_file_size_remaining, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("FILE_REMOVED %s\t%s\t%"G_GUINT64_FORMAT"\t%"G_GUINT64_FORMAT"\n", collection_name, subject, total_file_size, total_file_size_remaining));
}

static void
file_download_state_update (NNTPGrabCore *core, const char *collection_name, const char *subject, int num_parts_total, int num_parts_done, int num_parts_failed, guint64 file_size, guint64 file_size_remaining, guint64 total_file_size, guint64 total_file_size_remaining, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("FILE_DOWNLOAD_STATE_UPDATE %s\t%s\t%i\t%i\t%i\t%"G_GUINT64_FORMAT"\t%"G_GUINT64_FORMAT"\t%"G_GUINT64_FORMAT"\t%"G_GUINT64_FORMAT"\n", collection_name, subject, num_parts_total, num_parts_done, num_parts_failed, file_size, file_size_remaining, total_file_size, total_file_size_remaining));
}

static void
file_state_changed (NNTPGrabCore *core, const char *collection_name, const char *subject, const char *real_filename, TaskState old_state, TaskState new_state, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    const char *real_filename_str;
    char *old_state_str;
    char *new_state_str;

    if (real_filename) {
        real_filename_str = real_filename;
    } else {
        real_filename_str = "";
    }

    switch(old_state) {
        case TASK_STATE_WAITING_FOR_DOWNLOAD:
            old_state_str = "WAITING_FOR_DOWNLOAD";
            break;

        case TASK_STATE_DOWNLOADING:
            old_state_str = "DOWNLOADING";
            break;

        case TASK_STATE_WAITING_FOR_DECODE:
            old_state_str = "WAITING_FOR_DECODE";
            break;

        case TASK_STATE_DECODING:
            old_state_str = "DECODING";
            break;

        case TASK_STATE_SKIPPED:
            old_state_str = "SKIPPED";
            break;

        case TASK_STATE_FINISHED_INCOMPLETE:
            old_state_str = "FINISHED_INCOMPLETE";
            break;

        case TASK_STATE_FINISHED_NO_PARTS_AVAIL:
            old_state_str = "FINISHED_NO_PARTS_AVAIL";
            break;

        case TASK_STATE_FINISHED_COMPLETE:
        default:
            old_state_str = "FINISHED_COMPLETE";
            break;
    };

    switch(new_state) {
        case TASK_STATE_WAITING_FOR_DOWNLOAD:
            new_state_str = "WAITING_FOR_DOWNLOAD";
            break;

        case TASK_STATE_DOWNLOADING:
            new_state_str = "DOWNLOADING";
            break;

        case TASK_STATE_WAITING_FOR_DECODE:
            new_state_str = "WAITING_FOR_DECODE";
            break;

        case TASK_STATE_DECODING:
            new_state_str = "DECODING";
            break;

        case TASK_STATE_SKIPPED:
            new_state_str = "SKIPPED";
            break;

        case TASK_STATE_FINISHED_INCOMPLETE:
            new_state_str = "FINISHED_INCOMPLETE";
            break;

        case TASK_STATE_FINISHED_NO_PARTS_AVAIL:
            new_state_str = "FINISHED_NO_PARTS_AVAIL";
            break;

        case TASK_STATE_FINISHED_COMPLETE:
        default:
            new_state_str = "FINISHED_COMPLETE";
            break;
    };

    g_async_queue_push(queue, g_strdup_printf("FILE_STATE_CHANGED %s\t%s\t%s\t%s\t%s\n", collection_name, subject, real_filename_str, old_state_str, new_state_str));
}

static void
connection_connecting (NNTPGrabCore *core, const char *servername, int conn_id, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("CONNECTION_CONNECTING %s\t%i\n", servername, conn_id));
}

static void
connection_connected (NNTPGrabCore *core, const char *servername, int conn_id, const char *welcome_msg, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("CONNECTION_CONNECTED %s\t%i\t%s\n", servername, conn_id, welcome_msg));
}

static void
connection_disconnect (NNTPGrabCore *core, const char *servername, int conn_id, NNTPDisconnectType disconnect_type, const char *msg, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    char *disconnect_type_str;

    switch (disconnect_type) {
        case DISCONNECT_NORMAL:
            disconnect_type_str = "NORMAL";
            break;

        case DISCONNECT_NO_SUCH_HOST:
            disconnect_type_str = "NO_SUCH_HOST";
            break;

        case DISCONNECT_CONNECTION_REFUSED:
            disconnect_type_str = "CONNECTION_REFUSED";
            break;

        case DISCONNECT_TOO_MANY_CONNECTIONS:
            disconnect_type_str = "TOO_MANY_CONNECTIONS";
            break;

        case DISCONNECT_CONNECT_TIMEOUT:
            disconnect_type_str = "CONNECT_TIMEOUT";
            break;

        case DISCONNECT_READ_ERROR:
            disconnect_type_str = "READ_ERROR";
            break;

        case DISCONNECT_READ_TIMEOUT:
            disconnect_type_str = "READ_TIMEOUT";
            break;

        case DISCONNECT_WRITE_ERROR:
            disconnect_type_str = "WRITE_ERROR";
            break;

        case DISCONNECT_IDLE_TIMEOUT:
            disconnect_type_str = "IDLE_TIMEOUT";
            break;

        case DISCONNECT_INVALID_MSG:
            disconnect_type_str = "INVALID_MSG";
            break;

        case DISCONNECT_LOGIN_FAILURE:
            disconnect_type_str = "LOGIN_FAILURE";
            break;

        case DISCONNECT_ERROR_SSL_INITIALISE:
            disconnect_type_str = "ERROR_SSL_INITIALISE";
            break;

        case DISCONNECT_UNEXPECTED:
        default:
            disconnect_type_str = "UNEXPECTED";
            break;
    };

    g_async_queue_push(queue, g_strdup_printf("CONNECTION_DISCONNECT %s\t%i\t%s\t%s\n", servername, conn_id, disconnect_type_str, (msg) ? msg : ""));
}

static void
schedular_state_changed (NNTPGrabCore *core, SchedularState new_state, const char *reason, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    char *reason2 = NULL;

    // strip out any newlines
    if (reason) {
        char *ptr;

        reason2 = g_strdup(reason);
        ptr = reason2;

        while (*ptr) {
            if (*ptr == '\n') {
                *ptr = ' ';
            }
            ptr++;
        }
    }

    switch (new_state) {
        case SCHEDULAR_STATE_RUNNING:
            g_async_queue_push(queue, g_strdup("SCHEDULAR_STATE_CHANGED RUNNING\t\n"));
            break;

        case SCHEDULAR_STATE_STOPPING:
            if (reason2) {
                g_async_queue_push(queue, g_strdup_printf("SCHEDULAR_STATE_CHANGED STOPPING\t%s\n", reason2));
            } else {
                g_async_queue_push(queue, g_strdup("SCHEDULAR_STATE_CHANGED STOPPING\t\n"));
            }

            break;

        case SCHEDULAR_STATE_STOPPED:
        default:
            g_async_queue_push(queue, g_strdup("SCHEDULAR_STATE_CHANGED STOPPED\t\n"));
            break;
    };

    if (reason2) {
        g_free(reason2);
    }
}

static void
fatal_error (NNTPGrabCore *core, const char *errmsg, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    char *msg = g_strdup(errmsg);
    char *ptr = msg;

    // strip out any newlines
    while (*ptr) {
        if (*ptr == '\n') {
            *ptr = ' ';
        }
        ptr++;
    }

    g_async_queue_push(queue, g_strdup_printf("FATAL_ERROR %s\n", msg));

    g_free(msg);
}

static void
warning_message (NNTPGrabCore *core, const char *warn_msg, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    char *msg = g_strdup(warn_msg);
    char *ptr = msg;

    // strip out any newlines
    while (*ptr) {
        if (*ptr == '\n') {
            *ptr = ' ';
        }
        ptr++;
    }

    g_async_queue_push(queue, g_strdup_printf("WARNING_MESSAGE %s\n", msg));

    g_free(msg);
}

static void
debug_message (NNTPGrabCore *core, GThread *thread, const char *debug_msg, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    char *msg = g_strdup(debug_msg);
    char *ptr = msg;

    // strip out any newlines
    while (*ptr) {
        if (*ptr == '\n') {
            *ptr = ' ';
        }
        ptr++;
    }

    g_async_queue_push(queue, g_strdup_printf("DEBUG_MESSAGE %s\n", msg));

    g_free(msg);
}

static void
task_moved (NNTPGrabCore *core, const char *orig_collection_name, const char *subject, const char *new_collection_name, int old_position, int new_position, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("TASK_MOVED %s\t%s\t%s\t%i\t%i\n", orig_collection_name, subject, new_collection_name, old_position, new_position));
}

static void
collection_moved (NNTPGrabCore *core, const char *collection_name, int old_position, int new_position, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("COLLECTION_MOVED %s\t%i\t%i\n", collection_name, old_position, new_position));
}

static void
par2_repair_message (NNTPGrabCore *core, const char *msg, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    char *msg_stripped = g_strdup(msg);
    char *ptr = msg_stripped;

    // strip out any newlines
    while (*ptr) {
        if (*ptr == '\n') {
            *ptr = ' ';
        }
        ptr++;
    }

    g_async_queue_push(queue, g_strdup_printf("PAR2_REPAIR_MESSAGE %s\n", msg_stripped));

    g_free(msg_stripped);
}

static void
all_downloads_completed (NNTPGrabCore *core, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup("ALL_DOWNLOADS_COMPLETED\n"));
}

static void
unpack_progress_update (NNTPGrabCore *core, const char *collection_name, const char *filename, double progress, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("UNPACK_PROGRESS_UPDATE %s\t%s\t%.2f\n", collection_name, filename, progress));
}

static void
unpack_message_received (NNTPGrabCore *core, const char *collection_name, const char *filename, const char *message, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("UNPACK_MESSAGE_RECEIVED %s\t%s\t%s\n", collection_name, filename, message));
}

static void
unpack_working_archive_changed (NNTPGrabCore *core, const char *collection_name, const char *filename, const char *working_archive, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("UNPACK_WORKING_ARCHIVE_CHANGED %s\t%s\t%s\n", collection_name, filename, working_archive));
}

static void
par2_begin_verify (NNTPGrabCore *core, const char *collection_name, const char *active_par2_filename, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("PAR2_BEGIN_VERIFY %s\t%s\n", collection_name, active_par2_filename));
}

static void
par2_load_progress_update (NNTPGrabCore *core, const char *collection_name, const char *active_par2_filename, const char *filename, double progress, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("PAR2_LOAD_PROGRESS_UPDATE %s\t%s\t%s\t%.2f\n", collection_name, active_par2_filename, filename, progress));
}

static void
par2_recovery_file_loaded (NNTPGrabCore *core, const char *collection_name, const char *active_par2_filename, const char *filename, int num_new_packets, int num_blocks_found, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("PAR2_RECOVERY_FILE_LOADED %s\t%s\t%s\t%i\t%i\n", collection_name, active_par2_filename, filename, num_new_packets, num_blocks_found));
}

static void
par2_file_loaded (NNTPGrabCore *core, const char *collection_name, const char *active_par2_filename, char *filename, PAR2FileState state, int num_blocks_found, int num_blocks_expected, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    const char *state_str;

    switch (state) {
        case PAR2_FILE_STATE_MISSING:
            state_str = "MISSING";
            break;

        case PAR2_FILE_STATE_FOUND:
            state_str = "FOUND";
            break;

        case PAR2_FILE_STATE_DAMAGED:
            state_str = "DAMAGED";
            break;

        case PAR2_FILE_STATE_NO_NEW_BLOCKS_FOUND:
            state_str = "NO_NEW_BLOCKS_FOUND";
            break;

        default:
            g_print(__FILE__ ":%i Invalid state detected: %i\n", __LINE__, state);
            state_str = "MISSING";
    }

    g_async_queue_push(queue, g_strdup_printf("PAR2_FILE_LOADED %s\t%s\t%s\t%s\t%i\t%i\n", collection_name, active_par2_filename, filename, state_str, num_blocks_found, num_blocks_expected));
}

static void
par2_repair_progress_update (NNTPGrabCore *core, const char *collection_name, const char *active_par2_filename, double progress, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("PAR2_REPAIR_PROGRESS_UPDATE %s\t%s\t%.2f\n", collection_name, active_par2_filename, progress));
}

static void
par2_repair_failure (NNTPGrabCore *core, const char *collection_name, const char *active_par2_filename, int num_blocks_more_required, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("PAR2_REPAIR_FAILURE %s\t%s\t%i\n", collection_name, active_par2_filename, num_blocks_more_required));
}

static void
par2_repair_success (NNTPGrabCore *core, const char *collection_name, const char *active_par2_filename, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("PAR2_REPAIR_SUCCESS %s\t%s\n", collection_name, active_par2_filename));
}

static void
par2_no_repair_required (NNTPGrabCore *core, const char *collection_name, const char *active_par2_filename, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("PAR2_NO_REPAIR_REQUIRED %s\t%s\n", collection_name, active_par2_filename));
}

void
connect_signal_handlers(NNTPGrabCore *core, GAsyncQueue *queue)
{
    gpointer data = queue;

    ng_signal_connect(core, "config_changed", NG_CALLBACK(config_changed), data);
    ng_signal_connect(core, "part_download_start", NG_CALLBACK(part_download_start), data);
    ng_signal_connect(core, "part_done", NG_CALLBACK(part_done), data);
    ng_signal_connect(core, "part_failed", NG_CALLBACK(part_failed), data);
    ng_signal_connect(core, "traffic_monitor_update", NG_CALLBACK(traffic_monitor_update), data);
    ng_signal_connect(core, "part_progress_update", NG_CALLBACK(part_progress_update), data);
    ng_signal_connect(core, "collection_added", NG_CALLBACK(collection_added), data);
    ng_signal_connect(core, "collection_removed", NG_CALLBACK(collection_removed), data);
    ng_signal_connect(core, "collection_modified", NG_CALLBACK(collection_modified), data);
    ng_signal_connect(core, "file_added", NG_CALLBACK(file_added), data);
    ng_signal_connect(core, "file_removed", NG_CALLBACK(file_removed), data);
    ng_signal_connect(core, "file_download_state_update", NG_CALLBACK(file_download_state_update), data);
    ng_signal_connect(core, "file_state_changed", NG_CALLBACK(file_state_changed), data);
    ng_signal_connect(core, "connection_connecting", NG_CALLBACK(connection_connecting), data);
    ng_signal_connect(core, "connection_connected", NG_CALLBACK(connection_connected), data);
    ng_signal_connect(core, "connection_disconnect", NG_CALLBACK(connection_disconnect), data);
    ng_signal_connect(core, "schedular_state_changed", NG_CALLBACK(schedular_state_changed), data);
    ng_signal_connect(core, "fatal_error", NG_CALLBACK(fatal_error), data);
    ng_signal_connect(core, "warning_message", NG_CALLBACK(warning_message), data);
    ng_signal_connect(core, "debug_message", NG_CALLBACK(debug_message), data);
    ng_signal_connect(core, "task_moved", NG_CALLBACK(task_moved), data);
    ng_signal_connect(core, "collection_moved", NG_CALLBACK(collection_moved), data);
    ng_signal_connect(core, "par2_repair_message", NG_CALLBACK(par2_repair_message), data);
    ng_signal_connect(core, "all_downloads_completed", NG_CALLBACK(all_downloads_completed), data);
    ng_signal_connect(core, "unpack_progress_update", NG_CALLBACK(unpack_progress_update), data);
    ng_signal_connect(core, "unpack_message_received", NG_CALLBACK(unpack_message_received), data);
    ng_signal_connect(core, "unpack_working_archive_changed", NG_CALLBACK(unpack_working_archive_changed), data);
    ng_signal_connect(core, "par2_begin_verify", NG_CALLBACK(par2_begin_verify), data);
    ng_signal_connect(core, "par2_load_progress_update", NG_CALLBACK(par2_load_progress_update), data);
    ng_signal_connect(core, "par2_recovery_file_loaded", NG_CALLBACK(par2_recovery_file_loaded), data);
    ng_signal_connect(core, "par2_file_loaded", NG_CALLBACK(par2_file_loaded), data);
    ng_signal_connect(core, "par2_repair_progress_update", NG_CALLBACK(par2_repair_progress_update), data);
    ng_signal_connect(core, "par2_repair_failure", NG_CALLBACK(par2_repair_failure), data);
    ng_signal_connect(core, "par2_repair_success", NG_CALLBACK(par2_repair_success), data);
    ng_signal_connect(core, "par2_no_repair_required", NG_CALLBACK(par2_no_repair_required), data);
}

void
disconnect_signal_handlers(NNTPGrabCore *core, GAsyncQueue *queue)
{
    gpointer data = queue;

    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(config_changed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(part_download_start), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(part_done), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(part_failed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(traffic_monitor_update), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(part_progress_update), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(collection_added), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(collection_removed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(collection_modified), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(file_added), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(file_removed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(file_download_state_update), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(file_state_changed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(connection_connecting), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(connection_connected), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(connection_disconnect), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(schedular_state_changed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(fatal_error), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(warning_message), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(debug_message), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(task_moved), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(collection_moved), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(par2_repair_message), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(all_downloads_completed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(unpack_progress_update), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(unpack_message_received), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(unpack_working_archive_changed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(par2_begin_verify), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(par2_load_progress_update), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(par2_recovery_file_loaded), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(par2_file_loaded), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(par2_repair_progress_update), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(par2_repair_failure), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(par2_repair_success), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(par2_no_repair_required), data);
}
